/*	File:		vfp.h

	Contains:	Floating-Point Definitions and Declarations for Vector Math Lib. 

	Version:	1.0

	Written by:	Helder J Ramalho

	Copyright:	 1999 by Apple Computer, Inc., all rights reserved.

	File Ownership:

		DRI:				Helder J Ramalho

		Other Contact:		Anyone who desires

		Technology:			Vector Math Library

	Writers:

		(HJR)	Helder Ramalho

	Change History (most recent first):

	 <vMTH6>	 8/19/99	HJR		Cleaned up some typos.
	 <vMTH5>	 7/16/99	HJR		Added vTanf.
	 <vMTH4>	 7/16/99	HJR		Added hyperbolic functions and a couple others.
	 <vMTH3>	  6/3/99	HJR		Cleaned up comments a little.
	 <vMTH2>	  6/3/99	HJR		Added vRSqrtf, vASinf, vACosf, vATanf, vATan2f, vIPowf and
									vPowf.
		<1>	 	  2/11/99	HJR		New today.

*/

//
//										 										|
//    A collection of numerical functions designed to facilitate a wide         |
//    range of numerical programming for the Altivect Programming model.		|
//										 										|
//

//[ Computational Functions]
//										 										|
//	Divf		C = A  B  														|
//	Sqrtf		B = A													  		|
//	RSqrtf		B = 1/A													  	|
//										 										|
//

vector float 	vDivf (  vector float A, vector float B );
vector float 	vSqrtf ( vector float X );
vector float	vRSqrtf ( vector float X );

//[ Exponential Functions]
//										 										|
//	Exp			B = Exp(A)												  		|
//	ExpM1		ExpM1(x) = Exp(x) - 1.  But, for small enough arguments,  		|
//				ExpM1(x) is expected to be more accurate than Exp(x) - 1.		|
//	Log			B = Log(A)											  			|
//	Log1P		Log1P = Log(1 + x). But, for small enough arguments,  			|
//     			Log1P is expected to be more accurate than Log(1 + x).  		|
//	Logb		Extracts the exponent of its argument, as a signed integral		|
//      		value. A subnormal argument is treated as though it were first	|
//      		normalized. Thus:												|
//      			1   <=   x * 2^(-logb(x))   <   2 							|
//	Scalb		Computes x * 2^n efficently.  This is not normally done by		|
//      		computing 2^n explicitly. 										|
//										 										|
//

vector float 	vExpf ( vector float X );
vector float 	vExpM1f ( vector float X );

vector float 	vLogf ( vector float X );
vector float 	vLog1Pf ( vector float X );

vector float 	vLogbf ( vector float X );
vector float 	vScalbf ( vector float X, vector signed long n );

//[ Auxiliary Functions]
//										 										|
//	fab				Absolute value is part of the programming model, however 	|
//					completeness it is included in the library.					|				
// 	copysign		Produces a value with the magnitude of its first argument	|
//      			and sign of its second argument.  NOTE: the order of the 	|
//      			arguments matches the recommendation of the IEEE 754 		|
//					floating point standard,  which is opposite from the SANE	|
//					copysign function. 											|
// 	nextafter		Computes the next representable value after 'x' in the 		|
//					direction of 'y'.  if x == y, then y is returned.			|
//										 										|
//

vector float 	vFabf ( vector float v );
vector float 	vCopySignf ( vector float arg2, vector float arg1 );
vector float 	vNextafterf ( vector float x, vector float y );

//[ Inquiry Functions]
//										 										|
// 	classify		Returns one of the FP_ values.								|
//	signbit			Non-zero if and only if the sign of the argument x is		|
//     				negative.  This includes, NaNs, infinities and zeros. 		|
//																				|
//

vector unsigned long	vClassifyf ( vector float arg );
vector unsigned long	vSignBitf ( vector float arg );

//[ Transandental Functions]
//										 										|
// 	sin				B = Sin(A).													|
//	cos				B = Cos(A).                                          		|
//	tan				B = Tan(A).													|
//																				|
//

vector float	vSinf ( vector float arg );
vector float	vCosf ( vector float arg );
vector float	vTanf ( vector float arg );

//[ Trigonometric Functions]
//										 										|
//  asin        result is in [-pi/2,pi/2].                                      |
// 	acos        result is in [0,pi].                                            |
//  atan        result is in [-pi/2,pi/2].                                      |
//  atan2       Computes the arc tangent of y/x in [-pi,pi] using the sign of   |
//              both arguments to determine the quadrant of the computed value. |
//																				|
//

vector float	vASinf ( vector float arg );
vector float	vACosf ( vector float arg );
vector float	vATanf ( vector float arg );
vector float	vATan2f ( vector float arg1, vector float arg2 );

//[ Hyperbolic Functions]
//										 										|
//  Sinh        Sine Hyperbolic.                                                |
// 	Cosh        Cosine Hyperbolic.                                              |
//  Tanh        Tangent Hyperbolic.			                                    |
//																				|
//

vector float 	vSinhf ( vector float X );
vector float 	vCoshf ( vector float X );
vector float 	vTanhf ( vector float X );

//[ Remainder Functions]
//										 										|
// 	mod				B = X mod Y.												|
// 	remainder		IEEE 754 floating point standard for remainder.				|
// 	remquo			SANE remainder.  It stores into 'quotient' the 7 low-order	|
//					bits of the integer quotient x/y, such that:				|
//						-127 <= quotient <= 127. 								|
//																				|
//

vector float 	vFModf ( vector float X, vector float Y);
vector float 	vRemainderf ( vector float X, vector float Y);
vector float 	vRemQuof ( vector float X, vector float Y, vector unsigned long *QUO );

//[ Power Functions]
//										 										|
// 	ipow			Returns x raised to the integer power of y.  				|
// 	pow				Returns x raised to the power of y.  Result is more 		|
//					accurate than using exp(log(x)*y).							|
//																				|
//

vector float	vIPowf ( vector float X, vector signed long Y );
vector float	vPowf ( vector float X, vector float Y );

//
//	Useful								 										|
//

vector unsigned long 	vTableLookUp ( vector signed long Index_Vect, unsigned long * Table );
