/*#io
Socket ioDoc(
docCopyright("Steve Dekorte", 2004)
docLicense("BSD revised")
docDependsOn("SocketManager")
docObject("Socket")
docDescription("""Interface to network communication. Sockets will auto yields to other coroutines while waiting on a request. All blocking operations use the timeout settings of the socket. Reads are appended to the socket's read buffer which can be accessed using the readBuffer method. Example:

<pre>
  socket := Socket clone setHostName("www.yahoo.com") setPort(80) connect 
  if(socket error) then( write(socket error, "\n"); exit) 
  
  socket write("GET /\n\n")
  
  while(socket read, Nop)
  if(socket error) then(write(socket error, "\n"); exit) 
  
  write("read ", socket readBuffer length, " bytes\n")
</pre>""")
*/


/*#io
    docSlot("setHostName(hostName)", 
    "Translates hostName to an IP using asynchronous DNS and sets the host attribute. Returns self.")
*/
    
#include "IoSocket.h"
#include "IoSocketManager.h"
#include "IoState.h"
#include "IoNumber.h"
#include "IoNil.h"
#include "IoSeq.h"
#include "ByteArray.h"
#include "List.h"

#define DATA(self) ((IoSocketData *)self->data)

void IoSocket_tagCleanup(IoTag *self)
{
    #ifdef WIN32
    WSACleanup();
    #endif    
}

IoTag *IoSocket_tag(void *state)
{
    IoTag *tag = IoTag_newWithName_("Socket");
    tag->state = state;
    tag->freeFunc  = (TagFreeFunc *)IoSocket_free;
    tag->cloneFunc = (TagCloneFunc *)IoSocket_rawClone;
    tag->markFunc  = (TagMarkFunc *)IoSocket_mark;
    tag->tagCleanupFunc  = (TagMarkFunc *)IoSocket_tagCleanup;

    #ifdef WIN32
    {
    WSADATA wsd;
    WSAStartup(MAKEWORD(2,0), &wsd);
    }
    #endif

    return tag;
}

void IoSocket_init(IoSocket *self)
{
    IoSocketManager *sm = IoSocketManager_socketManagerForState_(IOSTATE);
    Descriptor *d = DATA(self)->socket->descriptor;
    DATA(self)->ioSocketManager = sm;
    Descriptor_socketManager_(d, IoSocketManager_rawSocketManager(sm));
    Descriptor_scheduler_(d,  IOSTATE->scheduler);
    Socket_externalBuffer_(DATA(self)->socket, IoSeq_rawByteArray(DATA(self)->buffer));
}

IoSocket *IoSocket_proto(void *state)
{
    IoObject *self = IoObject_new(state);
    
    self->data = calloc(1, sizeof(IoSocketData));
    self->tag = IoSocket_tag(state);
    
    DATA(self)->socket = Socket_new();
    DATA(self)->buffer = IoSeq_newWithCString_(state, "");
    IoSocket_init(self);
    
    IoState_registerProtoWithFunc_(state, self, IoSocket_proto);
    
    {
	IoMethodTable methodTable[] = {
	{"setHost", IoSocket_host_},
	{"host", IoSocket_host},
	{"setPort", IoSocket_port_},
	{"port", IoSocket_port},
	{"openFifo", IoSocket_openFifo_},
	    
	/* udp */
	{"udpOpen", IoSocket_udpOpen},
	{"udpFromIP", IoSocket_udpFromIP},
	
	/* server */
	{"serverOpen", IoSocket_serverOpen},
	{"serverWaitForConnection", IoSocket_serverWaitForConnection},
	    
	{"connect", IoSocket_connect},
	{"isOpen", IoSocket_isOpen},
	{"read", IoSocket_read},
	{"write", IoSocket_write},
	{"close", IoSocket_close},
	{"readBuffer", IoSocket_buffer},
	{"error", IoSocket_error},
	{"clearError", IoSocket_clearError},
	{"bytesPerSecondRead", IoSocket_transferRate},
	
	{"readTimeout", IoSocket_readTimeout},
	{"setReadTimeout", IoSocket_readTimeout_},
	
	{"connectTimeout", IoSocket_connectTimeout},
	{"setConnectTimeout", IoSocket_setConnectTimeout},
	
	{"bytesPerRead", IoSocket_readSize},
	{"setBytesPerRead", IoSocket_readSize_},
	{NULL, NULL},
	};
	IoObject_addMethodTable_(self, methodTable);
    }
    
    return self;
}

IoSocket *IoSocket_rawClone(IoSocket *proto) 
{ 
    IoObject *self = IoObject_rawClonePrimitive(proto);
    self->data = cpalloc(proto->data, sizeof(IoSocketData));
    
    DATA(self)->socket = Socket_new();
    Socket_host_(DATA(self)->socket, Socket_host(DATA(proto)->socket));
    Socket_port_(DATA(self)->socket, Socket_port(DATA(proto)->socket));
    DATA(self)->buffer = IoSeq_newWithCString_(IOSTATE, "");
    IoSocket_init(self);
    return self; 
}

IoSocket *IoSocket_new(void *state)
{
    IoObject *proto = IoState_protoWithInitFunction_(state, IoSocket_proto);
    return IOCLONE(proto);
}

/* ----------------------------------------------------------- */

IoSocket *IoSocket_newWithDescriptor_address_(void *state, int descriptor, struct sockaddr *address)
{
    IoSocket *self = IoSocket_new(state);
    DATA(self)->socket->descriptor->fd = descriptor;
    Socket_fromAddress_(DATA(self)->socket, address);
    return self;
}

IoSocket *IoSocket_newWithSocket_(void *state, Socket *socket)
{
    IoSocket *self = IoSocket_new(state);
    /* make this more efficient later */
    Socket_free(DATA(self)->socket);
    DATA(self)->socket = socket;
    IoSocket_init(self);
    return self;
}

void IoSocket_free(IoSocket *self) 
{
    Socket_free(DATA(self)->socket);
    free(self->data);
}

void IoSocket_mark(IoSocket *self) 
{ 
    /* make sure it doesn't get collected since our socket references it */
    IoObject_makeGrayIfWhite(DATA(self)->ioSocketManager); 
    IoObject_makeGrayIfWhite(DATA(self)->buffer); 
}

/* ----------------------------------------------------------- */

IoObject *IoSocket_host(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("host", 
    "Returns the host ip address. ")
    */
    
    return IOSYMBOL(Socket_host(DATA(self)->socket)); 
}

IoObject *IoSocket_host_(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("setHost(ipAddress)", 
    "Sets the host ip address. Returns self.")
    */
    
    IoSymbol *host = IoMessage_locals_symbolArgAt_(m, locals, 0);
    Socket_host_(DATA(self)->socket, CSTRING(host));
    return self; 
}

IoObject *IoSocket_port(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("port", 
    "Returns the port number.")
    */
    
    return IONUMBER(Socket_port(DATA(self)->socket)); 
}

IoObject *IoSocket_port_(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("setPort(aNumber)", 
    "Sets the port number. Returns self.")
    */
    
    IoNumber *host = IoMessage_locals_numberArgAt_(m, locals, 0);
    Socket_port_(DATA(self)->socket, IoNumber_asInt(host));
    return self; 
}

/*---------------*/

IoObject *IoSocket_readSize(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("readSize", 
    "Returns the max number of bytes to read per read call. ")
    */
    
    return IONUMBER(Socket_readSize(DATA(self)->socket)); 
}

IoObject *IoSocket_readSize_(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("setReadSize(aNumber)", 
    "Sets the max number of bytes to read per read call. The default is 8192. Returns self.")
    */
    
    IoNumber *host = IoMessage_locals_numberArgAt_(m, locals, 0);
    Socket_readSize_(DATA(self)->socket, IoNumber_asInt(host));
    return self; 
}

IoObject *IoSocket_readTimeout(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("readRimeout", 
    "Returns the number of seconds before socket will time out on reads. ")
    */
    
    return IONUMBER(Socket_readTimeout(DATA(self)->socket)); 
}

IoObject *IoSocket_readTimeout_(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("setReadTimeout(aNumber)", 
    "Sets the number of seconds before socket will time out on reads. Returns self.")
    */

    IoNumber *host = IoMessage_locals_numberArgAt_(m, locals, 0);
    Socket_readTimeout_(DATA(self)->socket, IoNumber_asInt(host));
    return self; 
}

IoObject *IoSocket_connectTimeout(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("connectTimeout", 
    "Returns the number of seconds before socket will time out on connect. ")
    */
    
    return IONUMBER(Socket_connectTimeout(DATA(self)->socket)); 
}

IoObject *IoSocket_setConnectTimeout(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("setConnectTimeout(aNumber)", 
    "Sets the number of seconds before socket will time out on connect. Returns self.")
    */
    
    IoNumber *host = IoMessage_locals_numberArgAt_(m, locals, 0);
    Socket_connectTimeout_(DATA(self)->socket, IoNumber_asInt(host));
    return self; 
}

/*-----------------------*/

IoObject *IoSocket_openFifo_(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("openFifo(path)", 
    "Opens a FIFO file on the specified file path. Returns self.")
    */
    
    IoSymbol *path = IoMessage_locals_seqArgAt_(m, locals, 0);
    Socket_openFifo_(DATA(self)->socket, CSTRING(path));
    return (strcmp(Descriptor_error(DATA(self)->socket->descriptor), "")!=0) ? IONIL(self) : self; 
}

/*--- udp --------------------*/

IoObject *IoSocket_udpOpen(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("udpOpen", 
    "Opens the socket for UDP communication. Returns self.")
    */
    
    Socket_isDatagram_(DATA(self)->socket, 1);
    Socket_open(DATA(self)->socket);
    return self;
}

/*
IoObject *IoSocket_udpWrite(IoSocket *self, IoObject *locals, IoMessage *m)
{
    IoSeq *data = IoMessage_locals_seqArgAt_(m, locals, 0);
    return IONUMBER(Socket_udpWrite(DATA(self)->socket, IoSeq_rawBytes(data), IoSeq_rawSize(data)));
}

IoObject *IoSocket_udpRead(IoSocket *self, IoObject *locals, IoMessage *m)
{
    Socket_udpRead(DATA(self)->socket); 
    return self;
}
*/

IoObject *IoSocket_udpFromIP(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("udpFromIP", 
    "Returns IP address that the last UDP packet was from. ")
    */
    
    char *ip = Socket_udpFromAddressIp(DATA(self)->socket);
    return ip ? IOSYMBOL(ip) : IONIL(self); 
}

/*--- tcp --------------------*/

IoObject *IoSocket_connect(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("connect", 
    "Connect to the remote host. Returns self on success or Nil on error.")
    */
    
    Socket_connect(DATA(self)->socket);
    return (strcmp(Descriptor_error(DATA(self)->socket->descriptor), "") != 0) ? IONIL(self) : self; 
}

IoObject *IoSocket_isOpen(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("isOpen", 
    "Returns self if open or Nil otherwise.")
    */
    
    return Socket_isOpen(DATA(self)->socket) ? self : IONIL(self); 
}

IoObject *IoSocket_read(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("read", 
    "Waits up to the readTimeout period (auto yielding to other actors) to be able to read from the socket. If it is able to read, it will read up to bytesPerRead bytes and append them to the socket's read buffer. Returns self. ")
    */
    
    Socket *s = DATA(self)->socket;
    int bytesRead = 0;
    
    IOASSERT(Socket_isOpen(DATA(self)->socket), "Socket not open");

    if (Socket_isDatagram(s))
    {
	bytesRead = Socket_udpRead(s); 
    }
    else
    {
	bytesRead = Socket_read(s);
    }
    return bytesRead < 1 ? IONIL(self) : self;
}

IoObject *IoSocket_write(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("write(aSequence)", 
    "Asynchronously writes the data in aSequence. Returns self immediately. ")
    */
    
    Socket *s = DATA(self)->socket;
    int isDatagram = Socket_isDatagram(s);
    int n, max = IoMessage_argCount(m);
    
    IOASSERT(Socket_isOpen(DATA(self)->socket), "Socket not open");
    
    for (n = 0; n < max; n ++)
    {
	IoSeq *data  = IoMessage_locals_seqArgAt_(m, locals, n);
	uint8_t *bytes  = IoSeq_rawBytes(data);
	uint32_t length = IoSeq_rawSize(data);
	
	if (isDatagram)
	{
	    Socket_udpWrite(s, bytes, length);
	}
	else
	{
	    Socket_write(s, bytes, length);
	}
    }
    return self; 
}

IoObject *IoSocket_close(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("close", 
    "Closes the socket if it is open. Returns self.")
    */
    
    Socket_close(DATA(self)->socket);
    return self; 
}

IoObject *IoSocket_buffer(IoSocket *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("readBuffer", 
    "Returns the receiver's read buffer.")
    */
    
    return DATA(self)->buffer; 
}

IoObject *IoSocket_error(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("error", 
    "Returns a description of the last Socket error, or Nil if there is none.")
    */
    
    char *error = Descriptor_error(DATA(self)->socket->descriptor);
    /*printf("IoSocket_error = %s\n", error);*/
    return (strcmp(error, "")!=0) ? IOSYMBOL(error) : IONIL(self);
}

IoObject *IoSocket_clearError(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("clearError", 
    "Sets the Socket error back to Nil. Returns self.")
    */
    
    Descriptor_error_(DATA(self)->socket->descriptor, "");
    return self; 
}

IoObject *IoSocket_transferRate(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("transferRate", 
    "Returns the bytes per second read rate of the receiver since the socket opened.")
    */
    
    int rate = Socket_activeBytesPerSecond(DATA(self)->socket);
    return IONUMBER(rate);
}

/* --- Server ---------- */

IoObject *IoSocket_serverOpen(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("serverOpen", 
    "Open the socket to listen for incoming requests. Returns self.")
    */
    
    Socket_serverOpen(DATA(self)->socket);
    return self;
}

IoObject *IoSocket_serverWaitForConnection(IoSocket *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("serverWaitForConnection", 
    "Waits for connection or read timeout. 
    Returns new Socket instance for new connections or Nil for timeouts.")
    */
    
    Socket *socket = Socket_acceptNewSocket(DATA(self)->socket);
    return socket ? IoSocket_newWithSocket_(IOSTATE, socket) : IONIL(self);
}

