/*
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
*/

#include "Socket.h"
#include "SocketManager.h"
#include <time.h>
#include <fcntl.h>

#ifdef WIN32
#define EINPROGRESS 36
#define EDQUOT 69
#endif

#define SOCKET_DEFAULT_BUFFER_SIZE 1024*8

Socket *Socket_new(void)
{
    Socket *self = (Socket *)calloc(1, sizeof(Socket));
    self->descriptor = Descriptor_new();
    self->buffer = ByteArray_new();
    self->ownsBuffer = 1;
    self->readSize = 1024 * 8;
    self->connectTimeout = 10;
    self->readTimeout = 10;
    self->writeTimeout = 10;
    
    ByteArray_sizeTo_(self->buffer, self->readSize);
    Socket_host_(self, "127.0.0.1");
    Socket_port_(self, 80);
    Socket_isDatagram_(self, 0);
    return self;
}

void Socket_free(Socket *self)
{
    /*
    if (Socket_isOpen(self))
    {
	printf("Socket warning: freeing an open socket!\n");
    }
    */
    
    Socket_close(self);
    if (self->host)  free(self->host);
    Descriptor_free(self->descriptor);
    if (self->ownsBuffer) ByteArray_free(self->buffer);
    free(self);
}

void Socket_externalBuffer_(Socket *self, ByteArray *ba)
{ 
    if (self->ownsBuffer) ByteArray_free(self->buffer);
    self->buffer = ba; 
    self->ownsBuffer = 0;
}

char *Socket_error(Socket *self)
{
    return Descriptor_error(self->descriptor);
}

int Socket_hasError(Socket *self)
{
    return (strlen(Descriptor_error(self->descriptor)) != 0);
}

void Socket_fromAddress_(Socket *self, struct sockaddr *address)
{
    struct sockaddr_in *sin = (struct sockaddr_in *)(address);
    struct in_addr in = sin->sin_addr;
    Socket_host_(self, inet_ntoa(in)); /* temporary hack */
}

void Socket_host_(Socket *self, const char *host)
{
    self->host = realloc(self->host, strlen(host)+1);
    strcpy(self->host, host);
    self->toAddress.sin_addr.s_addr = inet_addr(self->host);
    self->toAddress.sin_family = AF_INET;
}

char *Socket_host(Socket *self) 
{ 
    return self->host; 
}

void Socket_port_(Socket *self, unsigned int port) 
{ 
    self->port = port; 
    self->toAddress.sin_port = htons(self->port);
}

unsigned int Socket_port(Socket *self) 
{ 
    return self->port; 
}

void Socket_readSize_(Socket *self, int n) 
{ 
    self->readSize = (n>0) ? n : 1024; 
}

int Socket_readSize(Socket *self) 
{ 
    return self->readSize; 
}

void Socket_connectTimeout_(Socket *self, time_t timeout) 
{ 
    self->connectTimeout = timeout; 
}

time_t Socket_connectTimeout(Socket *self) 
{ 
    return self->connectTimeout; 
}

void Socket_readTimeout_(Socket *self, time_t timeout) 
{ 
    self->readTimeout = timeout; 
}

time_t Socket_readTimeout(Socket *self) 
{ 
    return self->readTimeout; 
}

int Socket_isOpen(Socket *self) 
{ 
    return Descriptor_isOpen(self->descriptor); 
}

void Socket_isDatagram_(Socket *self, int b)
{ 
    self->socketType = b ? SOCK_DGRAM : SOCK_STREAM; 
}

int Socket_isDatagram(Socket *self)
{ 
    return (self->socketType == SOCK_DGRAM); 
}

int Socket_openFifo_(Socket *self, const char *path)
{
#ifdef WIN32
    return 1;
#else
    self->descriptor->fd = open(path, O_RDONLY | O_NONBLOCK);
    
    if (self->descriptor->fd < 1)
    { 
	Descriptor_error_(self->descriptor, "unable to get a fifo descriptor"); 
	return 1; 
    }
    
    return 0;
#endif
}

void Socket_open(Socket *self)
{
    int val = 1;
    
    Socket_close(self);  
    
    self->descriptor->fd = socket(AF_INET, self->socketType, 0); 
    
    if (!Descriptor_isOpen(self->descriptor) )
    { 
	Descriptor_error_(self->descriptor, "unable to get a socket descriptor"); 
    }
    
    //if (setsockopt(self->descriptor->fd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(int)) < 0)
    if (setsockopt(self->descriptor->fd, SOL_SOCKET, SO_REUSEADDR, (char*)&val, sizeof(int)) < 0)
    { 
	Descriptor_error_(self->descriptor, "unable to set SO_REUSEADDR"); 
	return;
    }
    
    if (PortableSocket_makeAsync(self->descriptor->fd) !=0 )
    { 
	Descriptor_error_(self->descriptor, "unable to make socket async"); 
	return; 
    }
}

int Socket_connect(Socket *self)
{
    Socket_open(self);
    ByteArray_clear(self->buffer);
    
    Descriptor_error_(self->descriptor, "");
    
    {  
	int r = connect(self->descriptor->fd, 
	    (struct sockaddr *)&(self->toAddress), sizeof(self->toAddress));
	/* we connected immediately */
	
	if (r == 0) 
	{ 
	    self->connectTime = time(NULL); 
	    return 1; 
	}
	/* otherwise, we're waiting on connection */
	
	/* this will auto-pause the current coro until it's ready or timeout */
	errno = 0;
	
	do 
	{
	    Descriptor_waitOnWriteOrTimeout_(self->descriptor, self->connectTimeout);
	} while (errno == EINPROGRESS && !Descriptor_hasError(self->descriptor));
	
	r = connect(self->descriptor->fd, 
	    (struct sockaddr *)&(self->toAddress), 
	    sizeof(self->toAddress));
	
	if (Descriptor_hasError(self->descriptor)) 
	{
	    return 1;
	}
	
	if (r == 0 || myerrno == EISCONN) 
	{
	    self->connectTime = time(NULL);
	    return 1; 
	}
	//perror("Socket connect error");
	Descriptor_error_(self->descriptor, strerror(myerrno)); 
	Socket_close(self);
    }
    return 0;
}

size_t Socket_activeBytesPerSecond(Socket *self)
{
    int secs = difftime(time(NULL), self->connectTime);
    return (secs) ? ByteArray_size(self->buffer)/secs : 0;
}

size_t Socket_read(Socket *self)
{
    if (self->socketType == SOCK_DGRAM)
    {
	Descriptor_error_(self->descriptor, "can't read - use udpRead on UDP sockets"); 
	return 0;
    }
    
    Descriptor_waitOnReadOrTimeout_(self->descriptor, self->readTimeout);
    
    if (Descriptor_hasError(self->descriptor)) 
    {
	return 0;
    }
    
    return Socket_justRead(self);
}

int Socket_justRead(Socket *self)
{
    int readLength;
    int length = ByteArray_size(self->buffer);
    
    ByteArray_sizeTo_(self->buffer, ByteArray_size(self->buffer) + self->readSize);
    
    readLength = PortableSocket_read(self->descriptor->fd, 
	ByteArray_bytes(self->buffer) + length, 
	self->readSize);
    /*printf("Socket read %i bytes\n", readLength);*/
    
    if (readLength <= 0) 
    {
	Socket_close(self);
	return readLength; /* finished reading */
    }
    
    ByteArray_setSize_(self->buffer, length + readLength);
    return readLength;
}

int Socket_write(Socket *self, unsigned char *data, int dataLength)
{
    if (self->socketType == SOCK_DGRAM)
    {
	Descriptor_error_(self->descriptor, "can't write - use udpWrite on UDP sockets"); 
	return 0;
    }
    else
    {
	int writeLength = PortableSocket_write(self->descriptor->fd, data, dataLength);
	
	//Descriptor_waitOnWriteOrTimeout_(self->descriptor, self->writeTimeout);
	Descriptor_error_(self->descriptor, ""); 
	
	if (writeLength == -1) 
	{
	    char *msg = 0x0;
	    
	    switch (myerrno)
	    {
		case EBADF:	msg = "EBADF"; break; 
		case EPIPE:	msg = "EPIPE"; break; 
		case EFBIG:	msg = "EFBIG"; break; 
		case EFAULT:	msg = "EFAULT"; break; 
		case EINVAL:	msg = "EINVAL"; break; 
		case ENOSPC:	msg = "ENOSPC"; break; 
		case EDQUOT:	msg = "EDQUOT"; break; 
		case EIO:	msg = "EIO"; break; 
		case EINTR:	msg = "EINTR"; break; 
		case EAGAIN:	msg = "EAGAIN"; break; 
	    }

	    if (msg)
	    {
		Descriptor_error_(self->descriptor, msg); 
	    }
	    else
	    {
		Descriptor_error_(self->descriptor, strerror(myerrno));
	    }
	    
	    /*
	    if (myerrno == EPIPE)
	    {
		Socket_close(self);
	    }
	    */
	    //return -1;
	}
    }
    return 0;
}

void Socket_close(Socket *self) 
{ 
    Descriptor_close(self->descriptor); 
}

ByteArray *Socket_buffer(Socket *self) 
{ 
    return self->buffer; 
}

/* --- UDP ------ */

size_t Socket_udpWrite(Socket *self, unsigned char *data, int length)
{
    size_t bytesSent = sendto(self->descriptor->fd, 
	data, 
	length, 
	0, 
	(struct sockaddr *)&(self->toAddress), 
	sizeof(self->toAddress));
	
    return bytesSent;
}

size_t Socket_udpRead(Socket *self)
{
    /* fromAddress records where the message was from */
    int client_address_size = sizeof(self->fromAddress);
    int bytesRead;
    
    if (Descriptor_waitOnReadOrTimeout_(self->descriptor, self->readTimeout) == -1) 
    {
	return 0;
    }
    
    ByteArray_sizeTo_(self->buffer, ByteArray_size(self->buffer) + self->readSize);
    
    bytesRead = recvfrom(self->descriptor->fd, 
	ByteArray_bytes(self->buffer), 
	self->readSize, 
	0, 
	(struct sockaddr *) &(self->fromAddress), 
	&client_address_size);
    
    if (bytesRead <= 0)
    {
	if (!Descriptor_error(self->descriptor))
	{ 
	    Descriptor_error_(self->descriptor, "error on recvfrom"); 
	}
	return 0;
    }
    
    ByteArray_setSize_(self->buffer, bytesRead);
    /*printf("got upd message '%s' length: %i\n", ByteArray_bytes(self->buffer), bytesRead);*/
    return bytesRead;
}

char *Socket_udpFromAddressIp(Socket *self)
{
    char *ip = inet_ntoa(self->fromAddress.sin_addr);
    return ip;
}

/* --- Server ---------- */

void Socket_serverOpen(Socket *self)
{
    Socket_open(self);
    
    if (bind(self->descriptor->fd, 
	(struct sockaddr*)&(self->toAddress), 
	sizeof(self->toAddress)) < 0 ) 
    { 
	Descriptor_error_(self->descriptor, "binding failed"); 
	return; 
    }
    
    if (listen(self->descriptor->fd, SOMAXCONN))
    { 
	Descriptor_error_(self->descriptor, "listen failed"); 
	return; 
    }
}

Socket *Socket_acceptNewSocket(Socket *self)
{
    if (Descriptor_isOpen(self->descriptor))
    {
	Descriptor_waitOnReadOrTimeout_(self->descriptor, self->readTimeout);
	
	/*
	if (Descriptor_hasError(self->descriptor)) 
	{ 
	    return 0x0; 
	}
	else
	*/
	{
	    struct sockaddr address;
	    int len = sizeof(address);
	    int fd = accept(self->descriptor->fd, &address, &len);
	    
	    if (fd > 0) /* got a connection */
	    { 
		Socket *socket = Socket_new();
		socket->descriptor->fd = fd;
		Socket_fromAddress_(socket, &address);
		return socket;
	    }
	}
    }
    return 0x0;
}

