/*
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
*/

#include "SocketManager.h"
#include <signal.h>

/*
#define DEBUG_SOCKETMANAGER 1
*/

/*
 rlp.rlim_cur = 16000;
 setrlimit(RLIMIT_NOFILE, &rlp);
 printf("RLIMIT_NOFILE current:%i max:%i\n", rlp.rlim_cur, rlp.rlim_max);
 */

/* ----------------------------------------------------------- */

/*
#ifdef LIBEVENT_MULTISTATE
    #define EVENT_SET(ev, fd, flag, func, entry) event_set(self->ec, ev, fd, flag, func, entry)
    #define EVENT_LOOP(a) event_loop(self->ec, a)
    #define event_init() event_init(self->ec)
#else
*/
    #define EVENT_SET  event_set
    #define EVENT_LOOP event_loop
    #define EVENT_INIT event_init
/*
#endif
*/

void SocketManager_BrokenPipeSignalHandler(int v)
{
    /*
    printf("SocketManager catching broken pipe signal\n");
    */
}

SocketManager *SocketManager_new(void)
{
    SocketManager *self = (SocketManager *)calloc(1, sizeof(SocketManager));
#ifdef LIBEVENT_MULTISTATE
    self->ec = calloc(1, sizeof(struct eventctxt));
#endif
    EVENT_INIT();
    self->timeoutMicroseconds = 0; /* one milisecond */
    self->descriptors  = Hash_new();  
    
    /* 
    Some unixes raise a signal when a socket pipe is broken, 
    so we do this to avoid a crash.
    */
    #ifndef WIN32
    signal(SIGPIPE, SocketManager_BrokenPipeSignalHandler);
    #endif
    
    //SocketManager_setDescriptorLimitToMax(self);
    setDescriptorLimitToMax();
    
    return self;
}

void SocketManager_free(SocketManager *self)
{
#ifdef LIBEVENT_MULTISTATE
    /*event_term(self->ec);*/
    free(self->ec);
#endif
    Hash_do_(self->descriptors, (HashDoCallback *)Descriptor_removeFromSocketManager);
    Hash_free(self->descriptors);
    free(self);
}

void SocketManager_scheduler_(SocketManager *self, Scheduler *s) 
{ 
    self->scheduler = s; 
}

Scheduler *SocketManager_scheduler(SocketManager *self) 
{ 
    return self->scheduler; 
}

void SocketManager_timeoutSeconds_(SocketManager *self, float v) 
{ 
     self->timeoutMicroseconds = (unsigned int)(1000000*v); 
}

float SocketManager_timeoutSeconds(SocketManager *self)
{ 
    return ((float)self->timeoutMicroseconds)/1000000.0; 
}

void SocketManager_target_(SocketManager *self, void *target)
{ 
    self->target = target; 
}

void SocketManager_callback_(SocketManager *self, SocketManagerCallback *callback)
{ 
    self->callback = callback; 
}

/* ------------------------------------ */

int SocketManager_descriptorCount(SocketManager *self)
{ 
    return Hash_count(self->descriptors); 
}

void SocketManager_handleEvent(int fd, short eventType, void *entry)
{
    Descriptor *d = entry;
    SocketManager *self = Descriptor_socketManager(d);
    
    SocketManager_removeEntry_(self, entry);
    
    if (eventType == EV_TIMEOUT) 
    { 
	Descriptor_error_(entry, "timeout"); 
    }
    
    #ifdef DEBUG_SOCKETMANAGER
    printf("SocketManager handleEvent resuming %p\n", (void *)entry);
    #endif

    Descriptor_resume(entry);  
}

void SocketManager_addReadEntry_(SocketManager *self, Descriptor *entry)
{ 
    #ifdef DEBUG_SOCKETMANAGER 
    printf("SocketManager adding socket read entry %p\n", (void *)entry);
    #endif
    
    event_set(entry->ev, entry->fd, EV_READ, SocketManager_handleEvent, entry);
    
    if (entry->hasManagerReadEntry)
    {
	printf("SocketManager warning: entry %p already has read entry\n", (void *)entry);
    }

    event_add(entry->ev, entry->timeout); 
    
    Descriptor_hasManagerReadEntry_(entry, 1);
    Hash_at_put_(self->descriptors, entry, entry);
    
    if (self->callback) 
    {
	(self->callback)(self->target);
    }
    
    Descriptor_pause(entry);
    
    #ifdef DEBUG_SOCKETMANAGER
    printf("SocketManager adding socket read entry %p resumed\n", (void *)entry);
    #endif
}


void SocketManager_addWriteEntry_(SocketManager *self, Descriptor *entry)
{ 
    #ifdef DEBUG_SOCKETMANAGER 
    printf("SocketManager adding socket write entry %p\n", (void *)entry);
    #endif

    EVENT_SET(entry->ev, entry->fd, EV_WRITE, SocketManager_handleEvent, entry);
    event_add(entry->ev, entry->timeout); 
    
    Descriptor_hasManagerWriteEntry_(entry, 1);
    Hash_at_put_(self->descriptors, entry, entry);
    
    if (self->callback) 
    {
	(self->callback)(self->target);
    }
    
    Descriptor_pause(entry); 
}

void SocketManager_removeEntry_(SocketManager *self, Descriptor *entry)
{
    #ifdef DEBUG_SOCKETMANAGER
    printf("SocketManager removing socket entry %p\n", entry);
    #endif
    
    if (entry->hasManagerReadEntry && entry->hasManagerWriteEntry)
    {
	printf("SocketManager: write and read entry for same descriptor?\n");
	exit(1);
    }
    
    Descriptor_hasManagerReadEntry_(entry, 0);
    Descriptor_hasManagerWriteEntry_(entry, 0);
    Hash_removeKey_(self->descriptors, entry);
    
    if (self->callback) 
    {
	(self->callback)(self->target);
    }
}

#include <time.h>
#ifndef WIN32
 #include <unistd.h>
#endif

inline void SocketManager_sleepInline(SocketManager *self)
{
    /*#ifdef _TIMESPEC_DECLARED*/
    if (self->timeoutMicroseconds)
    {
	/*
	 struct timespec rqtp;  
	 struct timespec extra;
	 rqtp.tv_sec = 0;
	 rqtp.tv_nsec = self->nanoSleepPeriod; 
	 nanosleep(&rqtp, &extra);
	 */
	/* if unmasked signal is raised, nanosleep returns -1 */
	/*
	 while (nanosleep(&rqtp, &extra) == -1)
	 {
	     rqtp.tv_sec  = extra.tv_sec;
	     rqtp.tv_nsec = extra.tv_nsec;     
	 }
	 */
	
	usleep(self->timeoutMicroseconds);
	printf("SocketManager usleep(%i)\n", self->timeoutMicroseconds);
    }
    /*#endif*/
}

void SocketManager_timer(SocketManager *self)
{
    int hadEvents;
    /*
#ifdef DEBUG_SOCKETMANAGER
     printf("event_loop begin\n");
#endif
     */
     
    hadEvents = event_loop(EVLOOP_NONBLOCK); 

    /*
    if (!hadEvents) 
    {
	SocketManager_sleepInline(self);
    }
    */
    
    /*
#ifdef DEBUG_SOCKETMANAGER
     printf("event_loop -----\n");
#endif
     */
}

#ifdef WIN32
long setDescriptorLimitToMax(void) { return -1; }
#else

#include <sys/resource.h>

long setDescriptorLimitToMax(void)
{
    struct rlimit rlp;
 
    /* if we're root, we could also do: */
    /*
    system("sysctl -w kern.maxfiles=1000001");
    system("sysctl -w kern.maxfilesperproc=1000000");
    */
    
    if (getrlimit(RLIMIT_NOFILE, &rlp) != 0)
    {
	return -1;
    }
    
    //printf("rlp.rlim_cur = %i\n", (int)rlp.rlim_cur);
    //printf("rlp.rlim_max = %i\n", (int)rlp.rlim_max);
    
    rlp.rlim_cur = rlp.rlim_max;
    
    if (setrlimit(RLIMIT_NOFILE, &rlp) != 0)
    {
	return -2;
    }
    
     if (getrlimit(RLIMIT_NOFILE, &rlp) != 0)
    {
	return -3;
    }
    
    //printf("SocketManager max descriptors = %i\n", (int)rlp.rlim_cur);
      
    return (long)(rlp.rlim_cur);
}
#endif

/*
void SocketManager_setDescriptorLimitToMax(SocketManager *self)
{
    long max = SocketManager_descriptorLimitMax(self);
    printf("SocketManager: descriptor hard limit %i\n", (int)max);
    max = SocketManager_setDescriptorLimit_(self, 10000);
    printf("SocketManager: descriptor limit set to %i\n", (int)max);
}

long SocketManager_descriptorLimit(SocketManager *self)
{
    struct rlimit rlp;
    
    if (getrlimit(RLIMIT_NOFILE, &rlp) != 0)
    {
	return -1;
    }
    return rlp.rlim_cur;
}

long SocketManager_descriptorLimitMax(SocketManager *self)
{
    struct rlimit rlp;
    
    if (getrlimit(RLIMIT_NOFILE, &rlp) != 0)
    {
	return -1;
    }
    
    return rlp.rlim_max;
}

long SocketManager_setDescriptorLimit_(SocketManager *self, long n)
{
    struct rlimit rlp;
    
    rlp.rlim_cur == n;
    
    if (setrlimit(RLIMIT_NOFILE, &rlp) != 0)
    {
	return -1;
    }
    
    return SocketManager_descriptorLimit(self);
}
*/
