/*   
copyright: Steve Dekorte, 2004. All rights reserved.
license: See _BSDLicense.txt.

object: DNSResolver
prototype: Object

module: IoServer
description: With the DNSResolver you can set the DNS servers to be used for DNS lookups. Example;
<pre>
  DNSResolver addDNSServerIP("128.105.2.10")
  ipForYahoo := DNSResolver ipForHostName("yahoo.com")
</pre>
*/

DNSResolver := Object clone do(
    type := "DNSResolver"
    init := method(
	self dnsServers := List clone
	self queries := Map clone
	self cache := Map clone
    )
    init
    /*
    method: dnsServers
    description: Returns list of DNSServer objects used for lookups. 
    */
    
    /*
    method: queries
    description: Returns list of active DNSQuery objects. 
    */

    /*
    method: cache
    description: Returns Map containing lookup cache. 
    */
    
    /*
    method: emptyCache
    description: Empties the lookup cache. Returns self.
    */
    emptyCache := method(cache empty; self)

    /*
    method: addDNSServerIP(ipString)
    description: Adds a new DNSServer for the specified IP. 
    */
    addDNSServerIP := method(ip, 
	dnsServer := DNSServer clone setHost(ip)
	dnsServers append(dnsServer)    
    )
    
    /*
    method: localNameServersIPs
    description: Returns List of Strings containing local name server IPs as collected from the host platform.
    */
    
    localNameServersIPs := method(DNS localNameServersIPs)

    setupServerListIfNeeded := method(
	if (dnsServers size > 0, return)
    	localNameServersIPs foreach(ip, addDNSServerIP(ip))
    )
    /*
    method: ipForHostName(hostName)
    description: Returns a String containing the IP for the hostName or 'failed', if an error occurs.
    */
    ipForHostName := method(hostName,
	setupServerListIfNeeded
	ip := cache at(hostName)
	if (ip, return ip)
	
	q := queries at(hostName)
	if (q) then (
	    //writeln("sharing ", hostName, " lookup")
	    q setHostName(hostName) addActor(currentActor)
	    currentActor pause
	) else (
	    q := DNSQuery clone setHostName(hostName)
	    queries atPut(hostName, q)
	    ip := Nil
	    dnsServers foreach(dnsServer,
		try(
		    ip = dnsServer ipForHostName(hostName)
		) 
		if (ip, break)
	    )
	    if (ip == Nil, ip = "failed")
	    cache atPut(hostName, ip)
	    queries removeAt(hostName)
	    q resumeActors
	)
	cache at(hostName)
    )    
)
