/*   
copyright: Steve Dekorte, 2004. All rights reserved.
license: See _BSDLicense.txt.

object: UDPReceiver
prototype: Object

module: IoServer
description: A UDPReceiver listens on a specified port for UDP packets and send a handleUDPPacketFromIP(buffer, ip) message to itself to handle any received packets.

To use, clone it, set the port number and implement a handleUDPPacketFromIP() method. Here's an example of an UDP time server;
<pre>

  TimeServer := UDPReceiver clone
  TimeServer setPort(8080)
  TimeRequestHandler handleUDPPacketFromIP := method(packetBuffer, ip, 
      UDPSender setHost(ip) setPort(8080) write(Date now asString)
    )
  TimeServer start

</pre>
*/


UDPReceiver := Object clone do(
    type := "UDPReceiver"

    init := method(
	self socket := Socket clone setReadTimeout(356 * 24 * 60 * 60)
    )
    init
    
    /*
    method: setPort(portNumber)
    description: Sets the receiver's port number. Returns self.
    */
    setPort := method(port, socket setPort(port); self)
    
    /*
    method: port
    description: Returns the receiver's port number.
    */
    port := method(socket port)

    /*
    method: setTimeout(seconds)
    description: Sets the receiver's read timeout. Return self.
    */
    setTimeout := method(seconds, socket setReadTimeout(seconds); self)

    /*
    method: timeout
    description: Return the receiver's read timeout. 
    */    
    timeout := method(socket readTimeout)
    
    /*
    method: start
    description: Start's to listen for UDP packets. This method does not return until the receiver is stopped, so you may want to call start with an async message. 
    */    
    start := method(
	socket udpOpen
	while(socket isOpen,
	    socket read
	    socket handleUDPPacketFromIP(socket readBuffer, socket udpFromIP)
	)
    )

    /*
    method: stop
    description: Stops the receiver. Return self.
    */
    stop := method(socket close; self)
    
    handleUDPPacketFromIP := method(data, ip,
	Nop
    )
)


