/*#io
Exception ioDoc(
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
docObject("Exception")
docDescription("""The Exception proto is used to raise and catch exceptions. 
<p>
Example of raising an Exception:
<pre>Exception raise("some name", "some description")
</pre>

Example of catching an Exception:
<pre>try(
  // code
) catch(Warning, e, 
  // code
) catch(Error, e, 
  // code
)
</pre>
""")
*/

#define IoException_C
#include "IoException.h"
#undef IoException_C
#include "IoState.h"
#include "IoNil.h"
#include "IoCatch.h"
#include "IoObject.h"
#include "IoObject_actor.h"
#include "Scheduler.h"
#include "IoList.h"
#include "ByteArray.h"

#define DATA(self) ((IoExceptionData *)self->data)

IoTag *IoException_tag(void *state)
{
    IoTag *tag = IoTag_newWithName_("Exception");
    tag->state = state;
    tag->cloneFunc = (TagCloneFunc *)IoException_rawClone;
    tag->markFunc  = (TagMarkFunc *)IoException_mark;
    tag->freeFunc  = (TagFreeFunc *)IoException_free;
    /*
     tag->writeToStoreOnStreamFunc = (TagWriteToStoreOnStreamFunc *)IoException_writeToStore_stream_;
     tag->readFromStoreOnStreamFunc = (TagReadFromStoreOnStreamFunc *)IoException_readFromStore_stream_;
     */
    return tag;
}

IoException *IoException_proto(void *state)
{
    IoMethodTable methodTable[] = {
    {"raise", IoException_raise},
    {"raiseResumable", IoException_raiseResumable},
    {"isResumable", IoException_isResumable},
    {"catch", IoException_catch},
    {"pass", IoException_pass},
    {"backTraceString", IoException_backTraceString},
    {"resume", IoException_resume},
    {NULL, NULL},
    };
    
    
    IoObject *self = IoObject_new(state);
    self->tag = IoException_tag(state);
    
    self->data = calloc(1, sizeof(IoExceptionData));
    
    IoState_registerProtoWithFunc_(state, self, IoException_proto);
    
    IoObject_addMethodTable_(self, methodTable);
    return self;
}

IoException *IoException_rawClone(IoException *proto) 
{ 
    IoObject *self = IoObject_rawClonePrimitive(proto);
    self->data = cpalloc(proto->data, sizeof(IoExceptionData));
    return self; 
}

IoException *IoException_new(void *state)
{
    IoObject *proto = IoState_protoWithInitFunction_(state, IoException_proto);
    return IOCLONE(proto);
}

void IoException_free(IoException *self)
{
    if (DATA(self)->retainStack) Stack_free(DATA(self)->retainStack);
    if (DATA(self)->savedStack) free(DATA(self)->savedStack);
    free(DATA(self));
}

/*
 void IoException_writeToStore_stream_(IoDuration *self, IoStore *store, BStream *stream)
 {
 }
 
 void *IoException_readFromStore_stream_(IoDuration *self, IoStore *store, BStream *stream)
 {
     return self;
 }
 */

/* ------------------------------------------------------------------- */

IoException *IoException_newWithDescription_(void *state, IoSymbol *d)
{
    IoException *self = IoException_new(state);
    IoException_description_(self, d);
    return self;
}

void IoException_name_(IoException *self, IoSymbol *d)
{ 
    IoObject_setSlot_to_(self, IOSYMBOL("name"), d); 
}

IoSymbol *IoException_name(IoException *self)
{ 
    IoSymbol *s = IoObject_getSlot_(self, IOSYMBOL("name")); 
    return ISSYMBOL(s) ? s : IOSYMBOL("");
}

void IoException_description_(IoException *self, IoSymbol *d)
{ 
    IoObject_setSlot_to_(self, IOSYMBOL("description"), d); 
}

IoSymbol *IoException_description(IoException *self)
{ 
    IoSymbol *s = IoObject_getSlot_(self, IOSYMBOL("description")); 
    return ISSYMBOL(s) ? s : IOSYMBOL("");
}

void IoException_caughtMessage_(IoException *self, IoSymbol *d)
{ 
    IoObject_setSlot_to_(self, IOSYMBOL("caughtMessage"), d); 
}

IoSymbol *IoException_caughtMessage(IoException *self)
{ 
    return IoObject_getSlot_(self, IOSYMBOL("caughtMessage")); 
}

void IoException_callStack_(IoException *self, IoList *list)
{ 
    IoObject_setSlot_to_(self, IOSYMBOL("callStack"), list); 
}

IoList *IoException_callStack(IoException *self)
{ 
    return IoObject_getSlot_(self, IOSYMBOL("callStack")); 
}

void IoException_mark(IoException *self)
{
    if (DATA(self)->retainStack) 
    {
	Stack_do_(DATA(self)->retainStack, (ListDoCallback *)IoObject_makeGrayIfWhite);
    }
}

void IoException_show(IoException *self)
{
    printf("\n");
}

/* -------------------------------------------------------------- */

void IoException_justRaise(IoException *self)
{
    IoObject * actor = IOSTATE->currentActor;
    IoCoroutine *coro = actor->coro;
    IoCatch *c = IoCoroutine_exceptionCatch(coro);
    
    IoException_callStack_(self, IoObject_callStack(actor));
    
    if (!c)
    {
	fputs("IoException error: no exception catch to catch current exception\n", stderr);
	IoState_exception_(IOSTATE, self);
	exit(1);
    }
    
    coro->exception = self;
    /* will need to setjump here if we want to resume */
    IoObject_setSlot_to_(self, IOSYMBOL("actor"), actor);
    
    /*printf("self = %p\n", (void *)self);*/
    if (DATA(self)->isResumable)
    {
	if (!DATA(self)->savedStack)
	{ 
	    DATA(self)->savedStack = calloc(1, sizeof(jmp_buf)); 
	}
	
	if (!CSTACK_SETJMP(*(DATA(self)->savedStack)))
	{
	    if (DATA(self)->retainStack) Stack_free(DATA(self)->retainStack);
	    DATA(self)->retainStack = Stack_clone(IOSTATE->currentActor->coro->retainedValues);
	    IoCatch_longJump(c);
	}
    }
    else
    {
	IoCatch_longJump(c);
    }
    /*printf("self = %p\n", (void *)self);*/
}

IoObject *IoException_raise(IoException *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("raise(name, description)", 
    "Raise a new, unresumable exception and set it's name and description slots. ")
    */
    
    IoSymbol *n = IoMessage_locals_symbolArgAt_(m, locals, 0);
    IoSymbol *d = IoMessage_locals_symbolArgAt_(m, locals, 1);
    IoException *e = IOCLONE(self);
    IoException_name_(e, n);
    IoException_description_(e, d);
    IoException_caughtMessage_(e, m);
    IoException_justRaise(e);
    return self;
}

IoObject *IoException_raiseResumable(IoException *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("raiseResumable(name, description)", 
    "Same as raise() except the exception will be resumable. ")
    */
    
    DATA(self)->isResumable = 1;
    return IoException_raise(self, locals, m);
}


IoObject *IoException_catch(IoException *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("catch(exceptionProto, localSlot, doMessage)", 
    "Sets the localSlot to the receiver and performs the doMessage 
    if the receiver is the exceptionProto or has exceptionProto as a parent. ")
    */
    
    IoObject *exceptionProto = IoMessage_locals_valueArgAt_(m, locals, 0);
    int doArgIndex = 1;
    
    if (IoObject_rawHasProto_(self, exceptionProto))
    {
	if (IoMessage_argCount(m) == 3)
	{
	    IoMessage *eMessage = IoMessage_rawArgAt_(m, 1);
	    IoSymbol *slotName = IoMessage_name(eMessage);
	    IoObject_setSlot_to_(locals, slotName, self);
	    doArgIndex ++;
	}
	
	IoMessage_locals_valueArgAt_(m, locals, doArgIndex);
	return IONOP(self);
    }
    return self;
}
/*
 IoObject *IoException_catch(IoException *self, IoObject *locals, IoMessage *m)
 {
     if (IoMessage_argCount(m) > 1)
     {
	 IoObject *exceptionProto = IoMessage_locals_valueArgAt_(m, locals, 0);
	 if (!IoObject_rawHasProto_(self, exceptionProto))
	 {
	     return self;
	 }
     }
     
     if (IoMessage_argCount(m) == 2)
     {
	 IoMessage_locals_valueArgAt_(m, locals, 1);
	 return IONOP(self);
     }
     
     if (IoMessage_argCount(m) == 3)
     {
	 IoMessage *eMessage = IoMessage_rawArgAt_(m, 1);
	 IoSymbol *slotName = IoMessage_name(eMessage);
	 IoObject_setSlot_to_(locals, slotName, self);
	 IoMessage_locals_valueArgAt_(m, locals, 2);
	 return IONOP(self);
     }
     
     return self;
 }
 */

void IoException_rawPass(IoException *self)
{
    IoException_justRaise(self);
}

IoObject *IoException_pass(IoException *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("pass", 
    "Pass the exception up to the next try().")
    */
    
    IoException_justRaise(self);
    return self;
}

IoObject *IoException_resume(IoException *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("resume", 
    "Continue executing at the point where the exception 
    was raised (if the exception is resumable). ")
    */
    
    if (DATA(self)->isResumable && DATA(self)->savedStack) 
    { 
	Stack_copy_(IOSTATE->currentActor->coro->retainedValues, DATA(self)->retainStack);
	DATA(self)->isResumable = 0;
	Stack_free(DATA(self)->retainStack);
	DATA(self)->retainStack = 0x0;
	CSTACK_LONGJMP(*(DATA(self)->savedStack), 1);
    }
    return self;
}

IoObject *IoException_isResumable(IoException *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("isResumable", 
    "Returns self if the reciever is resumable, Nil otherwise. ")
    */
    
    return DATA(self)->isResumable ? self : IONIL(self);
}

/* --- Stack Trace ----------------------------------------------- */

ByteArray *IoException_backTraceDescription(IoException *self)
{
    ByteArray *ba = ByteArray_new();
    ByteArray_fromFormat_(ba, "%s: %s\n\n", 
	CSTRING(IoException_name(self)), 
	CSTRING(IoException_description(self))); 
    
    ByteArray_fromFormat_(ba, "Label %# Line \t Char \t Message\n", 20-5);
    /*IoState_print_(s, "----- %# ---- \t ---- \t -------\n", 20-5);*/
    ByteArray_appendCString_(ba, "------------------------------------------------\n");
    
    if (self && !ISNIL(IoException_caughtMessage(self)))
    {
	ByteArray *stackEntryDescription = 
	    IoMessage_asStackEntryDescription(IoException_caughtMessage(self));
	ByteArray_append_(ba, stackEntryDescription);
	ByteArray_free(stackEntryDescription);
    }
    
    {
	ByteArray *backTraceOnCallStackDescription = 
	    IoObject_backTraceOnCallStackDescription_(self, IoException_callStack(self));
	ByteArray_append_(ba, backTraceOnCallStackDescription);
	ByteArray_free(backTraceOnCallStackDescription);
    }
    return ba;
}

void IoException_printBackTrace(IoException *self)
{
    ByteArray *ba = IoException_backTraceDescription(self);
    IoState_justPrint_(IOSTATE, "\n");
    if (getenv("QUERY_STRING")) IoState_justPrint_(IOSTATE, "<pre>\n");
    IoState_justPrint_(IOSTATE, "\n");
    IoState_justPrint_(IOSTATE, ByteArray_asCString(ba));
    IoState_justPrint_(IOSTATE, "\n");
    ByteArray_free(ba);
}

IoSymbol *IoException_backTraceString(IoException *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("backTraceString", 
    "Returns a String containing a description of the back trace.")
    */
    
    ByteArray *ba = IoException_backTraceDescription(self);
    return IoState_symbolWithByteArray_copy_(IOSTATE, ba, 0);
}

