/*#io
Object ioDoc(
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
docObject("Object")
docInclude("_ioCode/Object.io")
docDescription("""An Object is a key/value dictionary with string keys and values of any type. The prototype Object contains a clone slot that is a CFuntion that creates new objects. When cloned, an Object will call it's init slot (with no arguments).

<p>
<b>Important note:</b> <p>The data structure used for objects is optimized for objects with relatively few slots (less than 100). Objects become very memory inefficient (exponentially so) when they have a large number of slots. Objects should not be used as large hash tables or lists. Use a Hash or List primitive instead.""")
*/

#include "IoState.h"
#define IOOBJECT_C
#include "IoObject.h"
#undef IOOBJECT_C 
#include "IoObject_actor.h"
#include "IoTag.h"
#include "IoCFunction.h"
#include "IoSeq.h"
#include "IoNumber.h"
#include "IoMessage.h"
#include "IoMessage_parser.h"
#include "IoCFunction.h"
#include "IoBlock.h"
#include "IoNil.h"
#include "IoList.h"
#include "IoObject.h"
#include "IoFile.h"
#include "IoSeq.h"
#include "IoFuture.h" /* for print */
#include <string.h>
#include <stddef.h>

IoObject *IoObject_activateFunc(IoObject *self, 
    IoObject *target, 
    IoObject *locals, 
    IoMessage *m, 
    IoObject *slotContext);

IoTag *IoObject_tag(void *state)
{
    IoTag *tag = IoTag_newWithName_("Object");
    tag->state = state;
    tag->cloneFunc = (TagCloneFunc *)IoObject_rawClone;
    tag->activateFunc = (TagActivateFunc *)0x0; /*IoObject_activateFunc;*/
    /* no hooks, as the Object methods are the ones that call the hooks */
    return tag;
}

IoObject *IoObject_proto(void *state)
{
    IoObject *self = calloc(1, sizeof(IoObject));
    
    IoObject_setupProtos(self);
    self->slots = PHash_new();
    self->mark.ownsSlots = 1;
    /*self->coro = IoCoroutine_new();*/ /* for mainActor */
    
    self->tag = IoObject_tag(state);
    self->state = state;
    self->data = 0x0;
    IoState_registerProtoWithFunc_(state, self, IoObject_proto);

    /*IoObject_testProtosCode(self);*/

    return self;
}

IoObject *IoObject_protoFinish(void *state)
{
    IoMethodTable methodTable[] = { 
    {"clone", IoObject_clone},
    {"shallowCopy", IoObject_shallowCopy},
    /*  {"print", IoObject_protoPrint},*/
    {"write", IoObject_protoWrite},
    {"writeln", IoObject_protoWriteLn},
    {"type", IoObject_type},
	
    /* logic - need these since Object is a special primitive */
    {"and", IoObject_and},
    {"or", IoObject_or},
	
    {"compare", IoObject_protoCompare},
    {"<", IoObject_isLessThan_},
    {">", IoObject_isGreaterThan_},
    {">=", IoObject_isGreaterThanOrEqualTo_},
    {"<=", IoObject_isLessThanOrEqualTo_},
	
    /* comparison */
    {"==", IoObject_equals},
    {"!=", IoObject_notEquals},
	
    /* introspection */
    /*  {"self", IoObject_self},*/
    {"setSlot", IoObject_protoSet_to_},
    /*  {"updateSlot", IoObject_protoSet_to_},*/
    {"updateSlot", IoObject_protoUpdateSlot_to_},
    {"getSlot", IoObject_protoGet_},
    {"hasSlot", IoObject_protoHasSlot},
    {"hasProto", IoObject_protoHasProto_},
    {"removeSlot", IoObject_protoRemoveSlot},
    {"slotNames", IoObject_protoSlotNames},
	
    /* method invocation */
    {"perform", IoObject_protoPerform},
    {"performWithArgList", IoObject_protoPerformWithArgList},
    {"ancestorWithSlot", IoObject_ancestorWithSlot},
    {"contextWithSlot", IoObject_contextWithSlot},
	
    /* control */
    {"block", IoObject_block},
    {"method", IoBlock_method},
    {"for", IoObject_for},
    {"if", IoObject_if},
    {"then", IoObject_then},
    {"", IoObject_evalArg},
    {"else", IoObject_nop},
    {"elseif", IoObject_nop},
	
    {"return", IoObject_return},
    {"loop", IoObject_loop},
    {"while", IoObject_while},
    {"break", IoObject_break},
    {"continue", IoObject_continue},
	
    /* utility */
    {"print", IoObject_lobbyPrint},
    {"do", IoObject_do},
    {"message", IoObject_message},
    {"doMessage", IoObject_doMessage},
    {"doString", IoObject_doString},
    {"doFile", IoObject_doFile},
	
    /* actor */
    {"@", IoObject_futurePerform},
    {"@@", IoObject_asyncPerform},
    {"yield", IoObject_actorYield},
    {"resume", IoObject_actorResume},
    {"pause", IoObject_actorPause},
    {"wait", IoObject_actorWait},
    {"isActive", IoObject_isActive},
    {"try", IoObject_try},
    
    /* --- Reflection --- */
    {"uniqueId", IoObject_uniqueId},
    
    /*
    {"memorySize", IoObject_memorySizeMethod},
    {"compact", IoObject_compactMethod},
    */
	
    {"init", IoObject_self},
	
    /* --- Enumeration --- */
    {"foreach", IoObject_foreach},
    {"-", IoObject_subtract},
    {"isNil", IoObject_isNil},
    {"ifNil", IoObject_ifNil},
    {"ifTrue", IoObject_ifTrue},
    {"ifFalse", IoObject_ifFalse},
    {"thisContext", IoObject_self},
    {"turnOnMessageDebugging", IoObject_turnOnMessageDebugging},
    {"turnOffMessageDebugging", IoObject_turnOffMessageDebugging},
    {"actorCallStack", IoObject_actorCallStack},
    
    /* --- protos --- */
    
    {"setProto", IoObject_setProto},
    {"setProtos", IoObject_setProtos},
    {"appendProto", IoObject_appendProto},
    {"prependProto", IoObject_prependProto},
    {"removeProto", IoObject_removeProto},
    {"removeAllProtos", IoObject_removeAllProtos},
    {"protos", IoObject_protos},
    {"proto", IoObject_objectProto},
    {"tailCall", IoObject_tailCall},
    {"setIsActivatable", IoObject_setIsActivatable},

    {0x0, 0x0},
    };

    IoObject *self = IoState_protoWithInitFunction_(state, IoObject_proto);
    
    IoObject_addMethodTable_(self, methodTable);
    return self;
}

#include <assert.h>

IoObject *IoObject_localsProto(void *state)
{
    IoObject *self = IoObject_new(state);
    IoObject *proto = IoObject_firstProto(self);
    assert(proto != 0x0);
    IoObject_createSlotsIfNeeded(self);
    PHash_copy_(self->slots, proto->slots);
    /*
    IoObject_removeSlot_(self, IOSYMBOL("proto")); 
    IoObject_removeSlot_(self, IOSYMBOL("parent")); 
    */
    IoObject_rawRemoveAllProtos(self);
    
    IoObject_removeSlot_(self, IOSYMBOL("delegate")); 
    IoObject_addMethod_(self, IOSYMBOL("forward"), IoObject_localsForward);
    IoObject_addMethod_(self, IOSYMBOL("setSlot"), IoObject_protoSet_to_);
    IoObject_addMethod_(self, IOSYMBOL("updateSlot"), IoObject_localsUpdateSlot);
    return self;
}

IoObject *IoObject_addMethod_(IoObject *self, IoSymbol *slotName, IoMethodFunc *fp)
{
    IoTag *t = self->tag;
    IoObject *proto = IoState_protoWithInitFunction_(IOSTATE, IoObject_proto);
    IoCFunction *f;
    
    if (t == proto->tag)
    {
	t = 0x0;
    }
    
    f = IoCFunction_newWithFunctionPointer_tag_(IOSTATE, (IoUserFunction *)fp, t);
    IoObject_setSlot_to_(self, slotName, f);
    return f;
}

void IoObject_addMethodTable_(IoObject *self, IoMethodTable *methodTable)
{
    IoMethodTable *entry = methodTable;
    
    while (entry->name) 
    {
	IoObject_addMethod_(self, IOSYMBOL(entry->name), entry->func);
	entry ++;
    }
}

void IoObject_addDeprecatedMethodTable_(IoObject *self, IoDeprecatedMethodTable *methodTable)
{
    IoDeprecatedMethodTable *entry = methodTable;
    char *tmp = malloc(1024);
    
    while (entry->name) 
    {	
    
	sprintf(tmp, "%s := method(Exception raise(\"deprecated\", \"%s is deprecated, use %s instead\"))", 
	    entry->name, entry->name, entry->newName);
	    
	IoObject_rawDoString_label_(self, IOSYMBOL(tmp), IOSYMBOL("[deprecated]"));

	entry ++;
    }
    
    free(tmp);
}

/*
void IoObject_addMethodDocTable_(IoObject *self, IoMethodDocTable *methodTable)
{
    IoMethodDocTable *entry = methodTable;
    
    while (entry->name) 
    {
	IoCFunction *f = IoObject_addMethod_(self, IOSYMBOL(entry->name), entry->func);
	
	if (entry->doc)
	{
	    IoObject_setSlot_to_(f, IOSYMBOL("_doc"), IOSYMBOL(entry->doc));
	}
	
	entry ++;
    }
}
*/

IoObject *IoObject_new(void *state)
{
    IoObject *proto = IoState_protoWithInitFunction_(state, IoObject_proto);
    return IOCLONE(proto);
}

IoObject *IoObject_justClone(IoObject *self)
{ 
    return (self->tag->cloneFunc)(self); 
}

/*
 void IoObject_createSlotsIfNeeded(IoObject *self)
 {
     if (!self->mark.ownsSlots)
     { 
	 IoObject_createSlots(self); 
     }
 }
 */

void IoObject_createSlots(IoObject *self)
{
    /*IoState *state = IOSTATE;*/
    /*printf("%p new slots\n", (void *)self);*/
    self->slots = PHash_new();
    self->mark.ownsSlots = 1;
}

inline void IoObject_freeData(IoObject *self) 
{
    if (!self->mark.doesNotOwnData)
    {
	TagFreeFunc *func = self->tag->freeFunc;
	
	if (func) 
	{ 
	    (*func)(self); 
	}
	else if (self->data) 
	{ 
	    free(self->data); 
	}
    }
    
    self->data = 0x0;
    self->extraData = 0x0;
}

#define IOOBJECT_RECYCLE

#ifdef IOOBJECT_RECYCLE

/*static int recycleHighPoint = 0;*/

inline IoObject *IoObject_alloc(IoObject *self)
{
    IoObject *child = List_pop(IOSTATE->recycledObjects);
    
    if (!child) 
    {
	child = (IoObject *)calloc(1, sizeof(IoObject));
	IoObject_setupProtos(child);
    }
        
    child->state = self->tag->state;
    return child;
}

inline void IoObject_unalloc(IoObject *self)
{
    if (IOSTATE->recycledObjects) 
    { 
	self->mark.isSymbol = 0;
	/*printf("recycling %p\n", (void *)self);*/
	List_append_(IOSTATE->recycledObjects, self);
	/*
	 if (List_size(IOSTATE->recycledObjects) > recycleHighPoint)
	 { 
	     recycleHighPoint = List_size(IOSTATE->recycledObjects); 
	     printf("recycleHighPoint %i\n", recycleHighPoint);
	 }
	 */
    }
    else
    { 
	IoObject_dealloc(self); 
    }
}

#else

inline IoObject *IoObject_alloc(IoObject *self)
{ 
    IoObject *child = calloc(1, sizeof(IoObject));
    IoObject_setupProtos(self);
    child->state = self->tag->state;
    return child;
}

#endif

inline void IoObject_setProtoTo_(IoObject *self, IoObject *proto)
{
    IoObject_rawRemoveAllProtos(self);
    IoObject_rawSetProto_(self, proto);

    if (!self->slots)
    {
	self->slots = proto->slots;
	self->mark.ownsSlots = 0; /* should be redundant */
    }
}

IoObject *IoObject_rawClone(IoObject *self)
{
    IoObject *child = IoObject_alloc(self);
    child->tag = self->tag;
    IoObject_setProtoTo_(child, self);
    child->data = self->data;
    child->mark.isDirty = 1;
    return child;
}

IoObject *IoObject_rawClonePrimitive(IoObject *proto)
{
    IoObject *self = IoObject_alloc(proto);
    self->tag = proto->tag;
    IoObject_setProtoTo_(self, proto);
    self->data = 0x0;
    self->mark.isDirty = 1;
    self->mark.knownToCollector = 0;
    return self;
}

/* --- protos --------------------------------------------- */

void IoObject_rawPrintProtos(IoObject *self)
{
    IoObject **proto = self->protos;
    int count = 0;
    
    while (*proto)
    {
	printf("%i : %p\n", count, (void *)(*proto));
	proto ++;
	count ++;
    }
    printf("\n");
}

void IoObject_setupProtos(IoObject *self)
{
    self->protos = (IoObject **)calloc(2, sizeof(IoObject *));
}

int IoObject_hasProtos(IoObject *self)
{
    return (self->protos[0] != 0x0);
}

int IoObject_rawProtosCount(IoObject *self)
{
    IoObject **proto = self->protos;
    int count = 0;
    
    while (*proto)
    {
	proto ++;
	count ++;
    }
    
    return count;
}

void IoObject_rawAppendProto_(IoObject *self, IoObject *p)
{
    int count = IoObject_rawProtosCount(self);
    
    self->protos = realloc(self->protos, (count + 2) * sizeof(IoObject *));
    self->protos[count] = IOREF(p);
    self->protos[count + 1] = 0x0;
}

void IoObject_rawPrependProto_(IoObject *self, IoObject *p)
{
    int count = IoObject_rawProtosCount(self);
    int oldSize = (count + 1) * sizeof(IoObject *);
    int newSize = oldSize + sizeof(IoObject *);
    
    /*IoObject_rawPrintProtos(self);*/
    
    self->protos = realloc(self->protos, newSize);
    {
	void *src = self->protos;
	void *dst = self->protos + 1;
	memmove(dst, src, oldSize);
    }
    
    /*IoObject_rawPrintProtos(self);*/
    self->protos[0] = IOREF(p);
}

void IoObject_rawRemoveProto_(IoObject *self, IoObject *p)
{
    IoObject **proto = self->protos;
    int count = IoObject_rawProtosCount(self);
    int index = 0;
    
    while (*proto)
    {
	if (*proto == p)
	{
	    memcpy(proto, proto + 1, (count - index) * sizeof(IoObject *)); 
	}
	
	proto ++;
	index ++;
    }
}

/*
IoObject *IoObject_rawProtoAt_(IoObject *self, int i)
{
    return self->protos[i];
}
*/

#include <assert.h>

void IoObject_testProtosCode(IoObject *self)
{
    IoObject *o1 = (IoObject *)0x1;
    IoObject *o2 = (IoObject *)0x2;
    int c;
    
    IoObject_rawRemoveAllProtos(self);
    c = IoObject_rawProtosCount(self);
    assert(c == 0);
    
    IoObject_rawAppendProto_(self, o1);
    assert(IoObject_rawProtoAt_(self, 0) == o1);
    assert(IoObject_rawProtoAt_(self, 1) == 0x0);
    assert(IoObject_rawProtosCount(self) == 1);
    
    IoObject_rawPrependProto_(self, (IoObject *)0x2);
    assert(IoObject_rawProtosCount(self) == 2);
    assert(IoObject_rawProtoAt_(self, 0) == o2);
    assert(IoObject_rawProtoAt_(self, 1) == o1);
    assert(IoObject_rawProtoAt_(self, 2) == 0x0);
    
    IoObject_rawRemoveAllProtos(self);
    c = IoObject_rawProtosCount(self);
    assert(c == 0);    
}

void IoObject_rawSetProto_(IoObject *self, IoObject *proto)
{
    self->protos[0] = IOREF(proto);
} 

IoObject *IoObject_objectProto(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("proto", 
    "Same as; method(self protos first)")
    */
    
    IoObject *proto = self->protos[0];
        
    return proto ? proto : IONIL(self);
} 

IoObject *IoObject_setProto(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("setProto(anObject)", 
    "Sets the first proto of the receiver to anObject, replacing the 
    current one, if any. Returns self.")
    */
    
    IoObject *proto = IoMessage_locals_valueArgAt_(m, locals, 0);
    
    IoObject_rawSetProto_(self, proto);    
    return self;
} 

IoObject *IoObject_appendProto(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("appendProto(anObject)", 
    "Appends anObject to the receiver's proto list. Returns self.")
    */
    
    IoObject *proto = IoMessage_locals_valueArgAt_(m, locals, 0);
    IoObject_rawAppendProto_(self, proto);
    return self;
}

IoObject *IoObject_prependProto(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("prependProto(anObject)", 
    "Prepends anObject to the receiver's proto list. Returns self.")
    */
    
    IoObject *proto = IoMessage_locals_valueArgAt_(m, locals, 0);
    IoObject_rawPrependProto_(self, proto);
    return self;
} 

IoObject *IoObject_removeProto(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("removeProto(anObject)", 
    "Remove's anObject from the receiver's proto list if it 
    is present. Returns self.")
    */
    
    IoObject *proto = IoMessage_locals_valueArgAt_(m, locals, 0);
    IoObject_rawRemoveProto_(self, proto);
    return self;
} 

IoObject *IoObject_removeAllProtos(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("removeAllProtos", 
    "Remove's all of the receiver's protos. Returns self. ")
    */
    
    IoObject_rawRemoveAllProtos(self);
    return self;
}

IoObject *IoObject_setProtos(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("setProtos(aList)", 
    "Replaces the receiver's protos with a copy of aList. Returns self.") 
    */
    
    IoList *ioList = IoMessage_locals_listArgAt_(m, locals, 0);
    List *list = IoList_rawList(ioList);
    
    IoObject_rawRemoveAllProtos(self);
    List_target_do_(list, self, (ListDoWithCallback *)IoObject_rawAppendProto_);
    
    return self;
}


IoObject *IoObject_protos(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("protos", 
    "Returns a copy of the receiver's protos list.")
    */
    
    IoList *ioList = IoList_new(IOSTATE);
    List *list = IoList_rawList(ioList);
    IoObject **proto = self->protos;
    
    while (*proto)
    {
	List_append_(list, *proto);
	proto ++;
    }

    return ioList;
} 

/* -------------------------------------------------------- */

inline void IoObject_freeCoro(IoObject *self) /* prepare for free and possibly recycle */
{  
    if (self->coro) 
    { 
	IoCoroutine_free(self->coro);
	self->coro = 0x0;
    }
}

inline void IoObject_freeSlots(IoObject *self) /* prepare for free and possibly recycle */
{  
    if (self->mark.ownsSlots) 
    { 
	PHash_free(self->slots);
	self->slots = 0x0;
	self->mark.ownsSlots = 0;
    }
    
    self->slots = 0x0;
}

void IoObject_free(IoObject *self) /* prepare for free and possibly recycle */
{  
    self->mark.isLocals = 0;
    self->mark.knownToCollector = 0; 
    //printf("IoObject_free(%p)\n", (void *)self);

    /*
    if (self->persistentId) IoStore_willFreeObject_(store, self); 
    */
    
    IoObject_freeData(self);
    IoObject_freeCoro(self);
    
    /*self->tag = 0x0;*/ /* new */
    IoObject_rawRemoveAllProtos(self);
    self->persistentId = 0;
    
#ifdef IOOBJECT_RECYCLE

    if (self->mark.ownsSlots)
    { 
	PHash_clean(self->slots); 
    }
    else 
    { 
	self->slots = 0x0; 
    }
    /*IoObject_freeSlots(self);*/
    IoObject_unalloc(self);
    
#else

    IoObject_dealloc(self);
    
#endif 
}

void IoObject_dealloc(IoObject *self) /* really free it */
{ 
    if (self->mark.ownsSlots) 
    {
	PHash_free(self->slots);
    }
    free(self->protos);
    memset(self, 0x0, sizeof(IoObject)); /* temp */
    free(self);
}

/* ---------------------------------------------------------------- */

IoObject *IoObject_protoCompare(IoObject *self, IoObject *locals, IoMessage *m) 
{
    /*#io
    docSlot("compare(anObject)", 
    "Returns a number containing the comparison value of the target with anObject.")
    */
    IOASSERT(IoMessage_argCount(m), "compare requires argument");
    
    {
    IoSymbol *other = IoMessage_locals_valueArgAt_(m, locals, 0);
    return IONUMBER(IoObject_compare(self, other));
    }
}

/* ------------------------------------------------------
* slot lookups with lookup loop detection 
* ------------------------------------------------------*/

unsigned int IoObject_rawHasProto_(IoObject *self, IoObject *p)
{ 
    if (self == p) 
    {
	return 1;
    }
    
    if (IoObject_hasDoneLookup(self)) 
    {
	return 0;
    }
    else
    {
	IoObject **proto = self->protos;
	
	IoObject_setHasDoneLookup_(self, 1);
	
	while (*proto) 
	{
	    if (IoObject_rawHasProto_(*proto, p))
	    {
		IoObject_setHasDoneLookup_(self, 0);
		return 1;
	    }
	    
	    proto ++;
	}
	
	IoObject_setHasDoneLookup_(self, 0);
	return 0;
    }
}

IoObject *IoObject_protoHasProto_(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("hasProto(anObject)", 
    "Returns self if anObject is found in the proto 
    path of the target, Nil otherwise.")
    */
    
    IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
    return IoObject_rawHasProto_(self, v) ? self : IONIL(self);
}

/* ------------------------------------------------------ */

IoObject *IoObject_getSlot_(IoObject *self, IoSymbol *slotName)
{ 
    IoObject *v = IoObject_rawGetSlot_(self, slotName);
    return v ? v : IONIL(self);
}

IoObject *IoObject_activateFunc(IoObject *self, 
    IoObject *target, 
    IoObject *locals, 
    IoMessage *m,
    IoObject *slotContext)
{
    IoState *state = IOSTATE;
    
    if (self->mark.isActivatable)
    {
	IoObject *context;
	IoObject *slotValue = IoObject_rawGetSlot_context_(self, state->activateString, &context);
	
	if (slotValue) 
	{
	    return IoObject_activate(slotValue, self, locals, m, context);
	}	
    }
    return self;
}

/* ----------------------------------------------------------- */

void IoObject_setSlot_to_(IoObject *self, IoSymbol *slotName, IoObject *value)
{ 
    IoObject_inlineSetSlot_to_(self, slotName, value); 
}

void IoObject_removeSlot_(IoObject *self, IoSymbol *slotName)
{     
    IoObject_createSlotsIfNeeded(self);
    PHash_removeKey_(self->slots, slotName); 
}

IoObject *IoObject_rawGetSlot_target_(
    IoObject *self, 
    IoSymbol *slotName, 
    IoObject **target)
{
    IoObject *slotValue = IoObject_rawGetSlot_(self, slotName);
    
    if (!slotValue)
    {
	IoObject *selfDelegate = IoObject_rawGetSlot_(self, IOSTATE->selfString);
	
	if (selfDelegate && selfDelegate != self) 
	{
	    slotValue = IoObject_rawGetSlot_(selfDelegate, slotName);
	    
	    if (slotValue) 
	    {
		*target = selfDelegate;
	    }
	}
    }
    return slotValue;
}

/*printf("%p %s\n",self,  CSTRING(IoMessage_name(m)));*/

IoObject *IoObject_localsForward(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*
    undocumented: yes
    docSlot("localsForward", 
    "CFunction used by Locals prototype for forwarding.")
    */
    /*IoObject *selfDelegate = IoObject_rawGetSlot_(self, IOSTATE->selfString);*/
    
    IoObject *selfDelegate = PHash_at_(self->slots, IOSTATE->selfString); /* cheating a bit here */
    
    if (selfDelegate && selfDelegate != self) 
    { 
	return IoObject_perform(selfDelegate, locals, m); 
    }
    
    return IONIL(self);
}

/* -- name ------------------------------------------------------ */

int IoObject_isObject(IoObject *self)
{
    return (strcmp(IoObject_name(self), "Object") == 0);
}

IoObject *IoObject_lobbyPrint(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("print", 
    "Prints a string representation of the object. Returns Nil.")
    */

    IoState *state = IOSTATE;
    const char *name = IoObject_name(self);
    IoSymbol *key;

    IoObject_createSlotsIfNeeded(self);

    key = PHash_firstKey(self->slots);
    IoState_print_(IOSTATE, "%s_%p do(\n", name, (void *)self, name);
    IoState_print_(IOSTATE, "  appendProto(");
    
    {
	IoObject **proto = self->protos;
	
	while (*proto)
	{
	    IoState_print_(IOSTATE, "%s_%p", name, (void *)*proto, name);
	    proto ++;
	    
	    if (*proto)
	    {
		IoState_print_(IOSTATE, ", ");
	    }
	}
    
    }
    IoState_print_(IOSTATE, ")\n");
    
    while (key)
    {
	IoObject *value = PHash_at_(self->slots, key);
	IoState_print_(state, "  %s := ", CSTRING(key));
	if (ISSYMBOL(value)) IoState_print_(state, "\"");
	IoObject_defaultPrint(value); 
	if (ISSYMBOL(value)) IoState_print_(state, "\"");
	IoState_print_(state, "\n");
	key = PHash_nextKey(self->slots);
    }
    IoState_print_(IOSTATE, ")\n");
    
    IoObject_printCoroInfo(self);
    /*IoState_print_(state, ")\n");*/
    return state->ioNil;
}

void IoObject_printCoroInfo(IoObject *self)
{
    if (self->coro /*&& Queue_length(self->coro->queue)*/)
    {
	Queue *q = self->coro->queue;
	int i;
	/*printf("  // async queue length = %i\n", Queue_length(q));*/
	printf("// %s actor %p label: %s\n", 
	    IoObject_name(self), (void *)self, Coro_label(self->coro->cid));
	    
	    printf("  // actor state: "); 
	
	if (IoObject_isPaused(self))
	{ 
	    printf("paused"); 
	}
	else if (Scheduler_currentCoro(IOSTATE->scheduler) == self->coro->cid)
	{
	    printf("active and current"); 
	}
	else
	{
	    printf("active but yielding"); 
	}
	printf("\n"); 
	
	{
	    printf("  // async queue:\n");
	    
	    for (i = 0; i < Queue_length(q); i ++)
	    {
		IoFuture *f = Queue_at_(q, i);
		/*IoSymbol *name = IoMessage_name(IoFuture_rawMessage(f));*/
		printf("  //   %i: ", i); 
		IoMessage_print(IoFuture_rawMessage(f));
		printf("\n"); 
	    }
	}
	printf("  // coro retaining: %i\n", IoCoroutine_retainingCount(self->coro)); 
    }
}

size_t IoObject_memorySizeFunc(IoObject *self)
{
    size_t t = sizeof(IoObject) + PHash_memorySize(self->slots);
    
    if (self->coro) 
    {
	t += IoCoroutine_memorySize(self->coro);
    }
    
    return t;
}

void IoObject_compactFunc(IoObject *self)
{
    PHash_compact(self->slots);
    
    if (self->coro) 
    {
	IoCoroutine_compact(self->coro);
    }
}

/* -- proto methods ---------------------------------------------- */

IoObject *IoObject_protoPerform(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("perform(methodName, <arg1>, <arg2>, ...)", 
    "Performs the method corresponding to methodName with the arguments supplied.") 
    */
    
    IoObject *slotName = IoMessage_locals_valueArgAt_(m, locals, 0);
    
    if (ISMESSAGE(slotName))
    {
	IOASSERT(IoMessage_argCount(m)==1, "perform takes a single argument when using a Message as an argument");
	return IoObject_perform(self, locals, slotName);  
    }
    
    IOASSERT(ISSYMBOL(slotName), "perform requires a String or Message argument");
    
    {
	IoObject *context;
	IoObject *v = IoObject_rawGetSlot_context_(self, slotName, &context);
	IoMessage *newMessage = IoMessage_newWithName_(IOSTATE, slotName);
	
	if (v)
	{
	    int i;
	    List *args = IoMessage_rawArgList(m);
	    
	    for (i = 1; i < List_size(args); i ++)
	    { 
		IoMessage_addArg_(newMessage, IoMessage_deepCopyOf_(List_at_(args, i))); 
	    }
	    
	    return IoObject_activate(v, self, locals, newMessage, context);
	}
	//IoState_error_description_(IOSTATE, newMessage, 
	//"Object perform", "Could not find slot '%s'", CSTRING(slotName));
    }
    
    return IoObject_forward(self, locals, m);
}

IoObject *IoObject_protoPerformWithArgList(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("performWithArgList(methodName, argList)", 
    "Performs the method corresponding to methodName with the arguments in the argList. ")
    */
    
    IoSymbol *slotName = IoMessage_locals_symbolArgAt_(m, locals, 0);
    IoList *args = IoMessage_locals_listArgAt_(m, locals, 1);
    List *argList = IoList_rawList(args);
    IoObject *context;
    IoObject *v = IoObject_rawGetSlot_context_(self, slotName, &context);
    
    if (v)
    {
	IoMessage *newMessage = IoMessage_newWithName_(IOSTATE, slotName);
	int i, max = List_size(argList);
	
	for (i = 0; i < max; i ++)
	{ 
	    IoMessage_addCachedArg_(newMessage, LIST_AT_(argList, i)); 
	}
	
	return IoObject_activate(v, self, locals, newMessage, context);
    }
    
    return IoObject_forward(self, locals, m);
}

IoObject *IoObject_protoWrite(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("write(<any number of arguments>)", 
    "Sends a print message to the evaluated result of each argument. Returns Nil.")
    */
    
    int n, max = IoMessage_argCount(m);
    IoState *state = IOSTATE;
    
    for (n = 0; n < max; n ++)
    {
	IoObject *v = IoMessage_locals_valueArgAt_(m, locals, n);
	IoMessage_locals_performOn_(state->printMessage, locals, v);
    }
    
    return IONIL(self);
}

IoObject *IoObject_protoWriteLn(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("writeln(<any number of arguments>)", 
    "Same as write() but also writes a return character at the end. Returns Nil.")
    */
    
    IoObject_protoWrite(self, locals, m);
    IoState_print_(IOSTATE, "\n");
    return IONIL(self);
}

inline IoObject *IoObject_initClone_(IoObject *self, 
    IoObject *locals, IoMessage *m, IoObject *newObject)
{
    IoState *state = IOSTATE;
    IoObject *context;
    IoObject *initSlotValue = IoObject_rawGetSlot_context_(newObject, state->initString, &context);
    
    if (initSlotValue) 
    {
	IoObject_activate(initSlotValue, newObject, locals, state->initMessage, context);
    }
    
    return newObject;
}

IoObject *IOCLONE(IoObject *self)
{
    IoState *state = IOSTATE;
    IoObject *newObject;
    IoState_pushCollectorPause(state);

    {
    TagCloneFunc *func = self->tag->cloneFunc;
    newObject = (func)(self);
    //printf("%p = IOCLONE()\n", (void *)newObject);
    /*return IoObject_initClone_(self, locals, m, newObject);*/
    }

    IoState_addValue_(state, newObject);
    IoState_popCollectorPause(state);
    return newObject;
}

IoObject *IoObject_clone(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("clone ", "Returns a clone of the receiver.")
    */
     
    IoObject *newObject = IOCLONE(self);
    return IoObject_initClone_(self, locals, m, newObject);
}

IoObject *IoObject_shallowCopy(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("shallowCopy", "Returns a shallow copy of the receiver.")
    */
    
    IoObject *newObject = IoObject_new(IOSTATE); 
    IoSymbol *key = PHash_firstKey(self->slots);
    
    while (key)
    {
	IoObject *value = PHash_at_(self->slots, key);
	IoObject_setSlot_to_(newObject, key, value);
	key = PHash_nextKey(self->slots);
    }  
    
    return newObject;
}

/* -- lobby methods ---------------------------------------------- */

IoObject *IoObject_protoSet_to_(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("setSlot(slotNameString, valueObject)", 
    "Sets the slot slotNameString in the receiver to 
    hold valueObject. Returns valueObject.")
    */
    
    IoSymbol *slotName  = IoMessage_locals_firstStringArg(m, locals);
    //IoObject *slotValue = IoMessage_locals_quickValueArgAt_(m, locals, 1);
    //IoSymbol *slotName  = IoMessage_locals_symbolArgAt_(m, locals, 0);
    IoObject *slotValue = IoMessage_locals_valueArgAt_(m, locals, 1);
    IoObject_inlineSetSlot_to_(self, slotName, slotValue);
    return slotValue;
}

IoObject *IoObject_localsUpdateSlot(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*
    undocumented: yes
    docSlot("updateLocalSlot(slotNameString, valueObject)", 
    "")
    */
    
    IoSymbol *slotName  = IoMessage_locals_firstStringArg(m, locals);
    //IoSymbol *slotName  = IoMessage_locals_symbolArgAt_(m, locals, 0);
    IoObject *obj = IoObject_rawGetSlot_(self, slotName);
    
    if (obj)
    {
	//IoObject *slotValue = IoMessage_locals_valueArgAt_(m, locals, 1);
        IoObject *slotValue = IoMessage_locals_quickValueArgAt_(m, locals, 1);
	IoObject_inlineSetSlot_to_(self, slotName, slotValue);
	return slotValue;  
    }
    else
    {
	IoObject *theSelf = IoObject_rawGetSlot_(self, IOSTATE->selfString);
	
	if (theSelf) 
	{
	    return IoObject_perform(theSelf, locals, m);
	    /*
	     obj = IoObject_rawGetSlot_(theSelf, slotName);
	     if (obj)
	     {
		 
		 IoObject_inlineSetSlot_to_(theSelf, slotName, slotValue);
		 return slotValue;
	     }
	     */
	}
    }
    
    IoState_error_description_(IOSTATE, m,
			       "Io Object updateSlot", 
			       "Slot with name `%s' not found in `%s'. Use := to create slots.",
			       CSTRING(slotName), IoObject_name(self));
    
    return IONIL(self);        
}

IoObject *IoObject_protoUpdateSlot_to_(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("updateSlot(slotNameString, valueObject)", 
    "Same as setSlot(), but raises an error if the slot does not 
    already exist in the receiver's slot lookup path.")
    */
    
    IoSymbol *slotName  = IoMessage_locals_firstStringArg(m, locals);
    IoObject *slotValue = IoMessage_locals_quickValueArgAt_(m, locals, 1);
    IoObject *obj = IoObject_rawGetSlot_(self, slotName);
    
    if (obj)
    { 
	IoObject_inlineSetSlot_to_(self, slotName, slotValue); 
    }
    else
    {
	IoSymbol *description =  IoSeq_newSymbolWithFormat_(IOSTATE, 
	    "Slot %s not found. Must define slot using := operator before updating.", 
	    CSTRING(slotName));
	IoState_raiseExceptionWithName_andDescription_(IOSTATE, 
	    IOSYMBOL("Object updateSlot"), description, m);
    }

    return slotValue;
}


IoObject *IoObject_protoGet_(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("getSlot(slotNameString)", 
    "Returns the value of the slot named slotNameString 
    (following the lookup path) or Nil if no such slot is found.")
    */
    
    IoSymbol *slotName = IoMessage_locals_symbolArgAt_(m, locals, 0);
    return IoObject_getSlot_(self, slotName);
}

IoObject *IoObject_protoHasSlot(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("hasSlot(slotNameString)", 
    "Returns self if the slot exists in the receiver or Nil otherwise.")
    */
    
    IoSymbol *slotName = IoMessage_locals_symbolArgAt_(m, locals, 0);
    IoObject_createSlotsIfNeeded(self);
    return PHash_at_(self->slots, slotName) ? self : IONIL(self);
}

IoObject *IoObject_protoRemoveSlot(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("removeSlot(slotNameString)", 
    "Removes the specified slot (only) in the receiver if it exists. Returns self.")
    */
    
    IoSymbol *slotName = IoMessage_locals_symbolArgAt_(m, locals, 0);
    IoObject_createSlotsIfNeeded(self);
    PHash_removeKey_(self->slots, slotName);
    return self;
}

IoObject *IoObject_protoSlotNames(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("slotNames", 
    "Returns a list of strings containing the names of the 
    slots in the receiver (but not in it's lookup path).")
    */
    
    IoObject_createSlotsIfNeeded(self);
    
    {
	IoList *slotNames = IoList_new(IOSTATE);
	IoSymbol *slotName = PHash_firstKey(self->slots);
	
	while (slotName)
	{ 
	    IoList_rawAppend_(slotNames, slotName);
	    slotName = PHash_nextKey(self->slots);
	}
	return slotNames;
    }
}

/* --- loops --- */

IoObject *IoObject_while(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("while(<condition>, <message>)", 
    "Keeps evaluating message until condition return Nil. 
    Returns the result of the last message evaluated or Nil if none were evaluated.")
    */
    
    IoMessage_assertArgCount_receiver_(m, 2, self);
    
    {
	IoMessage *condition = IoMessage_rawArgAt_(m, 0);
	IoMessage *doMessage = IoMessage_rawArgAt_(m, 1);
	IoObject *result = IONIL(self);
	IoState *state = IOSTATE;
	unsigned char c;
	
	IoMessage_resetStopStatus(m); 
	IoState_pushRetainPool(IOSTATE);
	
	for (;;)
	{ 
	    IoState_clearTopPool(IOSTATE);
	    c = !ISNIL(IoMessage_locals_performOn_(condition, locals, self));
	    
	    if (!c) 
	    {
		break;
	    }
	    
	    result = IoMessage_locals_performOn_(doMessage, locals, self);
	    
	    if (IoMessage_handleStatus(doMessage)) 
	    {
		goto done;
	    }
	}
done:
	IoState_popRetainPoolExceptFor_(state, result);
	return result;
    }
}

IoObject *IoObject_loop(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("loop(<message>)", 
    "Keeps evaluating message until a break.")
    */
    
    IoMessage_assertArgCount_receiver_(m, 1, self);
    {
	IoMessage *doMessage = IoMessage_rawArgAt_(m, 0);
	IoState *state = IOSTATE;
	IoObject *result;
	
	IoMessage_resetStopStatus(m); 
	IoState_pushRetainPool(IOSTATE);
	
	for (;;)
	{ 
	    IoState_clearTopPool(IOSTATE);
	    
	    result = IoMessage_locals_performOn_(doMessage, locals, self);
	    
	    if (IoMessage_handleStatus(doMessage)) 
	    {
		goto done;
	    }
	}
done:
	    IoState_popRetainPoolExceptFor_(state, result);
	return result;
    }
}


IoObject *IoObject_for(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("for(<counter>, <start>, <end>, <do message>)", 
    "See manual.html. ")
    */
    
    IoMessage_assertArgCount_receiver_(m, 4, self);
    
    {
	IoState *state = IOSTATE;
	IoMessage *indexMessage = IoMessage_rawArgAt_(m, 0);
	IoMessage *doMessage;
	IoObject *result = IONIL(self);
	double i;
	IoSymbol *slotName = IoMessage_name(indexMessage);
	double startValue = IoMessage_locals_doubleArgAt_(m, locals, 1);
	double endValue   = IoMessage_locals_doubleArgAt_(m, locals, 2);
	double increment  = 1;
	IoNumber *num = 0x0;
	
	if (IoMessage_argCount(m) > 4)
	{ 
	    increment = IoMessage_locals_doubleArgAt_(m, locals, 3); 
	    doMessage = IoMessage_rawArgAt_(m, 4);
	}
	else
	{ 
	    doMessage = IoMessage_rawArgAt_(m, 3); 
	}
	
	IoMessage_resetStopStatus(m); 
	
	if (startValue > endValue) 
	{
	    increment = -1;
	}
	
	if (startValue == endValue) return IONIL(self);
	
	/*IoObject_setIsReferenced_(result, 1);*/
	IoState_pushRetainPool(state);
	
	for (i = startValue; ; i += increment)
	{
	    if (increment > 0) 
	    { 
		if (i > endValue) break; 
	    }
	    else
	    { 
		if (i < endValue) break; 
	    }
	    
	    /*if (result != locals && result != self) IoState_immediatelyFreeIfUnreferenced_(state, result);*/
	    IoState_clearTopPool(state);
	    
	    #ifdef IO_BLOCK_RECYCLING
	    /* a little trick to reuse the counter Number object if it hasn't been referenced */
	    if (num && IoObject_isReferenced(num) == 0) 
	    {
		IoNumber_rawSet(num, i);
	    }
	    else
	    #endif
	    {
		/* the IoState_clearTopPool() call above ensures we don't accumulate these new numbers on the Io stack */
		num = IONUMBER(i);
		IoObject_addingRef_(locals, num);
		PHash_at_put_(locals->slots, slotName, num);
		#ifdef IO_BLOCK_RECYCLING
		IoObject_setIsReferenced_(num, 0);
		#endif
	    } 
	    
	    /*IoObject_setSlot_to_(locals, slotName, IONUMBER(i));*/
	    result = IoMessage_locals_performOn_(doMessage, locals, self);
	    
	    if (IoMessage_handleStatus(doMessage)) 
	    {
		goto done;
	    }
	}
	
done:
	    IoState_popRetainPoolExceptFor_(state, result);
	return result;
    }
}

IoObject *IoObject_if(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("if(<condition>, <trueMessage>, <optionalFalseMessage>)", 
    "Evaluates trueMessage if condition evaluates to a non-Nil. 
    Otherwise evaluates optionalFalseMessage if it is present. 
    Returns the result of the evaluated message or Nil if none was evaluated.")
    */
    
    int argCount = IoMessage_argCount(m);
    
    IOASSERT(argCount > 0, "Error: 'if' needs at least 1 argument\n");
    
    if (argCount == 1)
    { 
	return IoMessage_locals_valueArgAt_(m, locals, 0); 
    }
    
    if (!ISNIL(IoMessage_locals_valueArgAt_(m, locals, 0)))
    { 
	return IoMessage_locals_valueArgAt_(m, locals, 1); 
    }
    else if (argCount == 3)
    { 
	return IoMessage_locals_valueArgAt_(m, locals, 2); 
    }
    
    return IONIL(self);
}

IoObject *IoObject_return(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("return anObject", 
    "Return anObject from the current execution block.")
    */
    
    IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
    IoMessage_return_(m, v);
    return self;
}

IoObject *IoObject_break(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("break(optionalReturnValue)", 
    "Break the current loop, if any. ")
    */
    
    IoObject *v = IONIL(self);
    
    if (IoMessage_argCount(m) > 0)
    {
	v = IoMessage_locals_valueArgAt_(m, locals, 0);
    }
    
    IoMessage_break(m, v);
    return self;
}

IoObject *IoObject_continue(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("continue", 
    "Skip the rest of the current loop iteration and start on 
    the next, if any. ")
    */
    
    IoMessage_continue(m); 
    return self; 
}

IoObject *IoObject_forward_(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*
     docSlot("forward")
 docDescription("""Called when the receiver is sent a message it doesn't recognize. 
     Default implementation raises an "Object doesNotRespond" exception. 
     Subclasses can override this method to implement proxies or special error handling.""")
     
     Example:
     
    <pre>
    myProxy forward = method(
	messageName := thisMessage name
	arguments := thisMessage arguments
	myObject doMessage(thisMessage)
    )
    </pre>
     
     */
    IoState_error_description_(IOSTATE, m, 
	"Io Object doesNotRespond", "%s does not respond to message '%s'", 
	IoObject_name(self), 
	CSTRING(IoMessage_name(m)));
    return IONIL(self);
}

IoObject *IoObject_ancestorWithSlot(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("ancestorWithSlot(slotName)", 
    "Returns the first ancestor of the receiver that contains 
    a slot of the specified name or Nil if none is found.")
    */ 

    IoObject *slotName = IoMessage_locals_symbolArgAt_(m, locals, 0);
    IoObject **proto = self->protos;
    
    while (*proto)
    {
	IoObject *context = 0x0;
	IoObject *v = IoObject_rawGetSlot_context_((*proto), slotName, &context);
	
	if (v) 
	{
	    return context;
	}
	
	proto ++;
    }
    
    return IONIL(self);
}

IoObject *IoObject_contextWithSlot(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("contextWithSlot(slotName)", 
    "Returns the first context (starting with the receiver and following the lookup path) 
    that contains a slot of the specified name or Nil if none is found.")
    */ 

    IoObject *slotName = IoMessage_locals_symbolArgAt_(m, locals, 0);
    IoObject *context = 0x0;
    IoObject_rawGetSlot_context_(self, slotName, &context);
    return context ? context : IONIL(self);
}

/* --------------------------------------------------------------------------- */

IoObject *IoObject_rawDoString_label_(IoObject *self, IoSymbol *string, IoSymbol *label)
{
    IoState *state = IOSTATE;
    
    if(!ISSEQ(string))
    { 
	IoState_error_description_(IOSTATE, 0x0, "Io Object", 
	    "IoObject_rawDoString_label_ requires a string argument");
    }
    
    {
	IoSymbol *internal = IOSYMBOL("[internal]");
	IoMessage *cm = IoMessage_newWithName_label_(state, IOSYMBOL("Compiler"), internal);
	IoMessage *messageForString = IoMessage_newWithName_label_(state, IOSYMBOL("messageForString"), internal);
	IoMessage *newMessage; 
	
	IoMessage_rawSetAttachedMessage(cm, messageForString);
	IoMessage_addCachedArg_(messageForString, string);
	IoMessage_addCachedArg_(messageForString, label);
	
	//printf("cm = %p\n", (void *)cm);
	//printf("messageForString = %p\n", (void *)messageForString);
	newMessage = IoMessage_locals_performOn_(cm, self, self);    

	//IoMessage *newMessage = IoMessage_newFromText_label_(IOSTATE, CSTRING(string), CSTRING(label));
	
	if (newMessage) 
	{
	    return IoMessage_locals_performOn_(newMessage, self, self); 
	}
	
	IoState_error_description_(IOSTATE, 0x0, "Io Object doString", "no message compiled\n");
	return IONIL(self);
    }
}

IoObject *IoObject_doMessage(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("doMessage(aMessage, optionalContext)", 
    "Evaluates the message object in the context of the receiver. 
    Returns the result. optionalContext can be used to specific the locals 
    context in which the message is evaluated.")
    */
    
    IoMessage *aMessage = IoMessage_locals_messageArgAt_(m, locals, 0);
    IoObject *context = self;
    
    if (IoMessage_argCount(m) > 1)
    { 
	context = IoMessage_locals_valueArgAt_(m, locals, 1); 
    }
    
    return IoMessage_locals_performOn_(aMessage, context, self);
}

IoObject *IoObject_doString(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("doString(aString)", 
    "Evaluates the string in the context of the receiver. Returns the result. ")
    */
    
    IoSymbol *string = IoMessage_locals_seqArgAt_(m, locals, 0);
    IoSymbol *label;
    IoObject *result;
    
    if (IoMessage_argCount(m) > 1)
    {
	label = IoMessage_locals_symbolArgAt_(m, locals, 1);
    }
    else
    {
	label = IOSYMBOL("doString");
    }
    
    IoState_pushRetainPool(IOSTATE);
    result = IoObject_rawDoString_label_(self, string, label);
    IoState_popRetainPoolExceptFor_(IOSTATE, result);
    return result;
}

IoObject *IoObject_doFile(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("doFile(pathString)", 
    "Evaluates the File in the context of the receiver. Returns the result. ")
    */
    
    IoSymbol *path = IoMessage_locals_symbolArgAt_(m, locals, 0);
    IoFile *file = IoFile_newWithPath_(IOSTATE, path);
    IoSymbol *string = (IoSymbol *)IoFile_asString(file, locals, m);
    
    if (IoSeq_rawSize(string))
    {
	IoList *argsList = IoList_new(IOSTATE);
	int argn = 1;
	IoObject *arg = IoMessage_locals_valueArgAt_(m, locals, argn);
	
	while (arg && !ISNIL(arg))
	{
	    IoList_rawAppend_(argsList, arg);
	    argn++;
	    arg = IoMessage_locals_valueArgAt_(m, locals, argn);
	}
	
	if (IoList_rawSize(argsList))
	{ 
	    IoObject_setSlot_to_(self, IOSYMBOL("args"), argsList); 
	}
	
	return IoObject_rawDoString_label_(self, string, path);
    }
    
    return IONIL(self);
}

IoObject *IoObject_equals(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("== aValue", 
    "Returns the receiver is aValue is identical, Nil otherwise. ")
    */
    
    IOASSERT(IoMessage_argCount(m), "compare requires argument");
    
    {
    IoObject *other = IoMessage_locals_valueArgAt_(m, locals, 0);
    return IoObject_compare(self, other) == 0 ? self : IONIL(self);
    }
}

IoObject *IoObject_notEquals(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("!= aValue", 
    "Returns the receiver is aValue is not identical, nil otherwise. ")
    */
    
    IoObject *other = IoMessage_locals_valueArgAt_(m, locals, 0);
    return IoObject_compare(self, other) != 0 ? self : IONIL(self);
}

IoObject *IoObject_foreach(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("foreach([name,] value, message)", 
    """For each slot, set name to the slot's 
    name and value to the slot's value and execute message. Example:
    <pre>
    myObject foreach(n, v, 
	writeln("slot ", n, " = ", v type)
    )
    myObject foreach(v, 
	writeln("slot type ", v type)
    )
    </pre>
    """)
    */
    
    IoSymbol *keyName;
    IoSymbol *valueName;
    IoMessage *doMessage;
    IoObject *key = PHash_firstKey(self->slots);
    IoObject *result = IONIL(self);
    
    IoState_pushRetainPool(IOSTATE);
    IoMessage_foreachArgs(m, self, &keyName, &valueName, &doMessage);
    
    while (key)
    {
        IoState_clearTopPool(IOSTATE);
        
        {
            IoObject *value = PHash_at_(self->slots, key);
            if (!ISNIL(keyName))
            {
                IoObject_setSlot_to_(locals, keyName, key);
            }
            IoObject_setSlot_to_(locals, valueName, value);
            result = IoMessage_locals_performOn_(doMessage, locals, locals);
            
            if (IoMessage_handleStatus(doMessage)) 
            {
                goto done;
            }
            
            key = PHash_nextKey(self->slots);
        }
    }
done:
        IoState_popRetainPoolExceptFor_(IOSTATE, result);
    return result;
}

IoObject *IoObject_subtract(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("- (aNumber)", 
    "Returns the negative version of aNumber. 
    Raises an exception if argument is not a number.")
    */
    
    IoNumber *num = IoMessage_locals_numberArgAt_(m, locals, 0);
    return IONUMBER(- IoNumber_asDouble(num));
}

IoObject *IoObject_self(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("self ", 
    "Returns self.")
    */
    
    return self; 
}

/* --- message callbacks -------------------------------------- */

const char *IoObject_name(IoObject *self) 
{ 
    return IoTag_name(self->tag); 
}

int IoObject_compare(IoObject *self, IoObject *v)
{
    if (self == v) 
    {
	return 0;
    }
    
    if (self->tag->compareFunc) 
    { 
	return (self->tag->compareFunc)(self, v); 
    }
    
    return IoObject_defaultCompare(self, v);
}

int IoObject_defaultCompare(IoObject *self, IoObject *v)
{
    int d = -((ptrdiff_t)self->tag - (ptrdiff_t)v->tag);
    
    if (d == 0) 
    { 
	return ((ptrdiff_t)self) - ((ptrdiff_t)v); 
    }
    
    return d;
}

int IoObject_sortCompare(IoObject **self, IoObject **v) 
{ 
    return IoObject_compare(*self, *v); 
}

size_t IoObject_memorySize(IoObject *self)
{ 
    return (self->tag->memorySizeFunc) ? (self->tag->memorySizeFunc)(self) : 0; 
}

void IoObject_compact(IoObject *self)
{ 
    if (self->tag->compactFunc) 
    { 
	(self->tag->compactFunc)(self); 
    }
}


/* -- lobby methods ---------------------------------------------- */

IoObject *IoObject_memorySizeMethod(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*
    undocumented: yes
    docSlot("memorySize", 
    "Return the amount of memory used by the object.")
    */
    return IONUMBER(IoObject_memorySize(self)); 
}

IoObject *IoObject_compactMethod(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*
    undocumented: yes
    docSlot("compact", 
    "Compact the memory for the object if possible. Returns self.")
    */
    IoObject_compact(self); 
    return self; 
}

IoObject *IoObject_type(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("type", 
    "Returns a string containing the name of the type of Object (Number, String, etc).")
    */
    
    return IOSYMBOL((char *)IoObject_name(self)); 
}

void IoObject_defaultPrint(IoObject *self)
{
    if (ISSYMBOL(self))
    {
	IoSeq_rawPrint(self); 
    }
    else if (ISNUMBER(self))
    {
	IoNumber_print(self);
    }
    else
    {
	IoState_print_(IOSTATE, "%s_%p", IoObject_name(self), self); 
    }
}

void IoObject_print(IoObject *self) 
{
    IoMessage_locals_performOn_(IOSTATE->printMessage, self, self); /* using self as locals hack */
}

IoObject *IoObject_then(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("then (<message>)", 
    "Evaluates the argument and returns Nop.")
    */
     
    IOASSERT(IoMessage_argCount(m) > 0, "argument required for 'then'\n");
    /* eval the arg and return a non-Nil so an attached else() won't get performed */
    IoMessage_locals_valueArgAt_(m, locals, 0);
    return (((IoState *)IOSTATE)->ioNop);
}

IoObject *IoObject_evalArg(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("(<message>)", 
    """The "" method evaluates the argument and returns the result.""")
    */
    
    IOASSERT(IoMessage_argCount(m) > 0, "argument required\n");
    /* eval the arg and return a non-Nil so an attached else() won't get performed */
    return IoMessage_locals_valueArgAt_(m, locals, 0);
}

IoObject *IoObject_nop(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("else (<message>)", 
    "Returns self. (Does not evaluate argument)")
    */
    
    return self; 
}


IoObject *IoObject_and(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("and(<message>)", 
    "Evaluates argument and returns Nil if it is Nil or self otherwise.")
    */
    
    IoObject *other = IoMessage_locals_valueArgAt_(m, locals, 0);
    return ISNIL(other) ? IONIL(self) : self; 
}

IoObject *IoObject_or(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("or(<message>)", 
    "Returns self.")
    */
    
    return self; 
}

IoObject *IoObject_isLessThan_(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("< (<message>)", 
    "Evaluates argument and returns self if self is less or Nil if not.")
    */
    
    IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
    
    if (ISNIL(self) || ISNIL(v)) 
    {
	return IONIL(self);
    }
    
    return IoObject_compare(self, v) < 0 ? v : IONIL(self);
}

IoObject *IoObject_isLessThanOrEqualTo_(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("<= (<message>)", 
    "Evaluates argument and returns self if self is less 
    than or equal to it, or Nil if not.")
    */
    
    IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
    
    if (ISNIL(self) || ISNIL(v)) 
    {
	return IONIL(self);
    }
    
    return IoObject_compare(self, v) <= 0 ? v : IONIL(self);
}

IoObject *IoObject_isGreaterThan_(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("> (<message>)", 
    "Evaluates argument and returns self if self is greater than it, or Nil if not.")
    */
    
    IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
    return IoObject_compare(self, v) > 0 ? v : IONIL(self);
}

IoObject *IoObject_isGreaterThanOrEqualTo_(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot(">= (<message>)", 
    "Evaluates argument and returns self if self is greater 
    than or equal to it, or Nil if not.")
    */
    
    IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
    
    if (ISNIL(self) || ISNIL(v)) 
    {
	return IONIL(self);
    }
    
    return IoObject_compare(self, v) >= 0 ? v : IONIL(self);
}

IoObject *IoObject_uniqueId(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("uniqueId", 
    "Returns a Number containing a unique id for the receiver.")
    */
    
    return IONUMBER((double)((size_t)self)); 
}

IoObject *IoObject_do(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("do(<message>)", 
    "Evaluates the message in the context of the receiver. Returns self.")
    */
     
    if (IoMessage_argCount(m) != 0) 
    {
	IoMessage *argMessage = IoMessage_rawArgAt_(m, 0);
	IoMessage_locals_performOn_(argMessage, self, self);
    }
    
    return self;
}

IoObject *IoObject_message(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("message(<message>)", 
    "Return the message object for the argument or Nil if there is no argument.")
    */
    
    if (IoMessage_argCount(m) != 0) 
    {
	IoMessage *argMessage = IoMessage_rawArgAt_(m, 0);
	return argMessage;
    }
    return IONIL(self);
}


IoObject *IoObject_isNil(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("isNil", 
    "Returns Nil.")
    */
    
    return IONIL(self); 
}

IoObject *IoObject_ifNil(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("ifNil(<message>)", 
    "Returns self. (Does not evaluate the argument)")
    */
    
    return self; 
}

IoObject *IoObject_ifTrue(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("ifTrue(<message>)", 
    "Evaluates argument and returns result.")
    */
    
    return IoMessage_locals_valueArgAt_(m, locals, 0); 
}

IoObject *IoObject_ifFalse(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("ifFalse(<message>)", 
    "Returns Nil. (Does not evaluate the argument)")
    */
    
    return IONIL(self); 
}

IoObject *IoObject_tailCall(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("tailCall(<message>)", 
    """Does a tailCall on the currently executing Block. Example:
    
    <pre>
    Io> a := method(x, x = x + 1; if(x > 10, return x); tailCall(x))
    ==> method(x, updateSlot("x", x +(1));
    if(x >(10), return(x));
    tailCall(x))
    Io> a(1)
    ==> 11
    </pre>
    """)
    */
    
    IOSTATE->tailCallMessage = m;
    IoMessage_return_(m, IONIL(self));
    return self;
}

/* --- inline these --- */

IoObject *IoObject_waitOnFutureIfNeeded(IoObject *self)
{
    if (ISWAITINGFUTURE(self)) 
    {
	IoFuture_rawWaitOnResult(self);
    }
    
    return self;
}

int IoObject_hasCloneFunc_(IoObject *self, TagCloneFunc *func)
{
    if (ISWAITINGFUTURE(self)) 
    {
	IoFuture_rawWaitOnResult(self);
    }
    
    return (self->tag->cloneFunc == func);
}

/* -------------------------------------------- */

char *IoObject_markColorName(IoObject *self)
{
    unsigned char c = self->mark.color;
    if (c == IOMARK_WHITE) return "white";
    if (c == IOMARK_GRAY)  return "gray";
    if (c == IOMARK_BLACK) return "black";
    return "off-white";
}

void IoObject_show(IoObject *self)
{
    printf("  %p %s\n", (void *)self, IoObject_name(self));
}

/* --- Debugging --- */

int IoObject_debuggingOn(IoObject *self)
{
    return self->coro->debuggingOn;
}

IoObject *IoObject_turnOnMessageDebugging(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("turnOnMessageDebugging", 
    "Turns on message level debugging for this actor. When on, this 
    actor will send a vmWillSendMessage message to the Debugger object before 
    each message send and pause itself. See the Debugger object documentation 
    for more information. ")
    */
     
    IoState *state = IOSTATE;

    if (self == state->debugger)
    {
	IoState_error_description_(state, m, 
	    "Object turnOnMessageDebugging", "attempt to debug debugger");
    }
	    
    if (self->coro)
    {
	self->coro->debuggingOn = 1;
	IoState_updateDebuggingMode(state);
    }
    else
    {
	IoState_error_description_(state, m, 
	    "Object turnOnMessageDebugging", "this object has is not an active actor");
    }
    return self;
}

IoObject *IoObject_turnOffMessageDebugging(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("turnOffMessageDebugging", 
    "Turns off message level debugging for this actor. ")
    */

    if (self->coro)
    {
	self->coro->debuggingOn = 0;
	IoState_updateDebuggingMode(IOSTATE);
    }
    return self;
}

IoObject *IoObject_performWithDebugger(IoObject *self, IoObject *locals, IoMessage *m)
{
    IoState *state = IOSTATE;
    IoObject *currentActor = IoState_currentActor(state);
	
    if (currentActor->coro->debuggingOn)
    {
	if (state->debugger)
	{
	    /*
	    printf("IoObject_performWithDebugger(%i, %p, %s)\n", 
		(int)(void *)self, (void *)locals, CSTRING(IoMessage_name(m)));
	    */
	    IoMessage_setCachedArg_to_(state->vmWillSendMessage, 0, currentActor);
	    IoMessage_setCachedArg_to_(state->vmWillSendMessage, 1, m);
	    IoMessage_setCachedArg_to_(state->vmWillSendMessage, 2, self);
	    IoMessage_setCachedArg_to_(state->vmWillSendMessage, 3, locals);
	    IoObject_actorPerform(state->debugger, locals, state->vmWillSendMessage, 0);
	    IoObject_pause(currentActor);
	}
    }
    
    return IoObject_perform(self, locals, m);
}

IoObject *IoObject_actorCallStack(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("actorCallStack", 
    "Returns a copy of the receivers call stack if it is 
    an actor, or Nil, otherwise. ")
    */
    
    return IoObject_callStack(self);
}

IoObject *IoObject_setIsActivatable(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("setIsActivatable(aValue)", 
    "When called with a non-Nil aValue, sets the object 
    to call it's activate slot when accessed as a value. Turns this behavior 
    off if aValue is Nil. Only works on Objects which are not Activatable 
    Primitives (such as CFunction or Block). Returns self.")
    */
    
    IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
    IoObject *objectProto = IoState_protoWithInitFunction_(IOSTATE, (IoStateProtoFunc *)IoObject_proto);

    objectProto->tag->activateFunc = (TagActivateFunc *)IoObject_activateFunc;
    
    self->mark.isActivatable = ISNIL(v) ? 0 : 1;
    
    return self;
}


IoObject *IoObject_deprecated(IoObject *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("deprecated", 
    """Prints "The slot [slotName] has been deprecated".""")
    */
    
    IoState_error_description_(IOSTATE, m, 
	"Object deprecated", "The slot '%s' has been deprecated. Please see the docs for the new method name.", CSTRING(IoMessage_name(m)));
    
    return self;
}

IoObject *IoObject_rawDoMessage(IoObject *self, IoMessage *m)
{
    return IoObject_eval(self, m, self);
}

//#define IO_DEBUG_STACK

IoObject *IoObject_eval(IoObject *self, IoMessage *message, IoObject *locals)
{
    IoState *state = IOSTATE;
    IoObject *r = 0x0;
    IoMessage *m = message;
    IoObject *c = self;
    
    do
    {
	do
	{
	    r = IOMESSAGEDATA(m)->cachedResult;
	    
	    if (!r)
	    {
		//printf("%s\n", CSTRING(IoMessage_name(m)));
		r = c->tag->performFunc(c, locals, m);
    
    #ifdef IO_DEBUG_STACK
    IoState_stackRetain_(state, r); 
    #endif
    
		if (state->stopStatus != MESSAGE_STOP_STATUS_NORMAL)
		{
		    r = state->returnValue;
		    goto done;
		}
	    }
	    
	    c = r;
	    
	} while ((m = IOMESSAGEDATA(m)->attachedMessage));
	
	c = locals;
	
    } while ((m = IOMESSAGEDATA(m)->nextMessage));
    
    done:
    #ifdef IO_DEBUG_STACK
    IoState_stackRetain_(state, r); 
    #endif
    return r ? r : state->ioNil;
}


