/*#io
Object ioDoc(
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
*/

#include "IoObject_actor.h"
#include "IoState.h"
#include "IoNil.h"
#include "IoNumber.h"
#include "IoFuture.h"
#include "IoException.h"
#include "IoBlock.h"
#include "IoList.h"
#include <string.h>
#include <time.h>

void IoObject_markActor(IoObject *self)
{
    IoObject_makeGrayIfWhite(self);
    
    if (self->coro) 
    {
	IoObject_mark(self);
    }
}

void IoObject_createCoroIfAbsent(IoObject *self)
{
    if (!self->coro) 
    {
	self->coro = IoCoroutine_new();
	
	{
	    Coro *c = Scheduler_currentCoro(IOSTATE->scheduler);
	    Coro_label_(c, "main");
	    Coro_object_(c, self);
	    Coro_unyieldCallback_(c, (CoroUnyieldCallback *)IoObject_prepareToUnyield);
	    self->coro->cid = c;
	}
	
	IoState_addActor_(IOSTATE, self);
    }
}

IoObject *IoObject_actorPerform(IoObject *volatile self, 
				IoObject *locals, 
				IoMessage *m, 
				int returnFuture)
{
    /* add message to async message queue for this object */
    IoState *state = IOSTATE;
    IoMessage *sendMessage;
    IoFuture *future;
    
    /* 
	setup or resume coro - need to do this first to ensure it's in 
     the State's list of actors, in case the collector frees when
     we're creating the message object to put on the queue
     */
    
    if (!self->coro)
    {
	self->coro = IoCoroutine_new();
    }
    
    if (!self->coro->cid)
    { 
	IoObject *current = IoState_currentActor(state);
	Coro *coro = Scheduler_newCoro(state->scheduler);
	Coro_label_(coro, "Actor");
	Coro_object_(coro, self);
	Coro_unyieldCallback_(coro, (CoroUnyieldCallback *)IoObject_prepareToUnyield);
	self->coro->cid = coro;
	Scheduler_startCoro(state->scheduler, coro, (void *)self, IoObject_run);
	/*printf("self %p coro %p tid %p\n", self, self->coro, self->coro->cid);*/
	IoState_currentActor_(state, current);
	IoState_addActor_(IOSTATE, self);
    }
    else
    {
	if (IoObject_isPaused(self)) 
	{
	    IoObject_resume(self);
	}
    }
    
    
    /* we need to pre-evaluate arguments since the locals will be 
	gone by the time the message is processed */
    {
	int i, max = IoMessage_argCount(m);
	
	if (IoMessage_needsEvaluation(m))
	{ 
	    sendMessage = IoMessage_newWithName_(state, IoMessage_name(m)); 
	}
	else 
	{ 
	    sendMessage = m; 
	}
	
	for (i = 0; i < max; i++)
	{
	    IoMessage *arg = IoMessage_rawArgAt_(m, i);
	    IoObject *result = IoMessage_locals_performOn_(arg, locals, locals);
	    IoMessage_setCachedArg_to_(sendMessage, i, result);
	}
    }
    
    future = IoFuture_newWithMessage_creatorObject_(state, sendMessage, self);
    Queue_push_(self->coro->queue, IOREF(future));
    
    if (returnFuture) 
    {
	return future;
    }
    
    return IONIL(self);
}

IoObject *IoObject_asyncPerform(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    if (!IoMessage_argCount(m))
    {
	IoState_error_description_(IOSTATE, m, 
				   "Io Object asyncPerform", "argument is required for method '%s'", 
				   CSTRING(IoMessage_name(m)));  
    }
    
    return IoObject_actorPerform(self, locals, IoMessage_rawArgAt_(m, 0), 0); 
}

IoObject *IoObject_futurePerform(IoObject *self, IoObject *locals, IoMessage *m)
{
    if (!IoMessage_argCount(m))
    {
	IoState_error_description_(IOSTATE, m, 
				   "Io Object futurePerform", "argument is required for method '%s'", 
				   CSTRING(IoMessage_name(m)));  
    }
    
    return IoObject_actorPerform(self, locals, IoMessage_rawArgAt_(m, 0), 1); 
}

/* --- coroutine calls ----------------------------------------------- */

void *IoObject_processQueue(void *arg1, void *arg2, void *arg3)
{
    IoObject *self = arg1;
    IoState *state = IOSTATE;
    IoObject *lobby = IoState_lobby(state);
    IoFuture *f = 0x0;
    IoObject *result;
    
    for (;;)
    {
	while ((f = Queue_top(self->coro->queue)))
	{
	    IoMessage *m = IoFuture_rawMessage((IoFuture *)f); 
	    
	    if (state->currentActor != self) 
	    { 
		printf("wrong actor\n"); 
		exit(1); 
	    }
	    
	    IoState_pushRetainPool(state);
	    result = IoObject_perform(self, lobby, m); /* may have exception here */
	    IoFuture_rawSetValue_(f, result);
	    IoState_popRetainPool(state);
	    Queue_pop(self->coro->queue);
	    
	    IoObject_yield(self);
	}
	/*self->coro->actorSleepTime = time(NULL);*/
	IoObject_pause(self);
    }
    
    return 0x0;
}

void IoObject_run(void *context)
{
    IoObject *self = context;
    
    IoObject_yield(self); /* yield so Scheduler_createCoro can return new coro id */
    
    for (;;)
    {
	IoException *e;
	IoState_tryFunc_(IOSTATE, IoObject_processQueue, self, 0x0, 0x0, &e);
	
	if (e)
	{ 
	    IoFuture *f = Queue_pop(self->coro->queue);
	    IoFuture_rawSetValue_(f, IONIL(self));
	    IoState_popRetainPool(IOSTATE);
	    IoState_exception_(IOSTATE, e);
	}
    }
}

void IoObject_prepareToUnyield(IoObject *self)
{
    IoState_currentActor_(IOSTATE, self);
}

void IoObject_yield(IoObject *self)
{
    /*printf("yielding %p %s\n", self, Coro_label(Scheduler_currentCoro(IOSTATE->scheduler)));*/
    Scheduler_yield(IOSTATE->scheduler);
    /*printf("unyielding %p %s\n", self, Coro_label(Scheduler_currentCoro(IOSTATE->scheduler)));*/
    
    /* this may be called within another coroutine, so we can't just set the currentActor to self */
    /*
     {
	 Coro *coro = Scheduler_currentCoro(IOSTATE->scheduler);
	 IoObject *currentActor = (IoObject *)Coro_object(coro);
	 if (currentActor) 
	 {
	     IoState_currentActor_(IOSTATE, currentActor);
	 }
	 else
	 {
	     printf("IoObject_actor: no object attached to coro %p\n", (void *)coro);
	 }
     }
     */
}

void IoObject_abort(IoObject *self)
{ 
    if (self->coro)
    {
	IoState *state = IOSTATE;
	void *t = self->coro->cid;
	/*
	 IoCoroutine_free(self->coro);
	 self->coro = 0x0;
	 */
	IoState_removeActor_(state, self);
	/*printf("aborting %p\n", p);*/
	Scheduler_prepareToKillCoro_(state->scheduler, t);
	self->coro->cid = 0x0;
    }
}

void IoObject_prepareToPause(IoObject *self)
{
    if (self->coro)
    {
	IoState *state = IOSTATE;
	
	if (!Scheduler_hasCoro_(state->scheduler, self->coro->cid)) 
	{
	    return; /* doesn't work on main actor */
	}
	
	Scheduler_prepareToPauseCoro_(state->scheduler, self->coro->cid);
    }
}

void IoObject_pause(IoObject *self)
{
    if (self->coro)
    {
	/*printf("pausing %p\n", self);*/
	IoObject_prepareToPause(self);
	IoObject_yield(self);
    }
}

unsigned int IoObject_isPaused(IoObject *self)
{ 
    return (self->coro && !Scheduler_hasCoro_(IOSTATE->scheduler, (Coro *)self->coro->cid));
}

void IoObject_resume(IoObject *self)
{ 
    if (self->coro)
    {  
	IoState *state = IOSTATE;
	Coro *coro = self->coro->cid;
	/*if (Scheduler_hasCoro_(state->scheduler, coro)) return;*/
	/*printf("Object resuming %p\n", self);*/
	Scheduler_resumeCoro_(state->scheduler, coro);
	/*IoState_addActor_(state, self);*/
    }
}

/* --- proto methods ---------------------------------- */

IoObject *IoObject_actorYield(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    IoObject_yield(((IoState *)IOSTATE)->currentActor);
    return self;
}

IoObject *IoObject_actorResume(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    IoObject_resume(self);
    return self; 
}

IoObject *IoObject_actorPause(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    IoObject_pause(self); 
    return self; 
}

IoObject *IoObject_isActive(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    IoState *state = IOSTATE;
    
    if (!self->coro || !Scheduler_hasCoro_(state->scheduler, self->coro->cid)) 
    { 
	return state->ioNil; 
    }
    
    return self;
}

IoObject *IoObject_actorWait(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    double seconds = IoMessage_locals_doubleArgAt_(m, locals, 0);
    IoState *state = IOSTATE;
    IoObject *currentActor = IoState_currentActor(state);
    double waited = Scheduler_waitSeconds_(state->scheduler, seconds);
    IoState_currentActor_(state, currentActor);
    return IONUMBER(waited);
}

unsigned int IoObject_isPausing_(IoObject *self, IoObject *other)
{
    if (self->coro)
    {
	if (Queue_firstTrueFor_with_(self->coro->queue, 
				     (QueueFirstTrueForWithCallback *)IoFuture_isPausing_, other))
	{ 
	    return 1; 
	}
    }
    return 0;
}

void IoObject_showStack(IoObject *self)
{
    printf("Actor %p %s:\n", (void *)self, IoObject_name(self));
    printf("\n");
}

/* --- stack trace ------------------------------------------ */

IoObject *IoObject_try(IoObject *self, IoObject *locals, IoMessage *m)
{
    IoException *e;
    IoMessage *doMessage;
    IoMessage_assertArgCount_receiver_(m, 1, self);
    doMessage = IoMessage_rawArgAt_(m, 0);
    /*printf("%s\n", CSTRING(IoMessage_name(m)));*/
    IoState_tryFunc_(IOSTATE, (IoCatchCallback *)IoMessage_locals_performOn_, doMessage, locals, self, &e);
    return e ? e : IONOP(self);
}

/* --- stack trace ------------------------------------------ */

int IoObject_isBlockLocals(IoObject *self)
{
    IoState *state = IOSTATE;
    IoMessage *m   = IoObject_rawGetSlot_(self, state->thisMessageString); /* calling message */
    IoBlock *block = IoObject_rawGetSlot_(self, state->thisBlockString);
    return (m && block && ISMESSAGE(m) && ISBLOCK(block));
}

static void IoList_addIfBlockLocals_(IoList *self, IoObject *v)
{
    if (v && IoObject_isBlockLocals(v)) 
    {
	IoList_rawAppend_(self, v); 
    }
}

IoList *IoObject_callStack(IoObject *self)
{
    if (self->coro)
    {
	IoList *calls = IoList_new(IOSTATE);
	Stack_do_on_(self->coro->retainedValues, 
		     (StackDoOnCallback *)IoList_addIfBlockLocals_, calls);
	List_reverse(IoList_rawList(calls));
	return calls;
    }
    
    return IONIL(self);
}

/*
 IoList *IoObject_callStack(IoObject *self)
 {
     IoList *calls = IoList_new(IOSTATE);
     Stack *stack = Stack_newCopyWithNullMarks(self->coro->retainedValues);
     IoObject *item = stack->items;
     
     for (i = 0; i < Stack_count(stack); i ++)
     {
	 IoObject *item = stack->items[i];
	 if (item && IoObject_isBlockLocals(item)) 
	 {
	     IoList_rawAppend_(self, item);
	     while (i++ < Stack_count(stack) && stack->items[i] != 0x0) 
	     { }
	 }
     }
     Stack_free(stack);
     return calls;
 }
 */

void IoObject_appendStackEntryDescription(IoObject *self, ByteArray *ba)
{ 
    IoMessage *m = IoObject_rawGetSlot_(self, IOSTATE->thisMessageString);
    
    if (!m)
    {
	ByteArray_fromFormat_(ba, "0x0     %# ? \t ? \t ?\n", 20 - 10);
	return;
    }
    
    {
	ByteArray *stackEntryDescription = IoMessage_asStackEntryDescription(m);
	ByteArray_append_(ba, stackEntryDescription);
	ByteArray_free(stackEntryDescription);
    }
}

ByteArray *IoObject_backTraceOnCallStackDescription_(IoObject *self, IoList *calls)
{
    ByteArray *ba = ByteArray_new();
    List *list = IoList_rawList(calls);
    List_do_with_(list, (ListDoWithCallback *)IoObject_appendStackEntryDescription, ba);
    return ba;
}
