/*
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")
*/

#define IOSTATE_C 1
#include "IoState.h"
#undef IOSTATE_C 

#include "IoObject.h"
#include "IoObject_actor.h"
#include "IoNil.h"
#include "IoCatch.h"
#include "IoException.h"
#include "IoSeq.h"
#include "IoNumber.h"
#include "IoCFunction.h" 
#include "IoBlock.h"
#include "IoList.h"
#include "IoMap.h"
#include "IoFile.h"
#include "IoDate.h"
#include "IoDuration.h"
#include "IoSeq.h"
#include "IoMessage_parser.h"
#include "IoDynLib.h"
#include "CLI.h"
#include "IoWeakLink.h"
#include "IoFuture.h"

#include "IoSystem.h"
#include "IoCompiler.h"
#include "IoDebugger.h"
#include "IoCollector.h"
#include "IoScheduler.h"
#include "IoRandom.h"
#include "IoSandbox.h"

/*#include "IoPObject.h"*/
#include "IoDirectory.h"
#include "IoFile_stat.h"
    
#include <time.h>
#include <limits.h> 
#include <stdlib.h>

void IoVMCodeInit(IoState *self, IoObject *context);

int ioStateLiteralCmp(const char *key, unsigned int h, size_t size, IoSymbol *v);
unsigned int ioStateLiteralKey(IoSymbol *v);

IoState *IoState_new(void)
{
    IoCFunction *cFunctionProto;
    IoSeq *seqProto;
    IoState *self = (IoState *)calloc(1, sizeof(IoState));
        
    self->mainArgs = MainArgs_new();
    self->primitiveProtos = Hash_new();
    self->valueToWeakLink = Hash_new();
    /*self->collectedWeakLinks = Stack_new();*/
    
    self->freeGroup  = IoObjectGroup_newWithColor_(self, IOMARK_INITIAL_FREE);
    self->blackGroup = IoObjectGroup_newWithColor_(self, IOMARK_INITIAL_BLACK);
    self->grayGroup  = IoObjectGroup_newWithColor_(self, IOMARK_INITIAL_GRAY);
    self->whiteGroup = IoObjectGroup_newWithColor_(self, IOMARK_INITIAL_WHITE);
    
    IoState_updateCachedCollectorColors(self);
    
    self->recycledObjects = List_new();
    self->recycledObjectsSansSlots = List_new();
    self->newValueCount = 0;
    IoState_pushCollectorPause(self);

    /* Sandbox */
    self->messageCount = 0;
    self->messageCountLimit = 0;
    self->endTime = 0;
    
    self->sdbm = SkipDBM_new();
    self->symbols = SkipDBM_rootSkipDB(self->sdbm);
    /*SkipDB_setCompareFunction_(self->symbols, (SkipDBCompareFunc *)ioStateStringCmp); */
    /*SkipDB_objectMarkFunc_(self->symbols, (SkipDBObjectMarkFunc *)IoObject_makeGrayIfWhite);*/
    
    self->retainedValues = List_new();
    
    self->scheduler = Scheduler_new(self);
    Scheduler_callbackContext_(self->scheduler, self);
    Scheduler_updateCallback_(self->scheduler, 
			      (SchedulerUpdateCallback *)IoState_schedulerUpdate);
    
    self->actors = List_new();
    
    /* 
     * Problem:
     * - there are some interdependencies here:
     *   - creating instances requires a retain stack
     *   - retain stack is in Object->coro->exceptionStack, 
     *     so we need a proto Object(mainActor) to use for
     *     our retainStack
     *   - defining any primitive methods requires Strings and CFunctions
     *
     * Solution:
     * - create Object, CFunction and String protos sans methods. 
     * - then add methods to Object, CFunction and String
     */
    
    self->mainActor = IoObject_proto(self); /* need to do this first, so we have a retain stack */
    
    IoObject_createCoroIfAbsent(self->mainActor);
    self->mainActor->coro->cid = Scheduler_mainCoro(self->scheduler);
    
    /*IoState_addActor_(self, self->mainActor); IoObject_createCoroIfAbsent will add it*/
    IoState_currentActor_(self, self->mainActor);
    IoState_retain_(self, self->mainActor);
    
    seqProto = IoSeq_proto(self);
    
    /* --- per state globals, just here for quick access --- */
    
    self->activateString    = IoState_retain_(self, SIOSYMBOL("activate"));
    self->thisBlockString   = IoState_retain_(self, SIOSYMBOL("thisBlock"));
    self->cloneString	    = IoState_retain_(self, SIOSYMBOL("clone"));
    self->CompilerString    = IoState_retain_(self, SIOSYMBOL("Compiler"));
    self->debugString	    = IoState_retain_(self, SIOSYMBOL("debug"));  
    self->debugParserString = IoState_retain_(self, SIOSYMBOL("debugParser"));  
    self->delegateString    = IoState_retain_(self, SIOSYMBOL("delegate"));
    self->equalEqualString  = IoState_retain_(self, SIOSYMBOL("=="));  
    self->forwardString	    = IoState_retain_(self, SIOSYMBOL("forward"));
    self->initString	    = IoState_retain_(self, SIOSYMBOL("init"));
    self->thisMessageString = IoState_retain_(self, SIOSYMBOL("thisMessage"));
    self->objectString	    = IoState_retain_(self, SIOSYMBOL("Object"));  
    self->parentString	    = IoState_retain_(self, SIOSYMBOL("parent"));
    self->returnString	    = IoState_retain_(self, SIOSYMBOL("return"));
    self->selfString	    = IoState_retain_(self, SIOSYMBOL("self"));
    self->senderString	    = IoState_retain_(self, SIOSYMBOL("sender"));  
    self->setSlotString	    = IoState_retain_(self, SIOSYMBOL("setSlot"));
    self->slotContextString = IoState_retain_(self, SIOSYMBOL("slotContext"));
    self->updateSlotString  = IoState_retain_(self, SIOSYMBOL("updateSlot"));
    self->targetString      = IoState_retain_(self, SIOSYMBOL("target"));  
    
    IoObject_rawSetProto_(seqProto, self->mainActor); 
    
    cFunctionProto = IoCFunction_proto(self);
    self->localsUpdateSlotCFunc = IoState_retain_(self, 
	IoCFunction_newWithFunctionPointer_tag_(self, IoObject_localsUpdateSlot, 0x0));
    
    IoSeq_protoFinish(seqProto);
    IoObject_protoFinish(self);
    IoCFunction_protoFinish(self);
    
    self->setSlotBlock  = IoState_retain_(self, 
	IoObject_getSlot_(self->mainActor, SIOSYMBOL("setSlot")));  
    
    /* --- set up the lobby --- */
    {
	IoObject *object = self->mainActor; 
	IoObject *protos = IOCLONE(object);
	IoObject *vm     = IOCLONE(object);
	IoNumber *number = 0x0;
        IoSandbox *sandbox = 0x0;
	self->lobby      = IOCLONE(object);
	IoState_retain_(self, self->lobby);
	/*IoObject_initProto(self->mainActor, self);*/ /* need to do this after tags and lobby are ready */
	
	/* setup namespace */
	IoObject_setSlot_to_(self->lobby, SIOSYMBOL("Lobby"), self->lobby);
	IoObject_setSlot_to_(self->lobby, SIOSYMBOL("Protos"), protos);  
	IoObject_setSlot_to_(protos, SIOSYMBOL("IoVM"), vm);  
	IoObject_setSlot_to_(vm, SIOSYMBOL("Compiler"), IoCompiler_proto(self));
	IoObject_setSlot_to_(vm, SIOSYMBOL("System"), IoSystem_proto(self));
	
	IoObject_setSlot_to_(vm, SIOSYMBOL("Collector"), IoCollector_proto(self));
	IoObject_setSlot_to_(vm, SIOSYMBOL("Scheduler"), IoScheduler_proto(self));
	
	/* setup proto chain */
	
	IoObject_rawSetProto_(object, self->lobby); 
	IoObject_rawSetProto_(self->lobby, protos); 
	IoObject_rawSetProto_(protos, vm); 
	
	/* add protos to namespace */
	IoObject_setSlot_to_(vm, SIOSYMBOL("Object"), object);
	IoObject_setSlot_to_(vm, SIOSYMBOL("Sequence"), seqProto);
	number = IoNumber_proto(self);
	IoObject_setSlot_to_(vm, SIOSYMBOL("Number"), number); 
	IoObject_setSlot_to_(vm, SIOSYMBOL("Random"), IoRandom_proto(self)); 
	
	self->ioNil = IoNil_proto(self);
	IoObject_setSlot_to_(vm, SIOSYMBOL("Nil"), self->ioNil);
	IoObject_setSlot_to_(vm, SIOSYMBOL("Message"), IoMessage_proto(self));
	
	self->nilMessage  = IoMessage_newWithName_(self, SIOSYMBOL("Nil"));
	IoMessage_cachedResult_(self->nilMessage, self->ioNil);
	IoState_retain_(self, self->nilMessage);
	
	self->printMessage  = IoMessage_newWithName_(self, SIOSYMBOL("print"));
	IoState_retain_(self, self->printMessage);
	
	self->initMessage   = IoMessage_newWithName_(self, SIOSYMBOL("init"));
	IoState_retain_(self, self->initMessage);

	self->compareMessage = IoMessage_newWithName_(self, SIOSYMBOL("compare"));
	IoState_retain_(self, self->compareMessage);
	
	{
	IoObject *debugger = IoDebugger_proto(self);
	IoState_retain_(self, debugger);
	IoObject_setSlot_to_(vm, SIOSYMBOL("Debugger"), debugger);
	self->debugger = debugger;
	
	self->vmWillSendMessage  = IoMessage_newWithName_(self, SIOSYMBOL("vmWillSendMessage"));
	IoMessage_cachedResult_(self->nilMessage, self->ioNil);
	IoState_retain_(self, self->vmWillSendMessage);
	}
	
	IoObject_setSlot_to_(vm, SIOSYMBOL("Block"),      IoBlock_proto(self));
	IoObject_setSlot_to_(vm, SIOSYMBOL("List"),       IoList_proto(self));
	IoObject_setSlot_to_(vm, SIOSYMBOL("Map"),        IoMap_proto(self));
	IoObject_setSlot_to_(vm, SIOSYMBOL("File"),       IoFile_proto(self));
	IoFile_statInit(self);
	
	IoObject_setSlot_to_(vm, SIOSYMBOL("Directory"), IoDirectory_proto(self));
	IoObject_setSlot_to_(vm, SIOSYMBOL("Date"),       IoDate_proto(self));
	IoObject_setSlot_to_(vm, SIOSYMBOL("Duration"),   IoDuration_proto(self));
	IoObject_setSlot_to_(vm, SIOSYMBOL("WeakLink"),   IoWeakLink_proto(self));
        sandbox = IoSandbox_proto(self);
	IoObject_setSlot_to_(vm, SIOSYMBOL("Sandbox"),    sandbox);
#if !defined(__SYMBIAN32__)
	IoObject_setSlot_to_(vm, SIOSYMBOL("DynLib"),     IoDynLib_proto(self));
#endif
	
	self->store = IoStore_proto(self);
	IoStore_addMethods(self->store);
	IoObject_setSlot_to_(vm, SIOSYMBOL("Store"),      self->store);
	IoObject_setSlot_to_(vm, SIOSYMBOL("Future"),     IoFuture_proto(self));
	/*IoObject_setSlot_to_(vm, SIOSYMBOL("PObject"),    IoPObject_proto(self));*/
	IoObject_setSlot_to_(vm, SIOSYMBOL("CFunction"),  cFunctionProto);
	
	{
	    IoException *e = IoException_proto(self);
	    
	    self->errorException        = IOCLONE(e);
	    self->warningException      = IOCLONE(e);
	    self->notificationException = IOCLONE(e);
	    
	    IoObject_setSlot_to_(vm, SIOSYMBOL("Exception"),  e);
	    IoObject_setSlot_to_(vm, SIOSYMBOL("Error"),  self->errorException);
	    IoObject_setSlot_to_(vm, SIOSYMBOL("Warning"),  self->warningException);
	    IoObject_setSlot_to_(vm, SIOSYMBOL("Notification"),  self->notificationException);
	    
	}
	self->localsProto = IoState_retain_(self, IoObject_localsProto(self));
	IoObject_setSlot_to_(vm, SIOSYMBOL("Locals"),  self->localsProto);
	
	/* define Nop object */
	{ 
	    self->ioNop = IoObject_new(self);
	    IoObject_setSlot_to_(self->ioNop, self->forwardString,
				 IoCFunction_newWithFunctionPointer_tag_(self, (IoUserFunction *)IoObject_nop, self->lobby->tag));
	    IoObject_setSlot_to_(vm, SIOSYMBOL("Nop"), self->ioNop);
	    /*IoObject_removeSlot_(self->ioNop, SIOSYMBOL("proto"));*/
	    IoObject_rawRemoveAllProtos(self->ioNop);
	    IoState_retain_(self, self->ioNop);
	}
	
	IoObject_setSlot_to_(self->lobby, SIOSYMBOL("version"), IoNumber_newWithDouble_(self, IO_VERSION_NUMBER));
	IoObject_setSlot_to_(self->lobby, SIOSYMBOL("distribution"), SIOSYMBOL("Io"));
		
#if !defined(__SYMBIAN32__)
	//IoImporterInit(self, vm);
#endif
    
    self->stopStatus = MESSAGE_STOP_STATUS_NORMAL;
    self->returnValue = (void *)0x0;
    
    /*
    List_append_(self->retainedValues, IoObject_createSlotsIfNeeded);
    */
    
    IoState_gcInitPhase(self);
    self->newValueCountToUpdate = 10; 
    self->allocsPerSweep  = STATE_DEFAULT_COLECT_AFTER;
    
    IoState_clearRetainStack(self);
    IoState_popCollectorPause(self);
    IoState_gcSweepPhase(self);
    
    IoVMCodeInit(self, vm);
    IoState_clearRetainStack(self);
    }

    return self;
}

void IoState_init(IoState *self)
{
    if (self->bindingsInitCallback)
    {
	IoState_pushCollectorPause(self);
	{
	IoObject *iovm = IoObject_getSlot_(self->lobby, SIOSYMBOL("IoVM"));
	self->bindingsInitCallback(self, iovm);
	}
	//IoObject_setSlot_to_(self->lobby, SIOSYMBOL("distribution"), SIOSYMBOL("IoDesktop"));
	IoState_popCollectorPause(self);
	IoState_clearRetainStack(self);
    }
}

void IoState_registerProtoWithFunc_(IoState *self, IoObject *proto, IoStateProtoFunc *func)
{ 
    if (Hash_at_(self->primitiveProtos, (void *)func))
    {
	fputs("IoState_registerProtoWithFunc_() Error: attempt to add the same proto twice\n", stdout);
	exit(1);
    }
    
    IoState_retain_(self, proto);
    Hash_at_put_(self->primitiveProtos, (void *)func, proto); 
    /*printf("registered %s\n", IoObject_name(proto));*/
}

IoObject *IoState_protoWithName_(IoState *self, const char *name)
{
    IoObject *proto = Hash_first(self->primitiveProtos);
    
    while (proto)
    {
	if (!strcmp(IoObject_name(proto), name)) 
	{
	    return proto;
	}
	proto = Hash_next(self->primitiveProtos);
    }
    return 0x0;
}

/*
 static void showPrimitivesCallback(IoObject *proto)
 { 
     printf("tag %p %s\n", (void *)(proto->tag), proto->tag->name); 
 }
 */

static void addProtoToCollector(IoObject *proto)
{
    IoState *self = proto->tag->state;
    
    if (strcmp("Object", IoObject_name(proto)) == 0)
    {
	return;
    }
    
    /*printf("addProtoToCollector(%s)\n", IoObject_name(proto));*/
    
    if (!IoState_hasValue_(self, proto)) 
    {
	IoState_addValue_(self, proto);
    }
    
    Hash_removeValue_(self->primitiveProtos, proto);
    IoState_collectGarbage(self);
}

List *IoState_tagList(IoState *self) /* caller must free returned List */
{
    List *tags = List_new();
    void *k = Hash_firstKey(self->primitiveProtos);
    
    while (k)
    {
	IoObject *proto = Hash_at_(self->primitiveProtos, k);
	List_append_(tags, proto->tag);
	k = Hash_nextKey(self->primitiveProtos);
    }
    return tags;
}

void IoState_free(IoState *self)
{
    List *tags = IoState_tagList(self);
    /* we'll need to free the tags after freeing their protos */
    
    self->isCleaningUp = 1;
    
    MainArgs_free(self->mainArgs);
    self->mainArgs = 0x0;
        
    List_empty(self->actors);
    /*if (self->currentActor) IoState_retain_(self, self->currentActor); */
    List_empty(self->retainedValues);
    
    {
	IoObject *v;
	
	while ((v = List_pop(self->recycledObjects))) 
	{
	    IoObject_dealloc(v);
	}
	
	List_free(self->recycledObjects);
	self->recycledObjects = 0x0;
	
	while ((v = List_pop(self->recycledObjectsSansSlots))) IoObject_dealloc(v);
	List_free(self->recycledObjectsSansSlots);
	self->recycledObjectsSansSlots = 0x0;
    }
    
    self->lobby = 0x0;
    IoState_collectGarbage(self);
    IoState_collectGarbage(self); /* 2nd time to catch any non-white objects from last collect */
    
    /*Hash_do_(self->primitiveProtos, (HashDoCallback *)addProtoToCollector);*/
    
    while (Hash_first(self->primitiveProtos) && Hash_count(self->primitiveProtos) > 1)
    {
	addProtoToCollector(Hash_first(self->primitiveProtos));
    }
    
    /*Hash_clean(self->primitiveProtos);*/
    
    /*IoState_collectGarbage(self);*/ /* final collection */
    
    SkipDBM_free(self->sdbm);
    self->sdbm = 0x0;
    self->symbols = 0x0;
    
    List_free(self->retainedValues);
    self->retainedValues = 0x0;
    
    /* the protos that these tags were alloced 
	in were already collected so we free them directly 
    */
    List_do_(tags, (ListDoCallback *)IoTag_free); 
    Hash_free(self->primitiveProtos);
    self->primitiveProtos = 0x0;
    
    List_free(self->actors);
    self->actors = 0x0;
    
    Hash_free(self->valueToWeakLink);
    self->valueToWeakLink = 0x0;
    
    Scheduler_free(self->scheduler);
    self->scheduler = 0x0;
    
    /*Stack_free(self->collectedWeakLinks);*/  
    IoObjectGroup_free(self->freeGroup);
    self->freeGroup = 0x0;
    IoObjectGroup_free(self->blackGroup);
    self->blackGroup = 0x0;
    IoObjectGroup_free(self->grayGroup);
    self->grayGroup = 0x0;
    IoObjectGroup_free(self->whiteGroup);
    self->whiteGroup = 0x0;

    /*IoStore_free(self->store);*/

    List_free(tags);
    free(self);
}

IoObject *IoState_lobby(IoState *self) 
{ 
    return self->lobby; 
}

void IoState_setLobby_(IoState *self, IoObject *obj) 
{ 
    self->lobby = obj;
}

IoObject *IoState_protoWithInitFunction_(IoState *self, IoStateProtoFunc *func)
{ 
    IoObject *proto = Hash_at_(self->primitiveProtos, (void *)func); 
    
    if (!proto)
    {
	fputs("IoState_protoWithInitFunction_() Error: missing proto\n", stdout);
	exit(1);
    }
    return proto;
}

/* --- performing ----------------------------- */

unsigned char IoState_yieldIfNeeded(IoState *self)
{
    if (Scheduler_coroCount(self->scheduler) > 1) 
    { 
	IoObject_yield(self->mainActor); 
	return 1; 
    }
    
    return 0;
}

IoObject *IoState_rawDoCString_(void *arg1, void *arg2, void *arg3)
{
    IoState *self = arg1;
    IoObject *result = self->ioNil;

    IoState_pushRetainPool(self);
    
    result = IoObject_rawDoString_label_(self->lobby, SIOSYMBOL((char *)arg2), SIOSYMBOL((char *)arg3));

    IoState_popRetainPoolExceptFor_(self, result);
    
    return result;
}

IoObject *IoState_doCString_(IoState *self, const char *s)
{
    IoException *e;
    IoObject *r;
    
    self->messageCount = 0;
    self->endTime = 0;
    r = IoState_tryFunc_(self, (IoCatchCallback *)IoState_rawDoCString_, 
				   self, (void *)s, "[command line]", &e);
    if (e) 
    {
	IoState_exception_(self, e);
    }
    
    return r;
}

IoObject *IoState_doSandboxCString_(IoState *self, const char *s)
{
    IoException *e;
    IoObject *r = 0x0;

    {
        struct timeval startTv;
        double start;

        /* Get the start and current time */
        gettimeofday(&startTv, NULL);
        start = (double)startTv.tv_sec + ((double)startTv.tv_usec/1000000.0);
        
        /* Calculate the end of time */
        self->endTime = start + self->timeLimit;
    }
    
    self->messageCount = self->messageCountLimit;

    r = IoState_tryFunc_(self, (IoCatchCallback *)IoState_rawDoCString_, 
	self, (void *)s, "[command line]", &e);
	
    if (e) 
    {
	IoState_exception_(self, e);
    }
    
    return r;
}

double IoState_endTime(IoState *self)
{
    return self->endTime;
}

void *IoState_rawDoFile_(void *arg1, void *arg2, void *arg3)
{
    IoState *self = arg1;
    char *path = arg2;
    IoFile *file = IoFile_newWithPath_(self, SIOSYMBOL(path));
    return IoFile_rawDo(file, self->lobby);
    /*
     while (Scheduler_coroCount(self->scheduler) > 1) 
     { 
	IoObject_yield(self->mainActor); 
     }
     */
}

IoObject *IoState_doFile_(IoState *self, const char *path)
{
    IoException *e;
    IoObject *result = IoState_tryFunc_(self, (IoCatchCallback *)IoState_rawDoFile_,
					self, (void *)path, 0x0, &e);
    if (e) 
    {
	IoState_exception_(self, e);
    }
    
    return result;
}

/* --- actors ------------------------------------------------------ */

void IoState_addActor_(IoState *self, IoObject *actor) 
{ 
    /*IoObject_makeGrayIfWhite(actor);*/
    List_append_(self->actors, actor); 
}

void IoState_removeActor_(IoState *self, IoObject *actor) 
{ 
    List_remove_(self->actors, actor); 
}

IoObject *IoState_currentActor(IoState *self) 
{ 
    return self->currentActor; 
}

void IoState_currentActor_(IoState *self, IoObject *actor) 
{ 
    self->currentActor = actor; 
    self->currentRetainStack = actor->coro->retainedValues;
}

int IoState_activeCoroCount(IoState *self)
{
    return Scheduler_coroCount(self->scheduler);
}

void IoState_yield(IoState *self) 
{ 
    IoObject_yield(self->currentActor); 
}

void *IoState_prepareToPause(IoState *self) 
{ 
    IoObject_prepareToPause(self->currentActor);
    return self->currentActor;
}

void IoState_resume_(IoState *self, IoObject *actor) 
{ 
    IoObject_resume(actor);
}

void IoState_print_(IoState *self, const char *format, ...)
{
    ByteArray *ba;
    va_list ap;
    va_start(ap, format);
    ba = ByteArray_newWithVargs_(format, ap);
    IoState_justPrint_(self, ByteArray_asCString(ba));
    ByteArray_free(ba);
    va_end(ap);
}

/* --- command line ------------------------------------------------------ */

void IoState_argc_argv_(IoState *self, int argc, const char *argv[])
{ 
    MainArgs_argc_argv_(self->mainArgs, argc, argv); 
}

/* --- callbacks: print, error, exit --------------------------- */

void IoState_setBindingsInitCallback(IoState *self, IoStateBindingsInitCallback *callback) 
{ 
    self->bindingsInitCallback = callback; 
}

void IoState_callbackContext_(IoState *self, void *context) 
{ 
    self->callbackContext = context; 
}

void *IoState_callbackContext(IoState *self) 
{ 
    return self->callbackContext; 
}

/* --- print callback --------------------------- */

void IoState_printCallback_(IoState *self, IoStatePrintCallback *callback)
{ 
    self->printCallback = callback; 
}

void IoState_justPrint_(IoState *self, const char *s)
{ 
    if (self->printCallback) 
    { 
	self->printCallback(self->callbackContext, s); 
    }
    else 
    { 
	fputs(s, stdout); 
    }
}

/* --- exception callback --------------------------- */

void IoState_exceptionCallback_(IoState *self, IoStateExceptionCallback *callback)
{ 
    self->exceptionCallback = callback; 
}

void IoState_exception_(IoState *self, IoException *e)
{ 
    if (self->exceptionCallback) 
    { 
	self->exceptionCallback(self->callbackContext, e); 
    }
    else 
    { 
	IoException_printBackTrace(e);
    } 
}

/* --- exit callback --------------------------- */

void IoState_exitCallback_(IoState *self, IoStateExitCallback *callback)
{ 
    self->exitCallback = callback; 
}

void IoState_exit(IoState *self, int returnCode)
{ 
    fflush(stdout);
    
    if (self->exitCallback) 
    { 
	self->exitCallback(self->callbackContext, returnCode); 
    }
    else 
    { 
	IoState_free(self);
	/*abort();*/
	exit(returnCode);
    } 
}

/* --- active coro callback --------------------------- */

void IoState_activeCoroCallback_(IoState *self, IoStateActiveCoroCallback *callback)
{
    self->activeCoroCallback = callback; 
}

void IoState_schedulerUpdate(IoState *self, int count)
{
    if (self->activeCoroCallback)
    {
	self->activeCoroCallback(self->callbackContext, count);
    } 
}

/* --- Store --------------------------------------------------- */

IoStore *IoState_store(IoState *self)
{
    return self->store;
}

/* --- debugging --------------------------------------------------- */

void IoState_show(IoState *self)
{
    printf("--- state ----------------------------------\n");
    printf("State:\n");
    /*
    printf("black:\n");
    IoObjectGroup_show(self->blackGroup);
    printf("\n");

    printf("gray:\n");
    IoObjectGroup_show(self->grayGroup);  
    printf("\n");

    printf("white:\n");
    IoObjectGroup_show(self->whiteGroup);    
    printf("\n");
    */
    printf("stacks:\n");
    /*List_do_(self->actors, (ListDoCallback *)IoObject_showStack);*/
    printf("\n");  
}

IoObject *IoState_replacePerformFunc_with_(IoState *self, 
    TagPerformFunc *oldFunc, 
    TagPerformFunc *newFunc)
{
    IoObject *proto = Hash_first(self->primitiveProtos);
    
    while (proto)
    {
	if (proto->tag->performFunc == oldFunc)
	{
	    proto->tag->performFunc = newFunc;
	}
	
	proto = Hash_next(self->primitiveProtos);
    }
    
    return 0x0;
}

void IoState_debuggingOn(IoState *self)
{
    IoState_replacePerformFunc_with_(self, 
	(TagPerformFunc *)IoObject_perform, 
	(TagPerformFunc *)IoObject_performWithDebugger);
}

void IoState_debuggingOff(IoState *self)
{
    IoState_replacePerformFunc_with_(self, 
	(TagPerformFunc *)IoObject_performWithDebugger, 
	(TagPerformFunc *)IoObject_perform);
}

int IoState_hasDebuggingActor(IoState *self)
{
    return (List_detect_(self->actors, 
	(ListDetectCallback *)IoObject_debuggingOn) != (void *)0x0);
}

void IoState_updateDebuggingMode(IoState *self)
{
    if (IoState_hasDebuggingActor(self))
    {
	IoState_debuggingOn(self);
    }
    else
    {
	IoState_debuggingOff(self);
    }
}

