/*   
    copyright: Steve Dekorte, 2002. All rights reserved.
    license: See _BSDLicense.txt.
    
    object: BufferEnumerator
    prototype: Object
    module: IoVM
    description: 
 */

#include "IoBufferEnumerator.h"
#include "IoState.h"
#include "IoCFunction.h"
#include "IoObject.h"
#include "IoNil.h"
#include "IoNumber.h"
#include "IoMessage.h"
#include "IoList.h"
#include "IoString.h"
#include <ctype.h>

#define DATA(self) ((IoBufferEnumeratorData *)self->data)

IoTag *IoBufferEnumerator_tag(void *state)
{
    IoTag *tag = IoTag_newWithName_("BufferEnumerator");
    tag->state = state;
    tag->freeFunc = (TagFreeFunc *)IoBufferEnumerator_free;
    tag->cloneFunc = (TagCloneFunc *)IoBufferEnumerator_rawClone;
    tag->compareFunc = (TagCompareFunc *)IoBufferEnumerator_compare;
    /*tag->markFunc = (TagMarkFunc *)IoBufferEnumerator_mark;*/
    tag->writeToStoreOnStreamFunc  = (TagWriteToStoreOnStreamFunc *)IoBufferEnumerator_writeToStore_stream_;
    tag->readFromStoreOnStreamFunc = (TagReadFromStoreOnStreamFunc *)IoBufferEnumerator_readFromStore_stream_;
    return tag;
}

void IoBufferEnumerator_writeToStore_stream_(IoBufferEnumerator *self, IoStore *store, BStream *stream)
{
    //BStream_writeTaggedByteArray_(stream, BIVAR(self));
}

void IoBufferEnumerator_readFromStore_stream_(IoBufferEnumerator *self, IoStore *store, BStream *stream)
{
    //BStream_readTaggedByteArray_(stream, BIVAR(self));
}

IoBufferEnumerator *IoBufferEnumerator_proto(void *state)
{
    IoMethodTable methodTable[] = {
	{"setBuffer", IoBufferEnumerator_setBuffer},
	{"buffer", IoBufferEnumerator_buffer},

	{"setPosition", IoBufferEnumerator_setPosition},
	{"position", IoBufferEnumerator_position},
		
	{"setStopStrings", IoBufferEnumerator_setStopStrings},
	{"stopStrings", IoBufferEnumerator_stopStrings},
	
	{"next", IoBufferEnumerator_next},
	{NULL, NULL},
    };

    IoObject *self = IoObject_new(state);

    self->tag = IoBufferEnumerator_tag(state);
    self->data = ByteArray_new();
    IoState_registerProtoWithFunc_(state, self, IoBufferEnumerator_proto);
    IoObject_addMethodTable_(self, methodTable);

    return self;
}

IoBufferEnumerator *IoBufferEnumerator_rawClone(IoBufferEnumerator *self)
{
  IoBufferEnumerator *child = IoObject_rawClonePrimitive(self);
  child->data = ByteArray_clone(BIVAR(self));
  return child;
}

/* ----------------------------------------------------------- */

IoBufferEnumerator *IoBufferEnumerator_new(void *state)
{
  IoBufferEnumerator *proto = IoState_protoWithInitFunction_(state, IoBufferEnumerator_proto);
  return IoBufferEnumerator_rawClone(proto);
}

IoBufferEnumerator *IoBufferEnumerator_newWithBuffer_(void *state, IoBuffer *buffer)
{
  IoBufferEnumerator *self = IoBufferEnumerator_new(state);
  DATA(self)->buffer = IOREF(buffer);
  return self;
}

IoObject *IoBufferEnumerator_position(IoBufferEnumerator *self, IoObject *locals, IoMessage *m)
{
    /*
    method: buffer
    description: Returns the enumerator buffer.
    */
    return IONUMBER(DATA(self)->buffer);
}

IoObject *IoBufferEnumerator_setBuffer(IoBufferEnumerator *self, IoObject *locals, IoMessage *m)
{
    /*
    method: setBuffer(aBuffer)
    description: Sets the buffer to enumerate and resets the enumerator position to zero.
    */
    DATA(self)->buffer = IOREF(IoMessage_locals_bufferArgAt_(m, locals, 0));
    return self;
}

IoObject *IoBufferEnumerator_rewind(IoBufferEnumerator *self, IoObject *locals, IoMessage *m)
{
    /*
    method: rewind
    description: Sets the enumerator position to zero.
    */
    DATA(self)->position = 0;
    return self;
}


IoObject *IoBufferEnumerator_position(IoBufferEnumerator *self, IoObject *locals, IoMessage *m)
{
    /*
    method: position
    description: Returns the enumerator position.
    */
    return IONUMBER(DATA(self)->position);
}

IoObject *IoBufferEnumerator_setPosition(IoBufferEnumerator *self, IoObject *locals, IoMessage *m)
{
    /*
    method: setPosition(aNumber)
    description: Sets the enumerator position.
    */
    DATA(self)->position = IoMessage_locals_intArgAt_(m, locals, 0);
    return self;
}

int IoBufferEnumerator_matchesDatum_(IoBufferEnumerator *self, Datum *d)
{
    List *list = IoList_rawList(DATA(self)->list);
    return List_detect_withArg_(list, IoBuffer_rawBeginsWithDatum_, d);
}

IoObject *IoBufferEnumerator_nextString(IoBufferEnumerator *self, IoObject *locals, IoMessage *m)
{
    /*
    method: next
    description: 
    */
    size_t pos = DATA(self)->position;
    ByteArray *ba = IoBuffer_rawByteArray(DATA(self)->buffer);
    size_t length = ByteArray_length(ba);
    
    if (pos < length)
    {
	Datum d = ByteArray_datumAt_(ba, pos);
	
	int index = Datum_detect_withTarget_(&d, (DatumDetectWithFunc *)IoBufferEnumerator_matchesDatum, (void *)self);
	/*int index = Datum_datumUntil_withTarget_(&d, (DatumDetectWithFunc *)IoBufferEnumerator_matchesDatum, (void *)self);*/

	if (index == -1) 
	{
	    index = length; 
	}
	
	DATA(self)->position = index; 
	return IoState_stringWithCString_length_(IOSTATE, ba->bytes + pos, index - pos);
    }
    
    return IONIL(self);
}

IoObject *IoBufferEnumerator_setStopStrings(IoBufferEnumerator *self, IoObject *locals, IoMessage *m)
{
    /*
    method: setStopStrings(aList)
    description: Sets the stop-string list.
    */
    IoList *newList = IoMessage_locals_listArgAt_(m, locals, 0);
    int hasNonStringOfBuffer = List_detect_(IoList_rawList(newList), (ListDetectCallback *)IoObject_isNotStringOrBuffer);
    IOASSERT(hasNonStringOfBuffer, "list must only contain Strings or Buffers");
    return DATA(self)->stopStrings = IOREF(newList);    
}

IoObject *IoBufferEnumerator_stopStrings(IoBufferEnumerator *self, IoObject *locals, IoMessage *m)
{
    /*
    method: stopStrings
    description: Returns the stop-string list.
    */
    return DATA(self)->stopStrings;
}



