/*#io
Map ioDoc(
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")    
docObject("Coroutine")
docDescription("Object wrapper for an Io coroutine.")
*/
 
#include "IoCoroutine.h"
#include "IoObject.h"
#include "IoState.h"
#include "IoCFunction.h"
#include "IoNil.h"
#include "IoSeq.h"
#include "IoState.h"
#include "IoNumber.h"
#include "IoList.h"
#include "IoBlock.h"

#define IVAR(self) ((IoCoroutineData *)self->data)

IoTag *IoCoroutine_tag(void *state)
{
    IoTag *tag = IoTag_newWithName_("Coroutine");
    tag->state = state;
    tag->freeFunc  = (TagFreeFunc *)IoCoroutine_free;
    tag->cloneFunc = (TagCloneFunc *)IoCoroutine_rawClone;
    tag->markFunc  = (TagMarkFunc *)IoCoroutine_mark;
    tag->writeToStoreOnStreamFunc  = (TagWriteToStoreOnStreamFunc *)IoCoroutine_writeToStore_stream_;
    tag->readFromStoreOnStreamFunc = (TagReadFromStoreOnStreamFunc *)IoCoroutine_readFromStore_stream_;
    return tag;
}

void IoCoroutine_writeToStore_stream_(IoCoroutine *self, IoStore *store, BStream *stream)
{
    IoCoroutineData *data = IVAR(self);

}

void IoCoroutine_readFromStore_stream_(IoCoroutine *self, IoStore *store, BStream *stream)
{
    IoCoroutineData *data = IVAR(self);

}

IoCoroutine *IoCoroutine_proto(void *state)
{
    IoMethodTable methodTable[] = {
	{"retainedValues",	IoCoroutine_retainedValues},
	{"exceptionCatch",	IoCoroutine_exceptionCatch},
	{"exception",		IoCoroutine_exception},
	{NULL, NULL},
    };

    IoObject *self = IoObject_new(state);

    self->tag = IoCoroutine_tag(state);
    self->data = calloc(1, sizeof(IoCoroutineData));
    self->retainedValues = Stack_new();

    IoState_registerProtoWithFunc_(state, self, IoCoroutine_proto);

    IoObject_addMethodTable_(self, methodTable);
    return self;
}

IoCoroutine *IoCoroutine_rawClone(IoCoroutine *proto) 
{ 
    IoObject *self = IoObject_rawClonePrimitive(proto);
    self->data = calloc(1, sizeof(IoCoroutineData));
    self->retainedValues = Stack_new();
    return self; 
}

IoCoroutine *IoCoroutine_new(void *state)
{
    IoObject *proto = IoState_protoWithInitFunction_(state, IoCoroutine_proto);
    return IOCLONE(proto);
}

void IoCoroutine_free(IoCoroutine *self) 
{
    Stack_free(IVAR(self)->retainedValues);
	free(IVAR(self));
}

void IoCoroutine_mark(IoCoroutine *self) 
{ 
    Stack_do_(IVAR(self)->retainedValues, (ListDoCallback *)IoObject_makeGrayIfWhite);
    
    if (IVAR(self)->exception) 
    {
		IoObject_makeGrayIfWhite(IVAR(self)->exception);
    }
}

Stack *IoCoroutine_rawRetainedValues(IoCoroutine *self)
{ 
    return IVAR(self)->retainedValues; 
}

int IoCoroutine_rawRetainingCount(IoCoroutine *self)
{ 
    return Stack_totalSize(IVAR(self)->retainedValues); 
}

void IoCoroutine_rawShow(IoCoroutine *self)
{
    Stack_do_(IVAR(self)->retainedValues, (StackDoCallback *)IoObject_show);
    printf("\n");
}

/* ----------------------------------------------------------- */

IoObject *IoCoroutine_retainedValues(IoCoroutine *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("retainedValues", "Returns List of values on this coroutine's stack.")
    */
    
    return self; //IVAR(self)->retainedValues; 
}

/* ------------------------------------------------------------ */

void IoObject_createCoroIfAbsent(IoObject *self)
{
    IoState *state = IOSTATE;

    if (!IVAR(self)->cid)
    { 
		IoCoroutine *current = IoState_currentCoroutine(state);
		Coro *coro = Scheduler_newCoro(state->scheduler);
		Coro_label_(coro, "IoCoroutine");
		Coro_object_(coro, self);
		Coro_unyieldCallback_(coro, (CoroUnyieldCallback *)IoCoroutine_prepareToUnyield);
		IVAR(self)->cid = coro;
		Scheduler_startCoro(state->scheduler, coro, (void *)self, IoCoroutine_run);
		/*printf("self %p coro %p tid %p\n", self, self->coro, self->coro->cid);*/
		IoState_currentCoroutine_(state, current);
		IoState_addCoroutine_(state, self);
    }
}

IoMessage *IoMessage_asMessageWithEvaluatedArgs(IoMessage *self, IoObject *locals)
{	
    IoState *state = IOSTATE;
    IoMessage *sendMessage;
	int i, max = IoMessage_argCount(self);
	
	if (IoMessage_needsEvaluation(self))
	{ 
	    sendMessage = IoMessage_newWithName_(state, IoMessage_name(self)); 
	}
	else 
	{ 
	    sendMessage = self; 
	}
	
	for (i = 0; i < max; i++)
	{
	    IoMessage *arg = IoMessage_rawArgAt_(self, i);
	    IoObject *result = IoMessage_locals_performOn_(arg, locals, locals);
	    IoMessage_setCachedArg_to_(sendMessage, i, result);
	}
    
    return sendMessage;
}

/* --- coroutine calls ----------------------------------------------- */

/*

Object clone become(result)

Object actorProcessQueue := method(
	// this is called from within the object's coroutine
	
	while(f := actorFutureQueue removeFirst,
		getSlot("f") setResult(self doMessage(getSlot("f") message))
		actorCoroutine yield
	)
)

Object actorRun := method(
	
	self actorCoroutine := Coroutine clone
	actorCoroutine start(self, actorProcessQueue) // exception handling?
)

Object futureSend := method(
	m := thisMessage argAt(0) asMessageWithEvaluatedArgs(sender)
	f := Future clone setMessage(m) setTarget(self) 
	actorFutureQueue append(f)
	f setTrigger
	self actorRun
)

Object asyncSend := method(
	m := thisMessage argAt(0) asMessageWithEvaluatedArgs(sender)
	f := Future clone setMessage(m) setTarget(self) 
	actorFutureQueue append(f)
	Nil
)
*/

void IoCoroutine_prepareToUnyield(IoObject *self)
{
    IoState_currentCoroutine_(IOSTATE, self);
}

void IoCoroutine_yield(IoObject *self)
{
    /*printf("yielding %p %s\n", self, Coro_label(Scheduler_currentCoro(IOSTATE->scheduler)));*/
    Scheduler_yield(IOSTATE->scheduler);
    /*printf("unyielding %p %s\n", self, Coro_label(Scheduler_currentCoro(IOSTATE->scheduler)));*/
}

void IoCoroutine_abort(IoObject *self)
{ 
	IoState *state = IOSTATE;
	Coro *cid = IVAR(self)->cid;
	
    if (cid)
    {
		IoState_removeCoroutine_(state, self);
		/*printf("aborting %p\n", p);*/
		Scheduler_prepareToKillCoro_(state->scheduler, cid);
		IVAR(self)->cid = 0x0;
    }
}

void IoCoroutine_prepareToPause(IoObject *self)
{
	IoState *state = IOSTATE;
	Coro *cid = IVAR(self)->cid;
	
    if (cid)
	{
		if (!Scheduler_hasCoro_(state->scheduler, cid)) 
		{
			return; /* doesn't work on main actor */
		}
		
		Scheduler_prepareToPauseCoro_(state->scheduler, cid);
    }
}

void IoCoroutine_pause(IoObject *self)
{
	IoState *state = IOSTATE;
	Coro *cid = IVAR(self)->cid;
	
    if (cid)
    {
		/*printf("pausing %p\n", self);*/
		IoCoroutine_prepareToPause(self);
		IoCoroutine_yield(self);
    }
}

unsigned int IoCoroutine_isPaused(IoObject *self)
{ 
	IoState *state = IOSTATE;
	Coro *cid = IVAR(self)->cid;
	
    return (cid && !Scheduler_hasCoro_(state->scheduler, cid));
}

void IoCoroutine_resume(IoObject *self)
{
	IoState *state = IOSTATE;
	Coro *cid = IVAR(self)->cid;
	 
    if (cid)
    {  
		/*printf("Object resuming %p\n", self);*/
		Scheduler_resumeCoro_(state->scheduler, cid);
    }
}

/* --- proto methods ---------------------------------- */

IoObject *IoCoroutine_actorYield(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    IoCoroutine_yield(((IoState *)IOSTATE)->currentActor);
    return self;
}

IoObject *IoCoroutine_actorResume(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    IoCoroutine_resume(self);
    return self; 
}

IoObject *IoCoroutine_actorPause(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    IoCoroutine_pause(self); 
    return self; 
}

IoObject *IoCoroutine_isActive(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    IoState *state = IOSTATE;
    
    if (!self->coro || !Scheduler_hasCoro_(state->scheduler, self->coro->cid)) 
    { 
	return state->ioNil; 
    }
    
    return self;
}

IoObject *IoCoroutine_actorWait(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    double seconds = IoMessage_locals_doubleArgAt_(m, locals, 0);
    IoState *state = IOSTATE;
    IoObject *currentActor = IoState_currentActor(state);
    double waited = Scheduler_waitSeconds_(state->scheduler, seconds);
    IoState_currentActor_(state, currentActor);
    return IONUMBER(waited);
}

unsigned int IoCoroutine_isPausing_(IoObject *self, IoObject *other)
{
    if (self->coro)
    {
	if (Queue_firstTrueFor_with_(self->coro->queue, 
				     (QueueFirstTrueForWithCallback *)IoFuture_isPausing_, other))
	{ 
	    return 1; 
	}
    }
    return 0;
}

void IoCoroutine_showStack(IoObject *self)
{
    printf("Actor %p %s:\n", (void *)self, IoCoroutine_name(self));
    printf("\n");
}

/* --- stack trace ------------------------------------------ */

IoObject *IoCoroutine_try(IoObject *self, IoObject *locals, IoMessage *m)
{
    IoException *e;
    IoMessage *doMessage;
    IoMessage_assertArgCount_receiver_(m, 1, self);
    doMessage = IoMessage_rawArgAt_(m, 0);
    /*printf("%s\n", CSTRING(IoMessage_name(m)));*/
    IoState_tryFunc_(IOSTATE, (IoCatchCallback *)IoMessage_locals_performOn_, doMessage, locals, self, &e);
    return e ? e : IONOP(self);
}

/* --- stack trace ------------------------------------------ */

int IoObject_isBlockLocals(IoObject *self)
{
    IoState *state = IOSTATE;
    IoMessage *m   = IoObject_rawGetSlot_(self, state->thisMessageString); /* calling message */
    IoBlock *block = IoObject_rawGetSlot_(self, state->thisBlockString);
    return (m && block && ISMESSAGE(m) && ISBLOCK(block));
}

static void IoList_addIfBlockLocals_(IoList *self, IoObject *v)
{
    if (v && IoObject_isBlockLocals(v)) 
    {
	IoList_rawAppend_(self, v); 
    }
}

IoList *IoObject_callStack(IoObject *self)
{
    if (self->coro)
    {
	IoList *calls = IoList_new(IOSTATE);
	Stack_do_on_(self->coro->retainedValues, 
		     (StackDoOnCallback *)IoList_addIfBlockLocals_, calls);
	List_reverse(IoList_rawList(calls));
	return calls;
    }
    
    return IONIL(self);
}

void IoObject_appendStackEntryDescription(IoObject *self, ByteArray *ba)
{ 
    IoMessage *m = IoObject_rawGetSlot_(self, IOSTATE->thisMessageString);
    
    if (!m)
    {
	ByteArray_fromFormat_(ba, "0x0     %# ? \t ? \t ?\n", 20 - 10);
	return;
    }
    
    {
	ByteArray *stackEntryDescription = IoMessage_asStackEntryDescription(m);
	ByteArray_append_(ba, stackEntryDescription);
	ByteArray_free(stackEntryDescription);
    }
}

ByteArray *IoObject_backTraceOnCallStackDescription_(IoObject *self, IoList *calls)
{
    ByteArray *ba = ByteArray_new();
    List *list = IoList_rawList(calls);
    List_do_with_(list, (ListDoWithCallback *)IoObject_appendStackEntryDescription, ba);
    return ba;
}
