/*   
copyright: Steve Dekorte, 2002. All rights reserved.
license: See _BSDLicense.txt.

object: Block
prototype: Object
module: IoVM
description: Blocks are anonymous functions (messages with their own locals object).
They are typically used to represent object methods.
*/

#include "IoLocals.h"
#include "IoObject.h"
#include "IoObject_struct.h"
#include "IoState.h"
#include "IoMessage.h"
#include "IoMessage_parser.h"
#include "IoCFunction.h"
#include "IoString.h"
#include "IoNumber.h"
#include "IoNil.h"
#include "IoList.h"
#include "ByteArray.h"

#define DATA(self) ((IoLocalsData *)self->data)

IoTag *IoLocals_tag(void *state)
{
    IoTag *tag = IoTag_newWithName_("Block");
    tag->state = state;
    tag->cloneFunc = (TagCloneFunc *)IoLocals_rawClone;
    tag->markFunc = (TagMarkFunc *)IoLocals_mark;
    tag->freeFunc = (TagFreeFunc *)IoLocals_free;
    tag->activateFunc = (TagActivateFunc *)IoLocals_target_locals_call_;
    tag->writeToStoreOnStreamFunc  = (TagWriteToStoreOnStreamFunc *)IoLocals_writeToStore_stream_;
    tag->readFromStoreOnStreamFunc = (TagReadFromStoreOnStreamFunc *)IoLocals_readFromStore_stream_;
    return tag;
}

void IoLocals_copy_(IoLocals *self, IoLocals *other)
{
    DATA(self)->message = IOREF(DATA(other)->message);
    
    {
	List *l1 = DATA(self)->argNames;
	List *l2 = DATA(other)->argNames;
	int i, max = List_count(l2);
	
	List_clear(l1);
	
	for (i = 0; i < max; i ++)
	{
	    List_add_(l1, IOREF(List_at_(l2, i)));
	}
    }
    
    if (DATA(other)->scope) 
    {
	IOREF(DATA(other)->scope);
    }
    
    DATA(self)->scope = DATA(other)->scope;
}

void IoLocals_writeToStore_stream_(IoLocals *self, IoStore *store, BStream *stream)
{
    ByteArray *ba = IoLocals_justCode(self);
    BStream_writeTaggedByteArray_(stream, ba);
    /*printf("write block '%s'\n", ByteArray_asCString(ba));*/
    ByteArray_free(ba);
    
    if (DATA(self)->scope)
    {
	BStream_writeTaggedInt32_(stream, IoStore_pidForObject_(store, DATA(self)->scope));
    }
    else
    {
	BStream_writeTaggedInt32_(stream, 0);
    }
}

void IoLocals_readFromStore_stream_(IoLocals *self, IoStore *store, BStream *stream)
{
    IoLocals *newBlock = 0x0;
    ByteArray *ba = BStream_readTaggedByteArray(stream);
    
    /*printf("read block [[[%s]]]]\n", ByteArray_asCString(ba));*/
    newBlock = IoState_doCString_(IOSTATE, ByteArray_asCString(ba));
    
    if (!newBlock || !ISBLOCK(newBlock))
    {
	IoState_error_description_(IOSTATE, 0x0, "Store found bad block code", (char *)ByteArray_bytes(ba));
    }
    
    IoLocals_copy_(self, newBlock);
    
    {
	PID_TYPE pid = BStream_readTaggedInt32(stream);
	
	if (pid)
	{
	    DATA(self)->scope = IoStore_objectWithPid_(store, pid);
	}
	else
	{
	    DATA(self)->scope = 0x0;
	}
    }
}

IoLocals *IoLocals_proto(void *vState)
{
    IoState *state = (IoState *)vState;
    IoMethodTable methodTable[] = {
    {"print", IoLocals_print},
    {"code", IoLocals_code},
    {"message", IoLocals_message},
    {"setMessage", IoLocals_setMessage},
    {"argumentNames", IoLocals_argumentNames},
    {"setArgumentNames", IoLocals_argumentNames_},
    {"setScope", IoLocals_setScope_},
    {"scope", IoLocals_scope},
    /*{"performOn", IoLocals_performOn_}, depricated */
	/* expiremental Smalltalk stuff */
    {"ifTrue", IoLocals_ifTrue_}, 
    {"ifFalse", IoLocals_ifFalse_},
    {"whileTrue", IoLocals_whileTrue_},
    {"whileFalse", IoLocals_whileFalse_},
    {NULL, NULL},
    };
    
    IoObject *self = IoObject_new(state);
    
    self->data = calloc(1, sizeof(IoLocalsData));
    self->tag = IoLocals_tag(state);
    DATA(self)->message  = ((IoState *)state)->nilMessage;
    DATA(self)->argNames = List_new();
    DATA(self)->scope    = 0x0;
    IoState_registerProtoWithFunc_(state, self, IoLocals_proto);
    
    IoObject_addMethodTable_(self, methodTable);
    return self;
}

IoLocals *IoLocals_rawClone(IoLocals *self)
{
    IoObject *child = IoObject_rawClonePrimitive(self);
    child->data = calloc(1, sizeof(IoLocalsData));
    child->tag = self->tag;
    DATA(child)->message  = IoMessage_deepCopyOf_(DATA(self)->message);
    DATA(child)->argNames = List_clone(DATA(self)->argNames);
    DATA(child)->scope    = DATA(self)->scope;
    /*IoObject_setSlot_to_(child, IOSTRING("proto"), self);*/
    return child;
}

IoLocals *IoLocals_new(IoState *state)
{
    IoObject *proto = IoState_protoWithInitFunction_(state, IoLocals_proto);
    return IoLocals_rawClone(proto);
}

void IoLocals_mark(IoLocals *self)
{ 
    IoObject_makeGrayIfWhite(DATA(self)->message);
    if (DATA(self)->scope) IoObject_makeGrayIfWhite(DATA(self)->scope);
    List_do_(DATA(self)->argNames, (ListDoCallback *)IoObject_makeGrayIfWhite);
}

void IoLocals_free(IoLocals *self)
{
    if(self->data)
    {
	List_free(DATA(self)->argNames);;
	free(self->data);
    }
    
    self->data = NULL;
}

void IoLocals_message_(IoLocals *self, IoMessage *m) { DATA(self)->message = IOREF(m); }

/* --- parsing -------------------------------------------------------- */
IoLocals *IoLocals_newFromCString_(IoState *state, char *s)
{
    IoLocals *self = IoLocals_new(state);
    IoLocals_message_(self, IoMessage_newFromText_label_(state, s, 
							"[IoLocals_newFromCString_]"));
    return self;
}

/* --- calling -------------------------------------------------------- */

inline IoObject *IoLocals_target_locals_call_blockLocals_state_(
    IoLocals *self, 
    IoObject *target, 
    IoObject *locals, 
    IoMessage *m, 
    IoObject *blockLocals, 
    IoState *state)
{
    IoObject *result;
    IoLocalsData *selfData = DATA(self);
    PHash *slots;
    List *argNames = selfData->argNames;
    IoObject *scope = selfData->scope;
    
    if (!scope)
    { 
	scope = target;
    }
    
    /*printf("blockLocals %p\n", (void *)blockLocals); */
    IoObject_createSlotsIfNeeded(blockLocals);
    slots = blockLocals->slots;
    
    /* blockLocals must be white, so we don't need to use IOREF here */

    PHash_at_put_(slots, state->targetString, target);  
    PHash_at_put_(slots, state->selfString, scope); 
    PHash_at_put_(slots, state->senderString, locals);
    PHash_at_put_(slots, state->thisBlockString, self);
    PHash_at_put_(slots, state->thisMessageString, m);
    PHash_at_put_(slots, state->updateSlotString, state->localsUpdateSlotCFunc);
    
    /* eval args and put into locals */
    {
	int max = List_count(argNames);
	int i;
	
	for (i = 0; i < max; i++)  
	{
	    IoObject *arg = IoMessage_locals_valueArgAt_(m, locals, i);
	    IoString *name = (IoString *)LIST_AT_(argNames, i);
	    /* 
	    ok, so blockLocals is white *unless* the gc kicks 
	    in while evaling the args. So we need to IOREF the args to be safe 
	    */
	    IoObject_addingRef_(blockLocals, name);
	    IoObject_addingRef_(blockLocals, arg);

	    PHash_at_put_(slots, name, arg);
	    /*PHash_at_put_(slots, IOREF(name), IOREF(arg));*/ 
	}
    }  
    
    result = IoMessage_locals_performOn_(selfData->message, blockLocals, blockLocals);  
    
    state->stopStatus = MESSAGE_STOP_STATUS_NORMAL;
    
    return result;
}

IoObject *IoLocals_extTarget_locals_call_blockLocals_(IoLocals *self, 
						     IoObject *target, IoObject *locals, IoMessage *m, IoObject *blockLocals)
{ 
    return IoLocals_target_locals_call_blockLocals_state_(self, target, locals, m, blockLocals, IOSTATE); 
}

IoObject *IoLocals_target_locals_call_(IoLocals *self, 
				      IoObject *target, IoObject *locals, IoMessage *m)
{
    IoState *state = IOSTATE;
    IoObject *result;
    IoObject *blockLocals;
    IoState_pushRetainPool(state);
    {
	blockLocals = IoObject_rawClone(state->localsProto);
	blockLocals->mark.isLocals = 1;
	result = IoLocals_target_locals_call_blockLocals_state_(self, target, locals, m, blockLocals, state);
    }
    IoState_popRetainPoolExceptFor_(state, result);
    
#ifdef IO_BLOCK_RECYCLING
    if (result != blockLocals && !blockLocals->mark.isReferenced)
    { 
	IoState_immediatelyFree_(state, blockLocals); 
    }
#endif 
    
    return result;
}

/* ------------------------------------------------------------------------ */

IoObject *IoLocals_lobbyMethod(IoObject *target, IoObject *locals, IoMessage *m)
{
    IoLocals *self = IoLocals_new((IoState *)(target->tag->state));
    int nargs = IoMessage_argCount(m);
    int i;
    
    if (nargs == 0)
    { 
	IoState_error_description_(IOSTATE, m, 
	    "Io Block argumentMissing", "requires at least one argument"); 
    }
    
    for (i = 0; i < nargs - 1; i ++)
    {
	IoMessage *argMessage = IoMessage_rawArgAt_(m, i);
	IoString *name = IoMessage_name(argMessage);
	List_add_(DATA(self)->argNames, IOREF(name));
    }
    
    DATA(self)->message = (IoMessage *)IOREF(IoMessage_rawArgAt_(m, nargs-1));
    /*IoLocals_compile(self);*/
    return self;
}

IoObject *IoLocals_lobbyBlock(IoObject *target, IoObject *locals, IoMessage *m)
{
    IoLocals *self = (IoLocals *)IoLocals_lobbyMethod(target, locals, m);
    DATA(self)->scope = IOREF(locals);
    return self;
}

IoObject *IoLocals_print(IoLocals *self, IoObject *locals, IoMessage *m)
{ 
    /*
     method: print
     description: prints an Io source code representation of the block/method
     */
    ByteArray *ba = IoLocals_justCode(self);
    IoState_print_(IOSTATE, ByteArray_asCString(ba));
    return IONIL(self); 
}

ByteArray *IoLocals_justCode(IoLocals *self)
{ 
    ByteArray *ba = ByteArray_new();
    
    if (DATA(self)->scope)
    { 
	ByteArray_appendCString_(ba, "block("); 
    }
    else
    { 
	ByteArray_appendCString_(ba, "method("); 
    }
    
    {
	int i;
	
	for (i = 0; i < List_count(DATA(self)->argNames); i ++)
	{
	    IoString *argName = (IoString *)List_at_(DATA(self)->argNames, i);
	    ByteArray_append_(ba, IoString_rawByteArray(argName));
	    ByteArray_appendCString_(ba, ", ");
	}
    }
    
    {
	ByteArray *d = IoMessage_description(DATA(self)->message);
	ByteArray_append_(ba, d);
	ByteArray_free(d);
    }
    
    ByteArray_appendCString_(ba, ")");
    return ba; 
}

IoObject *IoLocals_code(IoLocals *self, IoObject *locals, IoMessage *m)
{ 
    /*
    method: code
    description: Returns a string containing the decompiled code of the receiver. 
    */
    
    return IoState_stringWithByteArray_copy_(IOSTATE, IoLocals_justCode(self), 0); 
}

IoObject *IoLocals_code_(IoLocals *self, IoObject *locals, IoMessage *m)
{ 
    /*
    undocumented: yes
    method: setCode
    description: 
    */
    
    IoString *string = IoMessage_locals_stringArgAt_(m, locals, 0);
    char *s = CSTRING(string);
    IoMessage *newM = IoMessage_newFromText_label_(IOSTATE, s, "[IoLocals_code_]");
    
    if (newM) 
    { 
	DATA(self)->message = IOREF(newM); 
    }
    else
    { 
	IoState_error_description_(IOSTATE, m, 
	    "Io Block compileError", "no messages found in compile string"); 
    }
    return self;
}

IoObject *IoLocals_message(IoLocals *self, IoObject *locals, IoMessage *m)
{
    /*
    method: message
    description: Returns the root message of the receiver. 
    */
    
    return DATA(self)->message ? (IoObject *)DATA(self)->message : IONIL(self); 
}

IoObject *IoLocals_setMessage(IoLocals *self, IoObject *locals, IoMessage *m)
{ 
    /*
    method: setMessage(aMessage)
    description: Sets the root message of the receiver to aMessage. 
    */
    
    IoMessage *message = IoMessage_locals_messageArgAt_(m, locals, 0);
    DATA(self)->message = IOREF(message);
    return self;
}

IoObject *IoLocals_argumentNames(IoLocals *self, IoObject *locals, IoMessage *m)
{
    /*
    method: argumentNames
    description: Returns a List of strings containing the argument names of the receiver. 
    */
    
    IoList *argsList = IoList_new(IOSTATE);
    int i;
    
    for (i = 0; i < List_count(DATA(self)->argNames); i ++)
    { 
	IoList_rawAdd_(argsList, (IoObject *)LIST_AT_(DATA(self)->argNames, i)); 
    }
    
    return argsList;
}

IoObject *IoLocals_argumentNames_(IoLocals *self, IoObject *locals, IoMessage *m)
{
    /*
     method: setArgumentNames(aListOfStrings)
     description: Sets the receiver's argument names to those specified in aListOfStrings. Returns self.  
     */
    IoList *newArgNames = IoMessage_locals_listArgAt_(m, locals, 0);
    IOASSERT(ISLIST(newArgNames), "argument must be a List");
    
    {
	unsigned int i;
	
	for (i = 0; i < IoList_rawCount(newArgNames); i ++)
	{ 
	    IOASSERT(ISSTRING(IoList_rawAt_(newArgNames, i)), "all argument names must be Strings"); 
	}
    }
    
    List_copy_(DATA(self)->argNames, IoList_rawList(newArgNames));
    return self;
}

IoObject *IoLocals_scope(IoLocals *self, IoObject *locals, IoMessage *m)
{ 
    /*
     method: scope
     description: Returns the scope used when the block is activated or Nil if the target of the message is the scope.   
     */
    IoObject *scope = DATA(self)->scope;
    return scope ? scope : IONIL(self); 
}

IoObject *IoLocals_setScope_(IoLocals *self, IoObject *locals, IoMessage *m)
{
    /*
     method: setScope(anObjectOrNil)
     description: If argument is an object, when the block is activated, it will set the proto and self slots of it's locals to the specified object. If Nil, it will set them to the target of the message. 
     */
    IoObject *scope = IoMessage_locals_valueArgAt_(m, locals, 0);
    DATA(self)->scope = ISNIL(scope) ? 0x0 : IOREF(scope);
    return self;
}

IoObject *IoLocals_ifTrue_(IoLocals *self, IoObject *locals, IoMessage *m)
{
    /*
     method: ifTrue(<message>)
     description: Evaluates message if receiver returns a non-Nil when activated in the current context. Returns self.
     */
    IOASSERT(IoMessage_argCount(m)==1, "method requires a single argument");
    
    if (!ISNIL(IoLocals_target_locals_call_(self, locals, locals, DATA(self)->message)))
    { 
	IoMessage_locals_performOn_(IoMessage_rawArgAt_(m, 0), locals, locals); 
    }
    return self;
}

IoObject *IoLocals_ifFalse_(IoLocals *self, IoObject *locals, IoMessage *m)
{
    /*
     method: ifFalse(<message>)
     description: Evaluates message if receiver returns a Nil when activated in the current context. Returns self.
     */
    IOASSERT(IoMessage_argCount(m)==1, "method requires a single argument");
    
    if (ISNIL(IoLocals_target_locals_call_(self, locals, locals, DATA(self)->message)))
    { 
	IoMessage_locals_performOn_(IoMessage_rawArgAt_(m, 0), locals, locals); 
    }
    return self;
}

IoObject *IoLocals_whileTrue_(IoLocals *self, IoObject *locals, IoMessage *m)
{
    /*
     method: whileTrue(<message>)
     description: Keeps evaluating message until receiver returns a Nil when activated in the current context. Returns self.
     */
    IOASSERT(IoMessage_argCount(m)==1, "method requires a single argument");
    
    while (!ISNIL(IoLocals_target_locals_call_(self, locals, locals, DATA(self)->message)))
    { 
	IoMessage_locals_performOn_(IoMessage_rawArgAt_(m, 0), locals, locals); 
    }
    return self;
}

IoObject *IoLocals_whileFalse_(IoLocals *self, IoObject *locals, IoMessage *m)
{
    /*
     method: whileFalse(<message>)
     description: Keeps evaluating message until receiver returns a non-Nil when activated in the current context. Returns self.
     */
    IOASSERT(IoMessage_argCount(m)==1, "method requires a single argument");
    
    while (ISNIL(IoLocals_target_locals_call_(self, locals, locals, DATA(self)->message)))
    { 
	IoMessage_locals_performOn_(IoMessage_rawArgAt_(m, 0), locals, locals); 
    }
    return self;
}

IoObject *IoLocals_performOn_(IoLocals *self, IoObject *locals, IoMessage *m)
{
    /*
     method: performOn(anObject)
     description: Activates the receiver in the target context of anObject. Returns the result.
     */
    IoObject *target = IoMessage_locals_valueArgAt_(m, locals, 0);
    
    if (IoMessage_argCount(m) > 1) 
    { 
	locals = IoMessage_locals_valueArgAt_(m, locals, 1); 
    }
    
    if (IoMessage_argCount(m) > 2) 
    { 
	m = IoMessage_locals_valueArgAt_(m, locals, 2); 
    }
    
    return IoLocals_target_locals_call_(self, target, locals, m);
}

