/*#io
Map ioDoc(
docCopyright("Steve Dekorte", 2002)
docLicense("BSD revised")    
docObject("Map")
docDescription("A key/value dictionary appropriate for holding large key/value collections.")
*/
 
#include "IoPMap.h"
#include "IoObject.h"
#include "IoState.h"
#include "IoCFunction.h"
#include "IoNil.h"
#include "IoSeq.h"
#include "IoState.h"
#include "IoNumber.h"
#include "IoList.h"
#include "IoBlock.h"

#define SDB(self) ((SkipDB *)self->data)

IoTag *IoPMap_tag(void *state)
{
    IoTag *tag = IoTag_newWithName_("PersistentMap");
    tag->state = state;
    tag->freeFunc  = (TagFreeFunc *)IoPMap_free;
    tag->cloneFunc = (TagCloneFunc *)IoPMap_rawClone;
    tag->markFunc  = (TagMarkFunc *)IoPMap_mark;
    tag->writeToStoreOnStreamFunc = (TagWriteToStoreOnStreamFunc *)IoPMap_writeToStore_stream_;
    tag->readFromStoreOnStreamFunc = (TagReadFromStoreOnStreamFunc *)IoPMap_readFromStore_stream_;
    return tag;
}

void IoPMap_writeToStore_stream_(IoMap *self, IoStore *store, BStream *stream)
{
    SkipDB *sdb = SDB(self);

}

void IoPMap_readFromStore_stream_(IoMap *self, IoStore *store, BStream *stream)
{
    SkipDB *sdb = SDB(self);
	
}


IoMap *IoPMap_proto(void *state)
{
    IoMethodTable methodTable[] = {
	{"empty",    IoPMap_empty}, 
	{"at",       IoPMap_at}, 
	{"atPut",    IoPMap_atPut}, 
	{"atIfAbsentPut", IoPMap_atIfAbsentPut}, 
	{"size",     IoPMap_size}, 
	{"keys",     IoPMap_keys}, 
	{"values",   IoPMap_values}, 
	{"foreach",  IoPMap_foreach}, 
	{"hasKey",   IoPMap_hasKey_}, 
	{"hasValue", IoPMap_hasValue_}, 
	{"removeAt", IoPMap_removeAt}, 
	{NULL, NULL},
    };

    IoObject *self = IoObject_new(state);

    self->tag = IoPMap_tag(state);
    self->data = SkipDB_new();

    IoState_registerProtoWithFunc_(state, self, IoPMap_proto);

    IoObject_addMethodTable_(self, methodTable);
    return self;
}

IoMap *IoPMap_rawClone(IoMap *proto) 
{ 
    IoObject *self = IoObject_rawClonePrimitive(proto);
    self->data = SkipDB_clone(SDB(proto));
    return self; 
}

IoMap *IoPMap_new(void *state)
{
    IoObject *proto = IoState_protoWithInitFunction_(state, IoPMap_proto);
    return IOCLONE(proto);
}

void IoPMap_free(IoMap *self) 
{
    SkipDB_free(SDB(self));
}

void IoPMap_mark(IoMap *self) 
{ 
    SkipDB_doOnKeyAndValue_(SDB(self), (ListDoCallback *)IoObject_makeGrayIfWhite); 
}

void IoPMap_rawAtPut(IoMap *self, IoSymbol *k, IoObject *v)
{
    SkipDB_at_put_(SDB(self), IOREF(k), IOREF(v));
}

Hash *IoPMap_rawSkipDB(IoMap *self)
{ 
    return SDB(self); 
}

/* ----------------------------------------------------------- */

IoObject *IoPMap_empty(IoMap *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("empty", "Removes all keys from the receiver. Returns self.")
    */
    
    SkipDB_clean(SDB(self)); 
    return self; 
}

IoObject *IoPMap_rawAt(IoMap *self, IoSymbol *k)
{
    void *result = SkipDB_at_(SDB(self), k);
    return (result) ? (IoObject *)result : IONIL(self);
}

IoObject *IoPMap_at(IoMap *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("at(keyString, optionalDefaultValue)", 
    "Returns the value for the key keyString. Returns nil if the key is absent. ")
    */
    
    IoSymbol *k = IoMessage_locals_symbolArgAt_(m, locals, 0);
    void *result = SkipDB_at_(SDB(self), k);
    
    if (!result && IoMessage_argCount(m) > 1) 
    { 
	return IoMessage_locals_valueArgAt_(m, locals, 1); 
    }
    
    return (result) ? (IoObject *)result : IONIL(self);
}

IoObject *IoPMap_atPut(IoMap *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("atPut(keyString, aValue)", 
    "Inserts/sets aValue with the key keyString. Returns self. ")
    */

    IoSymbol *k = IoMessage_locals_symbolArgAt_(m, locals, 0);
    IoObject *v  = IoMessage_locals_valueArgAt_(m, locals, 1);
    IoPMap_rawAtPut(self, k, v);
    return self;
}

IoObject *IoPMap_atIfAbsentPut(IoMap *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("atIfAbsentPut(keyString, aValue)", 
    "Inserts/sets aValue with the key keyString if the 
    key is not already present. Returns self. ")
    */
    
    IoSymbol *k = IoMessage_locals_symbolArgAt_(m, locals, 0);
    void *result = SkipDB_at_(SDB(self), k);
    
    if (result) 
    {
	return result;
    }
    else
    {
	IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 1);
	IoPMap_rawAtPut(self, k, v);
	return v;
    }
}

IoObject *IoPMap_size(IoMap *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("size", 
    "Returns the number of key/value pairs in the receiver.") 
    */
    
    return IONUMBER(SkipDB_count(SDB(self))); 
}

IoObject *IoPMap_hasKey_(IoMap *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("hasKey(keyString)", 
    "Returns self if the key is present or Nil otherwise.")
    */
    
    IoSymbol *k = IoMessage_locals_symbolArgAt_(m, locals, 0);
    return SkipDB_at_(SDB(self), k) ? (IoObject *)self : IONIL(self);
}

IoObject *IoPMap_removeAt(IoMap *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("removeAt(keyString)", 
    "Removes the specified keyString if present. Returns self. ")
    */
    
    IoSymbol *k = IoMessage_locals_symbolArgAt_(m, locals, 0);
    SkipDB_removeKey_(SDB(self), k);
    return self;
}

IoObject *IoPMap_hasValue_(IoMap *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("hasValue(aValue)", 
    "Returns self if the value is one of the Map's values or Nil otherwise.")
    */
    
    IoList *values = IoPMap_values(self, locals, m);
    return ISNIL(IoList_contains(values, locals, m)) ? IONIL(self) : self;
}

IoList *IoPMap_rawKeys(IoMap *self)
{
    IoList *list = IoList_new(IOSTATE);
    register IoObject *key = (IoObject *)SkipDB_firstKey(SDB(self));

    while (key)
    {
	IoList_rawAppend_(list, key);
	key = SkipDB_nextKey(SDB(self));
    }
    
    return list;
}

IoList *IoPMap_keys(IoMap *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("keys", "Returns a List of the receivers keys.")
    */
    
    return IoPMap_rawKeys(self);
}

IoObject *IoPMap_values(IoMap *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("values", 
    "Returns a List of the receivers values.")
    */
    
    IoList *list = IoList_new(IOSTATE);
    register IoObject *key = (IoObject *)SkipDB_firstKey(SDB(self));
    
    while (key)
    {
	IoList_rawAppend_(list, (IoObject *)SkipDB_at_(SDB(self), key));
	key = SkipDB_nextKey(SDB(self));
    }
    
    return list;
}

IoObject *IoPMap_foreach(IoMap *self, IoObject *locals, IoMessage *m)
{
    /*#io
    docSlot("foreach(optionalKey, value, message)", 
    """For each key value pair, sets the locals key to 
    the key and value to the value and executes message.
    Example:
    <pre>aMap foreach(k, v, writeln(k, " = ", v))
aMap foreach(v, write(v))</pre>
    
    Example use with a block:
    
    <pre>myBlock = block(k, v, write(k, " = ", v, "\n")) 
aMap foreach(k, v, myBlock(k, v))</pre>""")
    */

    IoState *state = IOSTATE;
    IoSymbol *keyName;
    IoSymbol *valueName;
    IoMessage *doMessage;
    IoObject *key = (IoObject *)SkipDB_firstKey(SDB(self));
    IoObject *result = IONIL(self);    
    IoMessage_foreachArgs(m, self, &keyName, &valueName, &doMessage);
    IoState_pushRetainPool(state);

    while (key)
    {
        IoState_clearTopPool(state);
        {
            IoObject *value = (IoObject *)SkipDB_at_(SDB(self), key);
            
            if (!ISNIL(keyName))
            {
                IoObject_setSlot_to_(locals, keyName, key);
            }
            IoObject_setSlot_to_(locals, valueName, value);
            IoMessage_locals_performOn_(doMessage, locals, locals);
            key = SkipDB_nextKey(SDB(self));
            
            if (IoMessage_handleStatus(doMessage)) 
            {
                goto done;
            }
        }
    }
    
    done:
    IoState_popRetainPoolExceptFor_(state, result);
    return result;
}
