/*   
    copyright: Steve Dekorte, 2002. All rights reserved.
    license: See _BSDLicense.txt.
    
    object: PObject
    prototype: Object
    module: IoVM
    description: Persistent Object used by Store as a placeholder for an on-disk object. When a message is sent to it, it loads itself from disk and becomes the deserialized version of the on-disk object.
*/

#include "IoPObject.h"
#include "IoObject.h"
#include "IoObject_persistence.h"
#include "IoState.h"
#include "IoStore.h"

#define DATA(self) ((IoPObjectData *)self->data)

IoTag *IoPObject_tag(void *state)
{
  IoTag *tag = IoTag_newWithName_("PObject");
  tag->state = state;
  tag->cloneFunc = (TagCloneFunc *)IoPObject_rawClone;
  tag->markFunc  = (TagMarkFunc *)IoPObject_mark;
  tag->freeFunc  = (TagFreeFunc *)IoPObject_free;
  return tag;
}

IoPObject *IoPObject_proto(void *state)
{
  IoMethodTable methodTable[] = {
    {"forward", IoPObject_forward},
    {"isWaitingPObject", IoObject_self},
    {"?", IoObject_question},
    /*
      method: ?(<message>)
      description: Same as Object ? operator.
    */
    {NULL, NULL},
  };
  
  IoObject *self = IoObject_new(state);

  self->data = calloc(1, sizeof(IoPObjectData));
  self->tag = IoPObject_tag(state);
  /* rest of data expected to be NULL */
  IoState_registerProtoWithFunc_(state, self, IoPObject_proto);
      
  IoObject_addMethodTable_(self, methodTable);
  return self;
}

IoPObject *IoPObject_rawClone(IoPObject *self) 
{ 
  IoPObject *child = IoObject_rawClonePrimitive(self);
  child->data = cpalloc(self->data, sizeof(IoPObjectData));
  {
  IoObject *forward = IoObject_getSlot_(self, IOSTATE->forwardString);
  
  IoObject_setSlot_to_(child, IOSTATE->forwardString, forward); /* so we have slots of our own */
  PHash_copy_(child->slots, self->slots);

  IoObject_removeSlot_(child, IOSTATE->protoString);
  }
  return child; 
}

IoPObject *IoPObject_new(void *state)
{
  IoPObject *proto = IoState_protoWithInitFunction_(state, IoPObject_proto);
  return IoPObject_rawClone(proto);
}

void IoPObject_free(IoPObject *self)
{
  /*IoPObject *proto = IoState_protoWithInitFunction_(IOSTATE, IoPObject_proto);*/
  if (NULL == DATA(self)) { return; }
  free(DATA(self));
  self->data = NULL;
}


IoPObject *IoPObject_newWithStore_pid_(IoState *state, IoStore *store, int pid)
{
  IoPObject *self = IoPObject_new(state);
  DATA(self)->store = IOREF(store);
  IoObject_setPersistentId_(self, pid);
  return self;
}

void IoPObject_mark(IoPObject *self)
{ 
  if (DATA(self)->store) IoObject_makeGrayIfWhite((IoObject *)(DATA(self)->store));
}

void IoPObject_rawSetValue_(IoPObject *self, IoObject *v) 
{   
  /* a little magic to turn the future into the result value */
  /* we empty the slots and set the proto slot to the result */
  IoObject_clean(self);
  /*IoObject_removeSlot_(self, IOSTATE->forwardString);*/
  IoObject_setSlot_to_(self, IOSTATE->protoString, v);
  
  /* set the tag and data to the result tag and data */
  /* need to add something to prevent proto slot from changing */
  /* in order to avoid freeing the data while we still need it */
  self->tag = v->tag;
  free(DATA(self));
  self->data = v->data;
  self->extraData = v->extraData;
  self->mark.doesNotOwnData = 1;
}


/* ----------------------------------------------------------- */

inline int IoPObject_rawIsReady(IoPObject *self)
{ return (self->proto != 0x0); }

int IoPObject_rawWaitOnResult(IoPObject *self)
{ 
  int pid = IoObject_persistentId(self);
  IoObject *obj = IoStore_objectWithPid_(DATA(self)->store, pid);
  IoPObject_rawSetValue_(self, obj);
  return 0;
}

IoObject *IoPObject_forward(IoPObject *self, IoObject *locals, IoMessage *m)
{ 
  /*
    method: forward
    description: Implements Future's waiting behavior. 
  */
  IoPObject_rawWaitOnResult(self);
  return IoObject_perform(self, locals, m);
}

